﻿using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Win32.SafeHandles;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Runtime.Serialization;
using System.Threading;
using System.Runtime.InteropServices;
using System.Security.AccessControl;
using System.Reflection;

using PapiWrapper;
using PapiWrapper.Common;
using PapiWrapper.Fmb;
using PapiWrapper.Dp;

using GlobalModule;
using HelperModule;
using FmbModule;
using DpModule;

namespace DpDemo
{
    /// <summary>
    /// Main class of DP Demo wrapper app.
    /// </summary>
    public class DPDemo
    {
        #region Fields

        private Global _global;
        private string _binFilePath;

        private Thread _keyboardThread;
        private Thread _readServiceInterfaceThread;
        private Thread _basicManagementThread;

        private static DPDemo _instance;

        private static WaitHandle[] _waitReadServiceHandles;

        #endregion

        #region Constructor
        private DPDemo()
        {
            _global = Global.Instance;
            _waitReadServiceHandles = new WaitHandle[2] { Global.EwhRead, Global.EwhExit };
        }
        #endregion

        #region Properties
        public static DPDemo Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new DPDemo();
                }
                return _instance;
            }
        }
        public static bool HasConfigFile { get; set; }

        #endregion

        #region Public methods
        /// <summary>
        ///------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        ///------------------------------------------------------------------------
        /// The main function controls the initialization and termination of the
        /// program.
        /// The BasicManagementThread starts the whole PROFIBUS activity.
        /// The KeyboardThread checks the keyboard for a termination character to
        /// quit the program at any time.
        /// If one of these two threads terminates then the whole application will
        /// be terminated.
        /// This function is responsible for calling the nested initialization and
        /// termination routines of the sub modules.
        ///
        /// Any 'open' devices of the PROFIBUS driver will be closed automatically
        /// after program termination. This will help to keep the thread structure
        /// incomplicated because there is no need to wait for several events at
        /// the same time (e.g. PROFIBUS message and QUIT message simultaneously).
        ///
        /// If you need some specific cleanup during program termination you can
        /// do this module-global by means of the TerminateXxx function.
        /// If you prefer to wait for multiple events please refer to the PROFIBUS
        /// driver manual. This task is not demonstrated here.
        ///------------------------------------------------------------------------
        /// </summary>
        public void Run(string[] args)
        {
            if (InitDpDemo())
            {
                ShowProgramInfo();
                if (ParseCommandLine(args))
                {
                    InitFmbLayer();

                    if (!OpenProfibusApi(2000,0))
                    {
                        Global.OutputString("Initializing PROFIBUS controller failed.");
                        DpLayer.Instance.TerminateDpLayer();
                        FmbLayer.Instance.TerminateFmbLayer();

                        TerminateDpDemo();
                        return;
                    }

                    if (HasConfigFile)
                    {
                        if (!ImportBinaryConfigurationFile(_binFilePath))
                        {
                            Global.OutputString("Import DP configuration parameters failed.");
                            DpLayer.Instance.TerminateDpLayer();
                            FmbLayer.Instance.TerminateFmbLayer();

                            TerminateDpDemo();
                            return;
                        }
                    }
                    else
                    {
                        DpLayer.Instance.DpSetPreConfigSlaveParaSets();
                    }


                    ShowProgramParameter();
                    ShowReleaseInfo();

                    _keyboardThread = new Thread(KeyboardThread);
                    _keyboardThread.Start();

                    _readServiceInterfaceThread = new Thread(ReadServiceInterfaceThread);
                    _basicManagementThread = new Thread(FmbLayer.Instance.BasicManagementThread);

                    _readServiceInterfaceThread.Start();
                    _basicManagementThread.Start();

                    _keyboardThread.Join();
                    _readServiceInterfaceThread.Join();
                    _basicManagementThread.Join();

                    //--- terminate sub modules and free globals -----------------------------
                    DpLayer.Instance.TerminateDpLayer();
                    FmbLayer.Instance.TerminateFmbLayer();

                    CloseProfibusApi();

                    TerminateDpDemo();
                }
            }
        }

        #endregion

        #region Private methods

        /// <summary>
        ///------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        ///------------------------------------------------------------------------
        /// The DpDemo application allows several command line parameters.
        /// Please refer to function ShowCommandLineUsage for details.
        ///------------------------------------------------------------------------
        /// </summary>
        private bool ParseCommandLine(string[] args)
        {
            int argc = args.Length;
            if (argc > 4)
            {
                ShowCommandLineUsage(args[0]);
            }

            for (int argIndex = 0; argIndex < argc; argIndex++)
            {
                if (args[argIndex].StartsWith("-b"))
                {
                    Global.BoardNumber = Convert.ToByte(args[argIndex].Remove(0,2));
                }
                else if (args[argIndex].StartsWith("-m"))
                {
                    Global.MasterAddr = Convert.ToByte(args[argIndex].Remove(0, 2));
                }
                else if (args[argIndex] == "-r0")
                {
                    Global.Baudrate = 0x00;  // 9.6 kBaud
                }
                else if (args[argIndex] == "-r1")
                {
                    Global.Baudrate = 0x01;  // 19.2 kBaud
                }
                else if (args[argIndex] == "-r2")
                {
                    Global.Baudrate = 0x02;  // 93.75 kBaud
                }
                else if (args[argIndex] == "-r2PF")
                {
                    Global.Baudrate = 0x0C;  // 93.75 kBaud
                }
                else if (args[argIndex] == "-r3")
                {
                    Global.Baudrate = 0x03;  // 187.5 kBaud
                }
                else if (args[argIndex] == "-r4")
                {
                    Global.Baudrate = 0x04;  // 500 kBaud
                }
                else if (args[argIndex] == "-r5")
                {
                    Global.Baudrate = 0x05;  // 750 kBaud
                }
                else if (args[argIndex] == "-r6")
                {
                    Global.Baudrate = 0x06;  // 1.5 MBaud
                }
                else if (args[argIndex] == "-r7")
                {
                    Global.Baudrate = 0x07;  // 3 MBaud
                }
                else if (args[argIndex] == "-r8")
                {
                    Global.Baudrate = 0x08;  // 6 MBaud
                }
                else if (args[argIndex] == "-r9")
                {
                    Global.Baudrate = 0x09;  // 12 MBaud
                }
                else if (args[argIndex] == "-rB")
                {
                    Global.Baudrate = 0x0B;  // 45.45 kBaud
                }
                else if (args[argIndex].Contains("-h") ||
                         args[argIndex].Contains("-?") ||
                         args[argIndex][0] == '-')
                {
                    ShowCommandLineUsage(args[0]);
                    return false;
                }
                else
                {
                    _binFilePath = args[argIndex];
                    HasConfigFile = true;
                }
            }

            if (!string.IsNullOrEmpty(_binFilePath))
            {
                if (Helper.GetFileAttributes(_binFilePath) == 0xFFFFFFFF)
                {
                    Global.OutputString(string.Format("The file \"{0}\" could not be found.", _binFilePath));
                    return false;
                }

                HasConfigFile = true;
                return true;

            } // if binfile

            return true;
        }  // ParseCommandLine



        //**************************************************************************
        private void ShowCommandLineUsage(string  pProgramPath)
        {
            Console.WriteLine("usage: [-b<BoardNumber>] [-m<MasterAddress>] [-r<baudrate>] [binfile]\n");//, pProgramPath);
            Console.WriteLine("       -b       use PROFIBUS board with \"Board Number\"\n");
            Console.WriteLine("       -m       master address \n");
            Console.WriteLine("       -r       baudrate  (default 1.5 MBaud)                   \n");
            Console.WriteLine("                -r0 (  9.6 kBaud)   -r1 ( 19.2 kBaud)   -r2   (93.75 kBaud) \n");
            Console.WriteLine("                                                        -r2PF (93.75 kBaud Pepperl&Fuchs) \n");
            Console.WriteLine("                -r3 (187.5 kBaud)   -r4 (  500 kBaud)   -r6   (  1.5 MBaud) \n");
            Console.WriteLine("                -r7 (    3 MBaud)   -r8 (    6 MBaud)   -r9   (   12 MBaud) \n");
            Console.WriteLine("       -h -?    show this message\n");
            Console.WriteLine("       binfile  use binary PROFIBUS DP configuration file\n");
            Console.WriteLine("\n");
        } // ShowCommandLineUsage



        //**************************************************************************

        private void ShowProgramInfo()
        {
           Global.OutputString("DpDemo V5.46 - Copyright (C) SOFTING Industrial Automation GmbH 2011-2012");
           Global.OutputString("");
        } /* ShowProgramInfo */


        //**************************************************************************


        private void ShowProgramParameter()
        {

           Global.OutputString("");
           Global.OutputString(string.Format("Board Number:     {0}", Global.BoardNumber));
           Global.OutputString(string.Format("Master Address:   {0}",Global.MasterAddr));


           switch (Global.Baudrate)
           {
             case 0x00: Global.OutputString("Baudrate:         9.6 kBaud") ; break;
             case 0x01: Global.OutputString("Baudrate:         19.2 kBaud"); break;
             case 0x02: Global.OutputString("Baudrate:         93.75 kBaud"); break;
             case 0x03: Global.OutputString("Baudrate:         187.5 kBaud"); break;
             case 0x04: Global.OutputString("Baudrate:         500 kBaud"); break;
             case 0x05: Global.OutputString("Baudrate:         750 kBaud"); break;
             case 0x06: Global.OutputString("Baudrate:         1.5 MBaud"); break;
             case 0x07: Global.OutputString("Baudrate:         3 MBaud"); break;
             case 0x08: Global.OutputString("Baudrate:         6 MBaud"); break;
             case 0x09: Global.OutputString("Baudrate:         12 MBaud"); break;
             case 0x0B: Global.OutputString("Baudrate:         45.45 kBaud"); break;
             case 0x0C: Global.OutputString("Baudrate:         93.75 kBaud Pepperl&Fuchs"); break;
             default:   Global.OutputString("wrong Baudrate"); break;
           }
        } /* ShowProgramParameter */


        /// <summary>
        ///------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        ///------------------------------------------------------------------------
        /// Place all module specific initialization here. This function is called
        /// before any communication is started or initialized.
        ///------------------------------------------------------------------------
        /// </summary>
        private void InitFmbLayer()
        {
            if (!HasConfigFile)
            {
                FmbLayer.Instance.InitFmbLayer(Global.Baudrate);
            }
        }

        /// <summary>
        ///------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        ///------------------------------------------------------------------------
        /// Place all module specific initialization here. This function is called
        /// before any communication is started or initialized.
        ///------------------------------------------------------------------------
        /// </summary>
        private bool InitDpDemo()
        {
            Global.Baudrate = 0x6; // 1.5 MB by default

            bool bResult = Global.OpenPapiAndLogger();
            if (bResult)
            {
                Global.OutputSetThreadIdString(" MAIN");
            }
            return bResult;
        }


        static bool bStopped;
        private void KeyboardThread()
        {
            Global.OutputSetThreadIdString("  KEY");
            Global.OutputString("KeyboardThread started");

            while (true)
            {
                ConsoleKeyInfo pressedKey = Console.ReadKey(true);
                if (pressedKey.KeyChar == 'Q' || pressedKey.KeyChar == 'X' ||
                    pressedKey.KeyChar == 'q' || pressedKey.KeyChar == 'x')
                {
                  bStopped = true;
                  Global.EwhExit.Set();
                  break;
                }
                else if (pressedKey.KeyChar == 'I' || pressedKey.KeyChar == 'i')
                {
                  DpLayer.EwhDPInputData.Set();
                }
                else if (pressedKey.KeyChar == 'O' || pressedKey.KeyChar == 'o')
                {
                  DpLayer.EwhDPOutputData.Set();
                }
            }

            Global.OutputString("KeyboardThread terminated");

        } // KeyboardThread


        /// <summary>
        /// Starts reading service states (indication/confirmation)
        /// </summary>
        public void ReadServiceInterfaceThread()
        {
            Global.OutputSetThreadIdString("RDSRV");
            Global.OutputString("ReadServiceInterfaceThread started");
            Global.CanReadServiceInterface = true;
            while (!bStopped)
            {
                int waitResult = WaitHandle.WaitAny(_waitReadServiceHandles, 1, false);
                if (waitResult == 0)
                {
                    if (Global.CanReadServiceInterface == true)
                    {
                       Global.ReadServiceInterface();
                    }
                }
                else if (waitResult == 1)
                {
                    Global.EwhExit.Set();
                    break;
                }
                else if (waitResult == WaitHandle.WaitTimeout)
                {
                    Global.EwhRead.Set();
                }
            }

            Global.OutputString("ReadServiceInterfaceThread terminated");

        } // ReadServiceInterfaceThread


        /// <summary>
        /// Show PAPI and PROFIBUS firmware version
        /// </summary>
        public void ShowReleaseInfo()
        {
            Global.OutputString("PAPI Version:     " + Global.GetPapiVersion().Substring(0, 50));
            Global.OutputString("Firmware Version: " + Global.GetFirmwareVersion().Substring(0, Math.Min(50, Global.GetFirmwareVersion().Length)));
            Global.OutputString("");
        } // ShowReleaseInfo



        //**************************************************************************
        //** INITIALIZATION / TERMINATION ******************************************
        //**************************************************************************


        //------------------------------------------------------------------------
        // FUNCTIONAL_DESCRIPTION
        //------------------------------------------------------------------------
        // - open logical PROFIBUS devices
        //------------------------------------------------------------------------
        private bool OpenProfibusApi(uint readTimeout, uint writeTimeout)
        {
            return Global.PapiInit(Global.BoardNumber, readTimeout, writeTimeout);
        } // OpenProfibusDevices

        /// <summary>
        ///------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        ///------------------------------------------------------------------------
        /// - close logical PROFIBUS devices
        ///------------------------------------------------------------------------
        /// </summary>
        private void CloseProfibusApi()
        {
            Global.PapiEnd(Global.BoardNumber);

        }  // CloseProfibusDevices

        /// <summary>
        ///------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        ///------------------------------------------------------------------------
        /// Place cleanup operations necessary for this module here.
        ///------------------------------------------------------------------------
        /// </summary>
        private void TerminateDpDemo()
        {
            Global.OutputString("main thread terminated");

            DpLayer.EwhDPInputData.Close();
            DpLayer.EwhDPOutputData.Close();
            Global.CloseLoggerAndHandlers();
        } // TerminateDpDemo

        //**************************************************************************

        /// <summary>
        ///------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        ///------------------------------------------------------------------------
        /// This function imports a Softing configuration file. These files are
        /// created by the Softing PROFIBUS DP Configurator. Also the MS DOS
        /// application CNF2BIN is able to convert ASCII configuration files (CNF)
        /// into the binary configuration format.
        ///
        /// The routine will load all descriptions (bus and slave parameter sets)
        /// found in the binary file into the respective working structures of
        /// this application. This is done by means of the functions:
        /// - FmbSetBusParaSet,
        /// - DpSetBusParaSet,
        /// - DpSetSlaveParaSet.
        ///
        /// Note, that there exist two versions of the binary configuration
        /// format: an aligned and a not aligned version. Both formats can be
        /// handled by this routine.
        ///------------------------------------------------------------------------
        /// </summary>

        private static bool ImportBinaryConfigurationFile(string binFile)
        {
           try
           {
              Global.OutputString(string.Format("Import DP configuration from file {0}", binFile));

              C_DP_BUS_PARA_SET                     dpBusParaSet    = null;
              Dictionary<byte, C_DP_SLAVE_PARA_SET> dpSlaveParaSets = new Dictionary<byte, C_DP_SLAVE_PARA_SET>();

              if (Global.ImportBinaryDpConfigurationFile(binFile, out dpBusParaSet, dpSlaveParaSets))
              {
                 // store the configuration parameters
                 FmbLayer.Instance.FmbSetBusParaSet(dpBusParaSet);
                 DpLayer.Instance.DpSetBusParaSet(dpBusParaSet);
                 DpLayer.Instance.DpUpdateSetSlaveParaSet(dpSlaveParaSets);
              }
              else
              {
                 Global.DisplayLastError();
                 return false;
              }
           }
           catch (Exception exc)
           {
              Global.OutputString(exc.Message);
              return false;
           }

           return true;
       } /* ImportBinaryConfigurationFile */

       #endregion
    }
}
