
using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Threading;
using PapiWrapper.Common;
using PapiWrapper.Dp;

using DpDemo;
using GlobalModule;
using HelperModule;


namespace DpModule
{
    /// <summary>
    /// DP slave description holder class
    /// </summary>
    public class DpSlaveDesc
    {
        #region Fields
        private byte  _slaveNumber;

        public byte   nNumberInputs;
        public byte   nNumberOutputs;
        public ushort OffsetInputs;
        public ushort OffsetOutputs;

        public C_DP_SLAVE_PARA_SET SlaveParaSet;
        #endregion

        #region Constructor
        public DpSlaveDesc(byte slaveNumber)
        {
           _slaveNumber = slaveNumber;
        }
        #endregion

        #region Properties
        public byte SlaveNumber
        {
           get { return _slaveNumber; }
        }
        #endregion
    };

    /// <summary>
    /// **************************************************************************
    /// ** PRECONFIGURED SLAVE PARAMETER SET DESCRIPTION *************************
    /// **************************************************************************
    /// If you have created a slave parameter set of your desired DP slave
    /// or you want to use one the above listed slaves just enter your slave
    /// in the 'dp_slave_descr_table'.
    /// All slaves of this table will be preconfigured slaves (so called preconfigured
    /// slaves). These DP slaves have their fixed settings and will not be
    /// used during the auto detection process (preconfigured slaves have higher
    /// priority than auto detected slaves).
    /// Use this table also if you have a DP slave which needs special USER_PRM
    /// data (see function 'DpSlaveDetection' for more details).
    ///**************************************************************************
    /// </summary>
    public class C_PRECONFIG_DP_SLAVE_DESCR
    {
      public byte                SlaveAdd;
      public C_DP_SLAVE_PARA_SET DPSlaveParaSetPtr;
    };


    /// <summary>
    /// This module contains all functionality which is relevant to the DP protocol.
    /// The module covers DP master initialization, seeking for DP slaves (auto detection),
    /// downloading slave parameter sets and handling of the DP master operational states
    /// (STOP, CLEAR, OPERATE).
    ///
    /// Responsible for all theses features is the DpManagementThread. If successful it
    /// will start the two threads which deal with DP slave input/output data
    /// (DpInputDataThread, DpOuputDataThread).
    /// </summary>
    public class DpLayer
    {
        #region Fields

        #region Constants

        private const uint DpMaxServiceRetry = 10;

        //--- CFG IDENTIFIER DEFINITIONS (EN 50170/DP) ---------------------------

        private const byte DpIdentifierConsistency = 0x80;
        private const byte DpIdentifierLen         = 0x40;

        private const byte DpIdentifierIOSpecial   = 0x00;
        private const byte DpIdentifierIOInp       = 0x10;
        private const byte DpIdentifierIOOutp      = 0x20;
        private const byte DpIdentifierIOInpOutp   = 0x30;
        private const byte DpIdentifierIO          = 0x30;

        private const byte DpIdentifierDataLen     = 0x0F;
        private const byte DpIdentifierSpecLen     = 0x3F;

        private const byte DpIdentifierFreePlace   = 0x00;
        private const byte DpIdentifierInpLen      = 0x40;
        private const byte DpIdentifierOutpLen     = 0x80;
        private const byte DpIdentifierInpOutpLen  = 0xC0;
        private const byte DpIdentifierSpecial     = 0xC0;

        #endregion

        #region Default object size(s)

        private const int SizeOfSlaveParaSet = 16;
        private const int WaitFailed         = 0x7FFFFFFF;

        #endregion

        #region DP data area

        private C_PRECONFIG_DP_SLAVE_DESCR[] _dpSlaveDescrTable;

        private C_DP_SLAVE_PARA_SET          _dpSlaveSiemensET200B;
        private C_DP_SLAVE_PARA_SET          _dpSlaveSoftingB205;
        private C_DP_SLAVE_PARA_SET          _dpSlaveSofting0B35;

        //--- DP specific bus parameter extensions ---------------------------------
        private static C_DP_SET_BUSPARAMETER_REQ  _dpBusparameter = null;

        //--- DP slave administration memory -------------------------------------
        C_DP_SLAVE_PARA_SET   _dpSlaveParaSetTemplate = new C_DP_SLAVE_PARA_SET();
        C_DP_SLAVE_PARA_SET[] _slaveParaSets          = new C_DP_SLAVE_PARA_SET[DPConstants.DefaultSlaveAddress + 1];

        private static DpSlaveDesc[] _slaveDescr = new DpSlaveDesc[DPConstants.DefaultSlaveAddress + 1];
        private static byte          _nNumberOfSlaves = 0;

        private static DPOPMode _nCurrentOpMode   = DPOPMode.Offline;
        private static DPOPMode _nRequestedOpMode = DPOPMode.Offline;
        private static DPOPMode _nDesiredOpMode   = DPOPMode.Offline;

        #endregion


        private        Thread          _dpOutputDataThread;
        private        Thread          _dpInputDataThread;

        private static DpLayer         _instance;

        private static EventWaitHandle _ewhDpInputData;
        private static EventWaitHandle _ewhDpOutputData;

        private static WaitHandle[]    _waitDpReadServiceHandles;
        private static WaitHandle[]    _waitDpInputDataHandles;
        private static WaitHandle[]    _waitDpOutputDataHandles;

        private C_PROFI_SERVICE_DESCR  _profiSdb = new C_PROFI_SERVICE_DESCR();
        #endregion

        #region Constructor
        private DpLayer()
        {
            InitDpLayer();

            _ewhDpInputData  = new EventWaitHandle(false, EventResetMode.AutoReset);
            _ewhDpOutputData = new EventWaitHandle(false, EventResetMode.AutoReset);

            _waitDpReadServiceHandles = new WaitHandle[2] { Global.EwhDpRead, Global.EwhExit };
            _waitDpInputDataHandles   = new WaitHandle[2] { _ewhDpInputData,  Global.EwhExit };
            _waitDpOutputDataHandles  = new WaitHandle[2] { _ewhDpOutputData, Global.EwhExit };

            Global.DPServiceConIndReceived += new EventHandler<EventArgs<C_PROFI_SERVICE_DESCR>>(DPServiceConIndReceived);
        }
        #endregion

        #region Properties
        public static DpLayer Instance
        {
           get
           {
              if (_instance == null)
              {
                 _instance = new DpLayer();
              }
              return _instance;
           }
        }

        public static EventWaitHandle EwhDPInputData
        {
           get { return _ewhDpInputData; }
        }

        public static EventWaitHandle EwhDPOutputData
        {
           get { return _ewhDpOutputData; }
        }
        #endregion


        #region Public methods

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This thread is responsible for initializing the DP protocol component
        /// and handles all DP specific events (e.g. diagnostic data).
        /// First the DP management device is opened.
        /// The DP slave detection process seeks for additional slaves on the
        /// PROFIBUS. All identified DP slaves are added to the slave description
        /// list ('SlaveDescr').
        ///
        /// This function initializes the DP master (DP_Init_Master service) and
        /// downloads all slaves recorded in the slave description list.
        /// After that the master can be brought in the operation mode OPERATE
        /// (via STOP and CLEAR).
        /// If the master is in OPERATE state the DP slave input/output data
        /// processing threads are started to handle the data exchange.
        /// ------------------------------------------------------------------------
        /// </summary>
        public void DpManagementThread()
        {
            byte[] profiData = new byte[Global.MaxPduDataBufferLength];

            byte nMasterAdd;
            uint deviceSlave;

            Global.OutputSetThreadIdString("   DP");
            Global.OutputString("DpManagementThread started.");

            // increase thread priority for dp layer thread
            Thread.CurrentThread.Priority++;

            // auto detection of DP slaves on the PROFIBUS
            if (!DPDemo.HasConfigFile)
            {
               if (!DpSlaveDetection())
               {
                  Global.OutputString("DP slave detection failed");
                  return;
               }
            }

            // initialize DP Master
            if (!DpInitMasterReq())
            {
               Global.DisplayLastError();
               return;
            }
            if (!DpReadServiceInterface())
            {
               Global.DisplayLastError();
               return;
            }
            DpHandleEvent();


            //  set DP specific bus parameters -------------------------------------
            nMasterAdd = Global.MasterAddr;

            if (!DpSetBusparameterReq())
            {
                Global.DisplayLastError();
                return;
            }
            if (!DpReadServiceInterface())
            {
                Global.DisplayLastError();
                return;
            }
            DpHandleEvent();


            // create logical DP slave devices and download the slave parameter sets

            List<DpSlaveDesc> slaveDescrList = new List<DpSlaveDesc>(_slaveDescr);
            DpSlaveDesc[] slaveDescrArray = slaveDescrList.FindAll(slaveActive => slaveActive.SlaveParaSet != null).ToArray();

            byte nSlaveAdd = 0;
            foreach(DpSlaveDesc slaveDescr in  slaveDescrArray)
            {
                if (slaveDescr.SlaveParaSet != null)    // slave available
                {
                    nSlaveAdd   = slaveDescr.SlaveNumber;
                    deviceSlave = nSlaveAdd;

                    // download slave parameter sets

                    if (!DpStartSeqLoc(nMasterAdd, nSlaveAdd))
                    {
                       return;
                    }

                    if (!DpDownloadLoc(nSlaveAdd, slaveDescr.SlaveParaSet))
                    {
                       return;
                    }

                    if (!DpEndSeqLoc(nMasterAdd))
                    {
                       return;
                    }

                    //--- get slave parameter slave info ---------------------------------
                    if (!DpGetSlaveParam(nSlaveAdd))
                    {
                       Global.DisplayLastError();
                       return; // (1);
                    }
                } // if pSlaveParaSet
            }  // for slaves


            //  enter operation mode STOP and lock DP configuration
            _nDesiredOpMode = DPOPMode.Operate;
            if (!DpActParamLocReq(DPOPMode.Stop))
            {
                Global.DisplayLastError();
                return;
            }
            if (!DpReadServiceInterface())
            {
                Global.DisplayLastError();
                return;
            }
            DpHandleEvent();


            //  create DP Slave data handling threads (possible from state STOP)
            _dpOutputDataThread = new Thread(DpOutputDataThread);
            _dpOutputDataThread.Start();

            _dpInputDataThread = new Thread(DpInputDataThread);
            _dpInputDataThread.Start();

            // start event handling for DP and change to OPERATE
            bool bResult;
            do
            {
               bResult = DpReadServiceInterface();
               if (bResult)
               {
                  DpHandleEvent();
               }
            } while (bResult);     // terminates in case of errors only

            Global.DisplayLastError();
            Global.OutputString("DpManagementThread terminated");
        }



        //  **************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function checks the given slave parameter set whether it is valid
        /// or not. In the positive case the slave parameter set is taken over
        /// into the working image 'SlaveDescr' and can be used for downloads to
        /// the DP master.
        /// ------------------------------------------------------------------------
        /// </summary>
        public bool DpSetSlaveParaSet(byte nSlaveAdd, C_DP_SLAVE_PARA_SET slaveParaSet)
        {
            if (Global.DpSetSlaveParaSet(nSlaveAdd, slaveParaSet))
            {
                byte nNumberInputs = 0;
                byte nNumberOutputs = 0;

                if (slaveParaSet.CfgData.DpConvertCfgToLength(ref nNumberInputs, ref nNumberOutputs))
                {
                    //memcpy(_slaveParaSet[nSlaveAdd], pSlaveParaSet, nSlaveParaLen);
                    _slaveParaSets[nSlaveAdd] = slaveParaSet;

                    if (_slaveDescr[nSlaveAdd].SlaveParaSet == null)
                    {
                        _nNumberOfSlaves++;
                    }

                    _slaveDescr[nSlaveAdd].SlaveParaSet   = _slaveParaSets[nSlaveAdd]; ;
                    _slaveDescr[nSlaveAdd].nNumberInputs  = nNumberInputs;
                    _slaveDescr[nSlaveAdd].nNumberOutputs = nNumberOutputs;

                } // if slave identified successfully
                else
                {
                    Global.OutputString(string.Format("inconsistent CFG data: slave {0}", nSlaveAdd));
                    return false;

                } // else slave identified successfully

                Global.OutputString(string.Format("- preconfigured DP slave [0x{0:X4}] address {1} with bytes: {2} in / {3} out",
                                                  Global.Swap16(slaveParaSet.PrmData.IdentNumber),
                                                  nSlaveAdd,
                                                  _slaveDescr[nSlaveAdd].nNumberInputs,
                                                  _slaveDescr[nSlaveAdd].nNumberOutputs));
            }

            return true;

        } // DpSetSlaveParaSet


        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// Update the slave parameter into the working image 'SlaveDescr' and
        /// can be used for downloads to the DP master.
        /// ------------------------------------------------------------------------
        /// </summary>
        public void DpUpdateSetSlaveParaSet(Dictionary<byte, C_DP_SLAVE_PARA_SET> dpSlaveParaSets)
        {
           byte nNumberInputs;
           byte nNumberOutputs;

           foreach (byte nSlaveAdd in dpSlaveParaSets.Keys)
           {
              C_DP_SLAVE_PARA_SET slaveParaSet = dpSlaveParaSets[nSlaveAdd];
              if (slaveParaSet != null)
              {
                 nNumberInputs = 0;
                 nNumberOutputs = 0;

                 if (slaveParaSet.CfgData.DpConvertCfgToLength(ref nNumberInputs, ref nNumberOutputs))
                 {
                    _slaveParaSets[nSlaveAdd] = slaveParaSet;

                    if (_slaveDescr[nSlaveAdd].SlaveParaSet == null)
                    {
                       _nNumberOfSlaves++;
                    }

                    _slaveDescr[nSlaveAdd].SlaveParaSet   = _slaveParaSets[nSlaveAdd];
                    _slaveDescr[nSlaveAdd].nNumberInputs  = nNumberInputs;
                    _slaveDescr[nSlaveAdd].nNumberOutputs = nNumberOutputs;

                    Global.OutputString(string.Format("- import DP slave [0x{0:X4}] address {1} with bytes: {2} in / {3} out",
                                        Global.Swap16(slaveParaSet.PrmData.IdentNumber),
                                        nSlaveAdd,
                                        _slaveDescr[nSlaveAdd].nNumberInputs,
                                        _slaveDescr[nSlaveAdd].nNumberOutputs));

                 }
              }
           }
        }

        //**************************************************************************


        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function converts the given standard DP bus parameters from the
        /// binary configuration file into the separated DP parameters.
        /// These parameters are not part of the FMB bus parameters which are
        /// valid for all protocol components (FDL, FMS, DP).
        /// The DP bus parameters are optional and will influence the specific
        /// parameters of the DP component only.
        /// In this application only the service FMB_Set_Busparameter is used to
        /// set the general parameters. Optional the service DP_Set_Busparameter
        /// is used to define the DP specific extensions. Note, that the service
        /// DP_Download_Loc will not be used to download a bus parameter set.
        /// ------------------------------------------------------------------------
        /// </summary>
        public bool DpSetBusParaSet(C_DP_BUS_PARA_SET dpBusParaSet)
        {
           ushort masterUserDataLen = Global.Swap16(dpBusParaSet.MasterUserDataLen);

           if (masterUserDataLen != DPConstants.MasterUserDataLen)
           {
              Global.OutputString("invalid master_user_data_len in DP bus parameters");
              return false;
           }

           Global.OutputString("- import DP bus parameters");
           _dpBusparameter.BPFlag            = dpBusParaSet.BpFlag;
           _dpBusparameter.MinSlaveInterval  = Global.Swap16(dpBusParaSet.MinSlaveInterval);
           _dpBusparameter.PollTimeout       = Global.Swap16(dpBusParaSet.PollTimeout);
           _dpBusparameter.DataControlTime   = Global.Swap16(dpBusParaSet.DataControlTime);
           _dpBusparameter.MasterUserDataLen = masterUserDataLen;

           Array.Copy(dpBusParaSet.MasterClass2Name, _dpBusparameter.MasterClass2Name, dpBusParaSet.MasterClass2Name.Length);

           return true;

        } // DpSetBusParaSet


        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function looks for preconfigured DP slaves (slave parameter sets)
        /// in the file DPDemoNT.h (so called preconfigured slaves). All slaves will be
        /// entered in the working image 'SlaveDescr' for use by the DpDownloadLoc
        /// routine.
        /// The 'pSlaveParaSet' within 'SlaveDescr' acts as indicator whether the
        /// slave is defined or not.
        /// The CFG data of the slaves will be converted into a length of bytes.
        /// This information can be used for data exchange in the respective data
        /// threads later.
        /// ------------------------------------------------------------------------
        /// </summary>
        public void DpSetPreConfigSlaveParaSets()
        {
           int nSlaveIndex;

           byte nSlaveAdd;
           C_DP_SLAVE_PARA_SET  pSlaveParaSet;

           for (nSlaveIndex = 0; nSlaveIndex < _dpSlaveDescrTable.Length; nSlaveIndex++)
           {
              nSlaveAdd     = _dpSlaveDescrTable[nSlaveIndex].SlaveAdd;
              pSlaveParaSet = _dpSlaveDescrTable[nSlaveIndex].DPSlaveParaSetPtr;

              // check for valid entries
              if ((nSlaveAdd >= 0) && (nSlaveAdd <= DPConstants.DefaultSlaveAddress) && (pSlaveParaSet != null))
              {
                 DpSetSlaveParaSet(nSlaveAdd, pSlaveParaSet);
              }
            } // for preconfigured slaves

        }  // DpSetPreConfigSlaveParaSets


        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// Place cleanup operations necessary for this module here.
        /// ----------------------------------
        /// </summary>
        public void TerminateDpLayer()
        {

        } // TerminateDpLayer

        #endregion

        #region Private methods


        /// <summary>
        /// DisplayLastError
        /// - retrieves the error code of the last function which failed
        /// - displays readable strings if an error message string is available
        public static StringBuilder GetDpErrorString( ushort status)
        {

            ushort        nErrorCode       = (ushort)(status & 0x00FF);
            ushort        nErrorExtension  = (ushort)(status & 0xFF00);
            StringBuilder resultString     = new StringBuilder();

            if (nErrorCode > 0)
            {
                switch ((DPServiceResult)nErrorCode)
                {
                    case DPServiceResult.UE: resultString.Append("remote user error "); break;
                    case DPServiceResult.RR: resultString.Append("remote resource insufficient "); break;
                    case DPServiceResult.RS: resultString.Append("remote service/SAP deactivated "); break;
                    case DPServiceResult.RA: resultString.Append("access of remote SAP blocked "); break;
                    case DPServiceResult.NA: resultString.Append("no reaction from remote station "); break;
                    case DPServiceResult.DS: resultString.Append("local entity disconnected "); break;
                    case DPServiceResult.NO: resultString.Append("not possible in this state "); break;
                    case DPServiceResult.LR: resultString.Append("local resource not available "); break;
                    case DPServiceResult.IV: resultString.Append("invalid parameter in request "); break;
                    case DPServiceResult.TO: resultString.Append("timeout expired "); break;
                    case DPServiceResult.FE: resultString.Append("format error in request frame "); break;
                    case DPServiceResult.NI: resultString.Append("function not implemented "); break;
                    case DPServiceResult.AD: resultString.Append("access denied "); break;
                    case DPServiceResult.EA: resultString.Append("area too large "); break;
                    case DPServiceResult.LE: resultString.Append("data block length exceeded "); break;
                    case DPServiceResult.RE: resultString.Append("format error in response frame "); break;
                    case DPServiceResult.IP: resultString.Append("invalid parameter "); break;
                    case DPServiceResult.SC: resultString.Append("sequence conflict "); break;
                    case DPServiceResult.SE: resultString.Append("sequence error "); break;
                    case DPServiceResult.NE: resultString.Append("area non-existent "); break;
                    case DPServiceResult.DI: resultString.Append("data incomplete "); break;
                    case DPServiceResult.NC: resultString.Append("not connected "); break;
                    default:                 resultString.Append(string.Format("ErrorCode = 0x{0:X2} ", nErrorCode)); break;
                } // switch nErrorCode
            } // if nErrorCode

            if (nErrorExtension > 0)
            {
                switch ((DPServiceResult)nErrorExtension)
                {
                    case DPServiceResult.DataAlignment:      resultString.Append("[DATA_ALIGNMENT] "); break;
                    case DPServiceResult.TooManySlaves:      resultString.Append("[TOO_MANY_SLAVES] "); break;
                    case DPServiceResult.WrongSlaveAdd:      resultString.Append("[WRONG_SLAVE_ADD] "); break;
                    case DPServiceResult.AAMNotSupported:    resultString.Append("[AAM_NOT_SUPPORTED] "); break;
                    case DPServiceResult.TooFewDiagEntries:  resultString.Append("[TOO_FEW_DIAG_ENTRIES] "); break;
                    case DPServiceResult.WrongPrmDataLen:    resultString.Append("[WRONG_PRM_DATA_LEN] "); break;
                    case DPServiceResult.WrongCfgDataLen:    resultString.Append("[WRONG_CFG_DATA_LEN] "); break;
                    case DPServiceResult.WrongDiagLen:       resultString.Append("[WRONG_DIAG_LEN] "); break;
                    case DPServiceResult.WrongBusParaLen:    resultString.Append("[WRONG_BUS_PARA_LEN] "); break;
                    case DPServiceResult.WrongSlaveParaLen:  resultString.Append("[WRONG_SLAVE_PARA_LEN] "); break;
                    case DPServiceResult.WrongIODataLen:     resultString.Append("[WRONG_IO_DATA_LEN] "); break;
                    case DPServiceResult.NotEnoughMemory:    resultString.Append("[NOT_ENOUGH_MEMORY] "); break;
                    case DPServiceResult.WrongUsifState:     resultString.Append("[WRONG_USIF_STATE] "); break;
                    case DPServiceResult.SlaveAccessDenied:  resultString.Append("[SLAVE_ACCESS_DENIED] "); break;
                    case DPServiceResult.WrongAreaCode:      resultString.Append("[WRONG_AREA_CODE] "); break;
                    case DPServiceResult.NotSupported:       resultString.Append("[NOT_SUPPORTED] "); break;
                    case DPServiceResult.PrmDataFault:       resultString.Append("[PRM_DATA_FAULT] "); break;
                    case DPServiceResult.CfgDataFault:       resultString.Append("[CFG_DATA_FAULT] "); break;
                    case DPServiceResult.AATDataFault:       resultString.Append("[AAT_DATA_FAULT] "); break;
                    case DPServiceResult.UserDataFault:      resultString.Append("[USER_DATA_FAULT] "); break;
                    case DPServiceResult.SlaveParaFault:     resultString.Append("[SLAVE_PARA_FAULT] "); break;
                    case DPServiceResult.AreaNotAccessed:    resultString.Append("[AREA_NOT_ACCESSED] "); break;
                    case DPServiceResult.WrongBaudrate:      resultString.Append("[WRONG_BAUDRATE] "); break;
                    case DPServiceResult.WrongBPFlag:        resultString.Append("[WRONG_BP_FLAG] "); break;
                    case DPServiceResult.WrongFDLState:      resultString.Append("[WRONG_FDL_STATE] "); break;
                    case DPServiceResult.WrongActivation:    resultString.Append("[WRONG_ACTIVATION] "); break;
                    case DPServiceResult.WrongMasterAdd:     resultString.Append("[WRONG_MASTER_ADD] "); break;
                    case DPServiceResult.DPRAMInitError:     resultString.Append("[DPRAM_INIT_ERROR] "); break;
                    case DPServiceResult.WrongLen:           resultString.Append("[WRONG_LEN] "); break;
                    case DPServiceResult.WrongIdentifier:    resultString.Append("[WRONG_IDENTIFIER] "); break;
                    case DPServiceResult.LoadBusParameter:   resultString.Append("[LOAD_BUS_PARAMETER] "); break;
                    case DPServiceResult.ActivateSAP:        resultString.Append("[ACTIVATE_SAP] "); break;
                    case DPServiceResult.WrongRemoteService: resultString.Append("[WRONG_REMOTE_SERVICE] "); break;
                    case DPServiceResult.IllegalIndex:       resultString.Append("[E_DP_ILLEGAL_INDEX]"); break;
                    case DPServiceResult.IllegalSlot:        resultString.Append("[E_DP_ILLEGAL_SLOT]"); break;
                    case DPServiceResult.IllegalLength:      resultString.Append("[E_DP_ILLEGAL_LENGTH]"); break;
                    case DPServiceResult.IllegalExtention:   resultString.Append("[E_DP_ILLEGAL_EXTENSION]"); break;

                    default:                               resultString.Append(string.Format("ErrorExtension = 0x{0:X2} ", nErrorExtension)); break;
                } // switch nErrorExtension
            } // if nErrorExtension

            resultString.Append("\r\n");

            return resultString;

        } // DisplayLastError




        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function opens all via 'SlaveDescr' configured DP slave devices
        /// for writing.
        /// This thread should be started when the DP master has entered the
        /// operation mode OPERATE. From that time it is possible to send output
        /// data to the DP slaves.
        /// All configured slaves (which have output data) will be entered in a
        /// list of slaves (nSlavesWithOutputsTable). This list is used to write
        /// cyclically to these slaves. To find out the number of output bytes of
        /// a respective slave the 'SlaveDescr' structure is used.
        /// As output pattern a simple array with a countered byte is used.
        ///
        /// The function displays a life sign character to indicate that the
        /// thread is working properly and the program resides in an endless
        /// loop.
        /// ------------------------------------------------------------------------
        /// </summary>
        private void DpOutputDataThread(object objParam)
        {
            int[] nLastErrorCodeTable = new int[DPConstants.DefaultSlaveAddress + 1];
            byte[] nSlavesWithOutputsTable = new byte[DPConstants.DefaultSlaveAddress + 1];
            byte nSlavesWithOutputs;

            uint nNumberOfBytesWritten;

            byte nOutput;
            byte[] nOutputTelegram = new byte[DPConstants.MaxTelegramLen];

            byte[] lifeSign = new byte[] { 0x2F, 0x2D, 0x5C, 0x7C };
            int nLifeSign;

            Global.OutputSetThreadIdString("  OUT");
            Global.OutputString("DpOutputDataThread started");

            // generate table with all slaves containing outputs
            List<DpSlaveDesc> slaveDescrList   = new List<DpSlaveDesc>(_slaveDescr);
            DpSlaveDesc[]     slaveDescrArray  = slaveDescrList.FindAll(slaveActive => slaveActive.SlaveParaSet != null && slaveActive.nNumberInputs > 0).ToArray();

            nSlavesWithOutputs = Convert.ToByte(slaveDescrArray.Length);

            // write output data to slaves
            nOutput = 0;
            nLifeSign = 0;

            nLastErrorCodeTable = Enumerable.Repeat(0, DPConstants.DefaultSlaveAddress + 1).ToArray();
            
            int waitResult = -1;
            while (true)
            {
               // report current input data if InputEvent received
               waitResult = WaitHandle.WaitAny(_waitDpOutputDataHandles, 10, false);
               switch (waitResult)
               {
                 case WaitHandle.WaitTimeout: // ignore: do not report output values
                      // set new output data automaticaly and write output data to slaves
                      nOutput++;
                      break;

                 case 0: // display output data in for-loop below
                      nOutput++;
                      Global.OutputString(OutputMode.OmScreen,string.Format("DP Slaves output value: 0x{0:X2}", nOutput));
                      break;

                   case 1: // exit
                        Global.OutputString("DpOutputDataThread terminated");
                        Global.EwhExit.Set();
                        return;

                   case WaitFailed:
                        Global.OutputString(OutputMode.OmScreen,
                           string.Format("DP Slaves output value: 0x{0:X2}", nOutput));
                        Global.DisplayLastError();
                        Global.OutputString("DpOutputDataThread terminated");
                        return;

                   default:
                        Global.OutputString(string.Format("Error reading OutputEvent = 0x{0:X8}", waitResult));
                        Global.OutputString("DpOutputDataThread terminated");
                        return;
                }


               foreach (DpSlaveDesc slaveDescr in slaveDescrArray)
               {
                   //--- prepare output data for this cycle -------------------------------

                   nOutputTelegram = Enumerable.Repeat(nOutput, DPConstants.MaxTelegramLen).ToArray();

                   //--- report current output data if OutputEvent received ---------------
                   if (!DpWriteDataInterface(nOutputTelegram,
                                             slaveDescr.OffsetOutputs,
                                             slaveDescr.nNumberOutputs,
                                             out nNumberOfBytesWritten))
                   {
                       if (Marshal.GetLastWin32Error() != nLastErrorCodeTable[slaveDescr.SlaveNumber])
                       {
                           nLastErrorCodeTable[slaveDescr.SlaveNumber] = Marshal.GetLastWin32Error();
                           Global.OutputString(string.Format("Writing outputs to DP slave {0}:", nSlavesWithOutputsTable[slaveDescr.SlaveNumber]));
                           Global.DisplayLastError();
                       }
                   } // if !WriteFile
                   else
                   {
                       if (nLastErrorCodeTable[slaveDescr.SlaveNumber] > 0)
                       {
                           Global.OutputString(string.Format("Writing outputs to DP slave {0}: failed", slaveDescr.SlaveNumber));
                           nLastErrorCodeTable[slaveDescr.SlaveNumber] = 0;
                       }
                   } // else !WriteFile
               } // for slaves with outputs

               //--- display life sign ------------------------------------------------
               if (nOutput % 5 == 0)            // delay life sign indication a little
               {
                   if (nLifeSign >= lifeSign.Length)
                   {
                       nLifeSign = 0;
                   }

                   Global.OutputString(OutputMode.OmSuppressId | OutputMode.OmScreen | OutputMode.OmSuppressLf,
                          string.Format("{0}\r", Convert.ToChar(lifeSign[nLifeSign++])));
               }

            }; // while true

        } // DpOutputDataThread

        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function opens all via 'SlaveDescr' configured DP slave devices
        /// for reading.
        /// This thread should be started when the DP master has entered the
        /// operation mode CLEAR or OPERATE. From that time it is possible to
        /// receive data from the DP slaves.
        /// All configured slaves (which have intput data) will be entered in a
        /// list of slaves (nSlavesWithInputsTable). This list is used to read
        /// cyclically from these slaves. To find out the number of input bytes of
        /// a respective slave the 'SlaveDescr' structure is used.
        /// ------------------------------------------------------------------------
        /// </summary>
        private void DpInputDataThread(object objParam)
        {
            int[] nLastErrorCodeTable = new int[DPConstants.DefaultSlaveAddress + 1];
            byte[] nSlavesWithInputsTable = new byte[DPConstants.DefaultSlaveAddress + 1];

            uint nNumberOfBytesRead;

            byte[] nInputTelegram = null;

            Global.OutputSetThreadIdString("   IN");
            Global.OutputString("DpInputDataThread started.");

            // generate table with all slaves containing inputs
            List<DpSlaveDesc> slaveDescrList  = new List<DpSlaveDesc>(_slaveDescr);
            DpSlaveDesc[]     slaveDescrArray = slaveDescrList.FindAll(slaveActive => slaveActive.SlaveParaSet != null && slaveActive.nNumberInputs > 0).ToArray();

            //--- read input data from slaves ----------------------------------------
            nLastErrorCodeTable = Enumerable.Repeat(0, DPConstants.MaxTelegramLen + 1).ToArray();

            int waitResult = -1;
            while (true)
            {
               // report current input data if InputDataEvent received
               waitResult = WaitHandle.WaitAny(_waitDpInputDataHandles, 10, false);
               switch (waitResult)
               {
                  case 0: // display input data in for-loop below

                        //--- read and report input data ---------------------------------------

                        foreach (DpSlaveDesc slaveDescr in slaveDescrArray)
                        {
                            if (!DpReadDataInterface(out nInputTelegram,
                                                     slaveDescr.OffsetInputs,
                                                     slaveDescr.nNumberInputs,
                                                     out nNumberOfBytesRead))
                            {
                                if (Marshal.GetLastWin32Error() != nLastErrorCodeTable[slaveDescr.SlaveNumber])
                                {
                                    nLastErrorCodeTable[slaveDescr.SlaveNumber] = Marshal.GetLastWin32Error();
                                    Global.OutputString(string.Format("Reading inputs from DP slave {0}:", slaveDescr.SlaveNumber));//nSlavesWithInputsTable[slaveDescr.SlaveNumber]));
                                    Global.DisplayLastError();
                                }
                            } // if !ReadFile
                            else
                            {
                                Global.OutputString(OutputMode.OmScreen, string.Format("Slave {0:D3}:", slaveDescr.SlaveNumber));
                                ShowHexDump(nInputTelegram, slaveDescr.nNumberInputs);

                                if (nLastErrorCodeTable[slaveDescr.SlaveNumber] > 0)
                                {
                                    Global.OutputString(string.Format("Reading inputs from DP slave {0}: OK", nSlavesWithInputsTable[slaveDescr.SlaveNumber]));
                                    nLastErrorCodeTable[slaveDescr.SlaveNumber] = 0;
                                }
                            } // else !ReadFile
                        } // for slaves with inputs
                        break;
                    case 1: // exit
                         Global.OutputString("DpInputDataThread terminated");
                         Global.EwhExit.Set();
                         return;

                    case WaitHandle.WaitTimeout: // ignore: do not report output values
                         break;

                    case WaitFailed:
                        Global.DisplayLastError();
                        Global.OutputString("DpInputDataThread terminated");
                        return;
                    default:
                        Global.OutputString(string.Format("Error reading InputEvent = 0x{0:X8}", waitResult));
                        Global.OutputString("DpInputDataThread terminated");
                        return;
                }

            }; // while true

        } // DpInputDataThread

        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// Place all module specific initialization here. This function is called
        /// before any communication is started or initialized.
        /// ------------------------------------------------------------------------
        /// </summary>
        private void InitDpLayer()
        {
            for (byte nSlaveAddr = 0; nSlaveAddr < _slaveDescr.Length; nSlaveAddr++)
            {
               _slaveDescr[nSlaveAddr] = new DpSlaveDesc(nSlaveAddr);
            }


            // Siemens ET200B Slave
            _dpSlaveSiemensET200B = new C_DP_SLAVE_PARA_SET()
            {
                //--- SLAVE_PARA_SET ----------------------------------------------------------------------------------------------------
                SlaveParaLen = Global.Swap16(38),                                   // slave_para_len (Big Endian Format)
                SlFlag       = (byte)DPSlaveFlags.Active,                           // sl_flag: slave activated
                SlaveType    = DPConstants.SlaveTypeDP,                             // slave_type: do not change

                //--- PRM_DATA ----------------------------------------------------------------------------------------------------------
                PrmData = new C_DP_PRM_DATA()
                {
                    PrmDataLen    = Global.Swap16(14),                              // prm_data_len (big endian format)
                    StationStatus = (byte)DPSlaveParamStatus.WDOn,                  // watchdog on
                    WdFact1       = 10,                                             // 10ms * wd_fact_1/2
                    WdFact2       = 10,                                             // --> 1s
                    MinTsdr       = 11,                                             // min Tsdr
                    IdentNumber   = Global.Swap16(0x000B),                          // ident number (big endian format)
                    GroupIdent    = 0,                                              // no group
                    UserPrmData   = new byte[] {0,0,0,0,0}                          // 5 byte user UserPrmData
                },

                //--- CFG_DATA ----------------------------------------------------------------------------------------------------------
                CfgData = new C_DP_CFG_DATA()
                {
                    CfgDataLen = Global.Swap16(4),                                  // cfg_data_len (big endian format)
                    CfgData    = new byte[] { 0x20, 0x10 }
                },

                //--- AAT_DATA -----------------------------------------------------------
                AATData = new C_DP_AAT_DATA()
                {
                    AATDataLen = Global.Swap16(2),                                  // AATDataLen (big endian format)
                    // using AAM IOP-Block-Mode
                },

                //--- SLAVE_USER_DATA ----------------------------------------------------
                UserData = new C_DP_SLAVE_USER_DATA()
                {
                    SlaveUserDataLen = Global.Swap16(2)                             // SlaveUserDataLen (big endian format)
                }
            }; // _dpSlaveSiemensET200B


            // Softing B205 PROFI104, PROFIboard-PCI as SP-Slave device
            _dpSlaveSoftingB205 = new C_DP_SLAVE_PARA_SET()
            {
                //--- SLAVE_PARA_SET ----------------------------------------------------------------------------------------------------
                SlaveParaLen = Global.Swap16(34),                                   // slave_para_len (Big Endian Format)
                SlFlag       = (byte)DPSlaveFlags.Active,                           // sl_flag: slave activated
                SlaveType    = DPConstants.SlaveTypeDP,                             // slave_type: do not change

                //--- PRM_DATA ----------------------------------------------------------------------------------------------------------
                PrmData = new C_DP_PRM_DATA()
                {
                    PrmDataLen    = Global.Swap16(9),                               // prm_data_len (big endian format)
                    StationStatus = (byte)DPSlaveParamStatus.LockReq |              // locked by master
                                    (byte)DPSlaveParamStatus.WDOn,                  // station_status: watchdog enabled
                    WdFact1       = 100,                                            // 10ms * wd_fact_1/2
                    WdFact2       = 10,                                             // --> 10s
                    MinTsdr       = 11,                                             // min Tsdr
                    IdentNumber   = Global.Swap16(0xB205),                          // ident number
                    GroupIdent    = 0,                                              // no group
                },

                //--- CFG_DATA ----------------------------------------------------------------------------------------------------------
                CfgData = new C_DP_CFG_DATA()
                {
                    CfgDataLen = Global.Swap16(5),                                  // cfg_data_len (big endian format)
                    CfgData    = new byte[] { 0x10, 0xD1, 0x21 }
                },

                //--- AAT_DATA -----------------------------------------------------------
                AATData = new C_DP_AAT_DATA()
                {
                    AATDataLen = Global.Swap16(2)                                   // AATDataLen (big endian format)
                    // using AAM IOP-Block-Mode
                },

                //--- SLAVE_USER_DATA ----------------------------------------------------
                UserData = new C_DP_SLAVE_USER_DATA()
                {
                    SlaveUserDataLen = Global.Swap16(2)                             // SlaveUserDataLen (big endian format)
                }
            }; // _dpSlaveSoftingB205


            // Softing 0B35 PPpro as SP-Slave device
            _dpSlaveSofting0B35 = new C_DP_SLAVE_PARA_SET()
            {
                //--- SLAVE_PARA_SET ----------------------------------------------------------------------------------------------------
                SlaveParaLen = Global.Swap16(34),                                   // slave_para_len (Big Endian Format)
                SlFlag       = (byte)DPSlaveFlags.Active,                           // sl_flag: slave activated
                SlaveType    = DPConstants.SlaveTypeDP,                             // slave_type: do not change

                //--- PRM_DATA ----------------------------------------------------------------------------------------------------------
                PrmData = new C_DP_PRM_DATA()
                {
                    PrmDataLen    = Global.Swap16(9),                               // prm_data_len (big endian format)
                    StationStatus = (byte)DPSlaveParamStatus.LockReq |              // locked by master
                                    (byte)DPSlaveParamStatus.WDOn,                  // station_status: watchdog enabled
                    WdFact1       = 100,                                            // 10ms * wd_fact_1/2
                    WdFact2       = 10,                                             // --> 10s
                    MinTsdr       = 11,                                             // min Tsdr
                    IdentNumber   = Global.Swap16(0x0B35),                          // ident number
                    GroupIdent    = 0,                                              // no group
                },

                //--- CFG_DATA ----------------------------------------------------------------------------------------------------------
                CfgData = new C_DP_CFG_DATA()
                {
                    CfgDataLen = Global.Swap16(5),                                  // cfg_data_len (big endian format)
                    CfgData    = new byte[] { 0x10, 0xD1, 0x21 }
                },

                //--- AAT_DATA -----------------------------------------------------------
                AATData = new C_DP_AAT_DATA()
                {
                    AATDataLen = Global.Swap16(2)                                   // AATDataLen (big endian format)
                    // using AAM IOP-Block-Mode
                },

                //--- SLAVE_USER_DATA ----------------------------------------------------
                UserData = new C_DP_SLAVE_USER_DATA()
                {
                    SlaveUserDataLen = Global.Swap16(2)                             // SlaveUserDataLen (big endian format)
                }
            }; // _dpSlaveSofting0B35


            _dpSlaveDescrTable = new C_PRECONFIG_DP_SLAVE_DESCR[]
            {
              // slave address (SlaveAdd), pointer to the slave parameter set
              // just insert your desired DP slave here with slave address and para set!

              new C_PRECONFIG_DP_SLAVE_DESCR()
              {
                 SlaveAdd = 0x02,
                 DPSlaveParaSetPtr = _dpSlaveSoftingB205
              },
              new C_PRECONFIG_DP_SLAVE_DESCR()
              {
                 SlaveAdd = 0x03,
                 DPSlaveParaSetPtr = _dpSlaveSofting0B35
              },
              new C_PRECONFIG_DP_SLAVE_DESCR()
              {
                 SlaveAdd = 0x04,
                 DPSlaveParaSetPtr = _dpSlaveSiemensET200B
              },
              new C_PRECONFIG_DP_SLAVE_DESCR()
              {
                 SlaveAdd = 0xFF,
                 DPSlaveParaSetPtr = null
              } // do not remove

            }; // DpSlaveDescrTable



            _dpSlaveParaSetTemplate = new C_DP_SLAVE_PARA_SET()
            {
                //--- SLAVE_PARA_SET -----------------------------------------------------
                SlaveParaLen = Global.Swap16(36),                                  // slave_para_len (big endian format)
                SlFlag       = (byte)DPSlaveFlags.Active,                          // sl_flag: slave activated
                SlaveType    = DPConstants.SlaveTypeDP,                            // slave_type: do not change

                //--- PRM_DATA -----------------------------------------------------------
                PrmData = new C_DP_PRM_DATA()
                {
                   PrmDataLen    = Global.Swap16(14),                             // prm_data_len (big endian format)
                   StationStatus = (byte)DPSlaveParamStatus.WDOn,                 // station_status: watchdog enabled
                   WdFact1       = 50,                                            // 10ms * wd_fact_1/2
                   WdFact2       = 10,                                            // --> 5s
                   UserPrmData   = new byte[5]                                    // user_prm_data: 5 Byte for LSPM2-Slaves (Siemens)
                },

                //--- CFG_DATA -----------------------------------------------------------
                CfgData = new C_DP_CFG_DATA()
                {
                   CfgDataLen = Global.Swap16(2),                                 // cfg_data_len (big endian format)
                },

                //--- AAT_DATA -----------------------------------------------------------
                AATData = new C_DP_AAT_DATA()
                {
                   AATDataLen = Global.Swap16(2),                                 // aat_data_len (big endian format)
                },

                //--- SLAVE_USER_DATA ----------------------------------------------------
                UserData = new C_DP_SLAVE_USER_DATA()
                {
                  SlaveUserDataLen = Global.Swap16(2),                            // slave_user_data_len (big endian format)
                }
            }; // _dpSlaveParaSetTemplate


            _dpBusparameter = new C_DP_SET_BUSPARAMETER_REQ()
            {
                BPFlag            = 0,      // bp_flag (bus parameter flags): DP_BP_ERROR_ACTION (AUTO_CLEAR)
                MinSlaveInterval  = 10,                 // USIGN16 min_slave_interval: 1..2^16-1 [100 micros]
                PollTimeout       = 10000,                          // USIGN16 poll_timeout: 1..2^16-1 [1 ms]
                DataControlTime   = 100,                      // USIGN16 data_control_time: 1..2^16-1 [10 ms]
                MasterUserDataLen = (ushort)DPConstants.MasterUserDataLen,                      // USIGN16 master_user_data_len
                MasterClass2Name  = Helper.ConvertStringToSByte("Softing DP-Master V5.45 ", 32) // master_class2_name [32]

            }; // _dpSetBusparameterReqData



        } // InitDpLayer

        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function searches for DP slaves on the PROFIBUS. Addresses with
        /// already defined slaves (preconfigured slaves) will be omitted.
        /// In case of a found and valid DP slave a new slave parameter set with
        /// standard settings will be created for this slave.
        /// The template slave parameter set from file DPDemoNT.h will be used.
        ///
        /// Prerequisites:
        /// The function must be called in the DP master OFFLINE state, i.e.
        /// right after FMB initialization. This routine will bring the DP master
        /// in the previous state when terminating the detection process with
        /// no errors (thus, the startup routine can continue without regards).
        ///
        /// Slave detection process:
        /// To identify a DP slave the following services are used:
        /// - DP_Slave_Diag: request the current slave state and the ident_number,
        /// - DP_Get_Cfg   : retrieves the current slave configuration (the size
        ///                  of slave inputs/outputs), the so called 'real CFG'.
        /// Unfortunately it is not possible to request the parameterization data
        /// (PRM data + USER_PRM data) from the DP slave. Many DP slaves on the
        /// market are not able to run with default parameters!
        /// This function assumes that the slave is a SIEMENS-LSPM2-Asic based
        /// device (e.g. ET200B, ...). The slave para set template contains the
        /// USER_PRM data (0,0,0,0,0) for this reason. If a slave reports a
        /// PRM_FAULT during startup the slave most likely needs special USER_PRM
        /// data. In this case you should modify the slave template slave para set
        /// or create different templates and identify them by means of the slave
        /// ident_number. Of course you can also define a preconfigured slave.
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpSlaveDetection()
        {
            C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();

            C_DP_SLAVE_PARA_SET   slaveParaSet  = new C_DP_SLAVE_PARA_SET();
            C_DP_PRM_DATA         prmData       = new C_DP_PRM_DATA();
            C_DP_CFG_DATA         cfgData       = new C_DP_CFG_DATA();
            C_DP_AAT_DATA         aatData       = new C_DP_AAT_DATA();
            C_DP_SLAVE_USER_DATA  slaveUserData = new C_DP_SLAVE_USER_DATA();

            C_DP_SLAVE_DIAG_CON   slaveDiagCon  = new C_DP_SLAVE_DIAG_CON(); ;
            C_DP_DIAG_DATA        diagData      = new C_DP_DIAG_DATA();
            C_DP_GET_CFG_CON      getCfgCon     = new C_DP_GET_CFG_CON();

            ushort nPrmDataLength = 0;
            ushort nSlaveParaLength = 0;

            ushort nRealCfgDataLength;

            int    cfgIndex;
            byte[] pCfgIdentifier;
            string realCfgIdentifierString;

            StringBuilder sbDpManagement = new StringBuilder();

            //--- initialize DP Master -----------------------------------------------

            if (!DpInitMasterReq())
            {
                Global.DisplayLastError();
                return false;
            }

            if (!DpReadServiceInterface())
            {
                Global.DisplayLastError();
                return false;
            }

            DpHandleEvent();

            //--- enter operation mode STOP (enter PROFIBUS) -------------------------

            _nDesiredOpMode = DPOPMode.Stop;

            if (!DpActParamLocReq(DPOPMode.Stop))
            {
                Global.DisplayLastError();
                return false;
            }

            if (!DpReadServiceInterface())
            {
                Global.DisplayLastError();
                return false;
            }

            DpHandleEvent();

            //--- scan any bus address for DP slave detection (except preconfigured slaves) --
            Global.OutputString (OutputMode.OmScreen | OutputMode.OmStartLf, "Scanning PROFIBUS for not configured DP slave devices");
            Thread.Sleep(2000);

            do
            {
                for (byte nSlaveAdd = 0; nSlaveAdd <= DPConstants.DefaultSlaveAddress; nSlaveAdd++)
                {
                    if (_slaveDescr[nSlaveAdd].SlaveParaSet == null)
                    {
                        //--- check for diagnostic data at any address (DP slave?) -----------

                        if (!DpSlaveDiagReq(nSlaveAdd))
                        {
                            Global.DisplayLastError();
                            return false;
                        }

                        if (!DpReadServiceInterface())
                        {
                            Global.DisplayLastError();
                            return false;
                        }

                        //--- only DP_Slave_Diag.con shall be received here ------------------

                        try
                        {
                            profiSdb = _profiSdb;

                            // pSlaveDiagCon = (T_DP_SLAVE_DIAG_CON*)(pProfiSdb + 1);
                            slaveDiagCon = profiSdb.Data as C_DP_SLAVE_DIAG_CON;
                            if (slaveDiagCon == null || slaveDiagCon.DiagData == null)
                            {
                                Global.OutputString(OutputMode.OmScreen | OutputMode.OmStartLf, string.Format("invalid DP_Slave_Diag response at slave {0}", nSlaveAdd));
                                return false;
                            }

                            // pDiagData = (T_DP_DIAG_DATA*)(pSlaveDiagCon + 1);
                            diagData = slaveDiagCon.DiagData;

                            if ((profiSdb.Service != (byte)DPService.SlaveDiag) || (profiSdb.Primitive != (byte)ServicePrimitive.Con))
                            {
                                Global.OutputString(OutputMode.OmScreen | OutputMode.OmStartLf, string.Format("invalid service response at slave {0} on DP_Slave_Diag.req", nSlaveAdd));
                                return false;
                            }

                            if (slaveDiagCon.Status == (ushort)(DPServiceResult.WrongSlaveAdd | DPServiceResult.IV))
                            {
                               Global.OutputString(OutputMode.OmScreen | OutputMode.OmStartLf, string.Format("local DP-Master at address {0}", nSlaveAdd));
                            }
                            else if (slaveDiagCon.Status == (ushort)DPServiceResult.OK)    // DP slave available here
                            {
                                // copy slave para set template as base for a new one
                                _slaveParaSets[nSlaveAdd] = _dpSlaveParaSetTemplate;
                                slaveParaSet              = _slaveParaSets[nSlaveAdd];

                                //pPrmData = (T_DP_PRM_DATA*)(pSlaveParaSet + 1);
                                prmData = slaveParaSet.PrmData;
                                nPrmDataLength = Global.Swap16(prmData.PrmDataLen);

                                //pCfgData = (T_DP_CFG_DATA*)(((OCTET*)pPrmData) + nPrmDataLength);
                                cfgData = slaveParaSet.CfgData;

                                prmData.IdentNumber = diagData.IdentNumber;

                                //--- get configuration data from slave now ----------------------

                                    if (!DpGetCfgReq(nSlaveAdd))
                                    {
                                        Global.DisplayLastError();
                                        return false;
                                    }

                                    if (!DpReadServiceInterface())
                                    {
                                        Global.DisplayLastError();
                                        return false;
                                    }

                                    profiSdb = _profiSdb;

                                    //// --- only DP_Get_Cfg.con shall be received here -----------------

                                    //pGetCfgCon = (T_DP_GET_CFG_CON*)(pProfiSdb + 1);
                                    getCfgCon = profiSdb.Data as C_DP_GET_CFG_CON;
                                    if (getCfgCon == null || getCfgCon.RealCfgData == null)
                                    {
                                        Global.OutputString(OutputMode.OmScreen | OutputMode.OmStartLf,
                                            string.Format("invalid DP_Get_Cfg response at slave {0}", nSlaveAdd));

                                        return false;
                                    }

                                    if ((profiSdb.Service != (byte)DPService.GetCfg) ||
                                        (profiSdb.Primitive != (byte)ServicePrimitive.Con))
                                    {
                                        Global.OutputString(OutputMode.OmScreen | OutputMode.OmStartLf,
                                            string.Format("invalid service response at slave {0} on DP_Get_Cfg.req", nSlaveAdd));

                                        return false;
                                    }

                                    if (getCfgCon.RealCfgData.DpConvertCfgToLength(
                                              ref _slaveDescr[nSlaveAdd].nNumberInputs,
                                              ref _slaveDescr[nSlaveAdd].nNumberOutputs))
                                    {
                                        //--- adapt REAL_CFG from slave to new slave para set ----------

                                        nRealCfgDataLength = Global.Swap16(getCfgCon.RealCfgData.CfgDataLen);
                                        if (nRealCfgDataLength > 2)
                                        {
                                            cfgData.CfgData = new byte[nRealCfgDataLength - 2];
                                            cfgData.CfgDataLen = getCfgCon.RealCfgData.CfgDataLen;
                                            Buffer.BlockCopy(getCfgCon.RealCfgData.CfgData, 0, cfgData.CfgData, 0, nRealCfgDataLength - 2);

                                        }

                                        //--- complete new slave para set and recalculate length -------

                                        //pAatData = (T_DP_AAT_DATA*)(((OCTET*)pCfgData) + nRealCfgDataLength);
                                        aatData = slaveParaSet.AATData;
                                        aatData.AATDataLen = Global.Swap16(2);    // sizeof (aat_data_len)

                                        //pSlaveUserData = (T_DP_SLAVE_USER_DATA*)(((OCTET*)pAatData) + 2);
                                        slaveUserData.SlaveUserDataLen = Global.Swap16(2);

                                        nSlaveParaLength = (ushort)(SizeOfSlaveParaSet +
                                                           nPrmDataLength +
                                                           nRealCfgDataLength +
                                                           4);     // aat_data_len + slave_user_data_len
                                        slaveParaSet.SlaveParaLen = Global.Swap16(nSlaveParaLength);

                                        slaveParaSet.AATData = aatData;
                                        slaveParaSet.CfgData = cfgData;
                                        slaveParaSet.PrmData = prmData;
                                        slaveParaSet.UserData = slaveUserData;
                                        _slaveDescr[nSlaveAdd].SlaveParaSet = slaveParaSet;
                                        _nNumberOfSlaves++;          // slave identified successfully

                                        Global.OutputString(OutputMode.OmScreenFile | OutputMode.OmStartLf,
                                            string.Format("DP slave [0x{0:X4}] found at address {1} with bytes: {2} in / {3} out",
                                                 Global.Swap16(prmData.IdentNumber),
                                                 nSlaveAdd,
                                                 _slaveDescr[nSlaveAdd].nNumberInputs,
                                                 _slaveDescr[nSlaveAdd].nNumberOutputs));

                                        //--- write REAL CFG data to log file --------------------------

                                        sbDpManagement.Remove(0, sbDpManagement.Length);
                                        sbDpManagement.Append(string.Format("REAL_CFG [0x{0:X4}]: ",
                                            Global.Swap16(prmData.IdentNumber)));

                                        pCfgIdentifier = getCfgCon.RealCfgData.CfgData;
                                        int cfgIdentIndex = 0;
                                        for (cfgIndex = 2; cfgIndex < nRealCfgDataLength; cfgIndex++)
                                        {
                                            if (cfgIndex == nRealCfgDataLength - 1)
                                            {
                                                realCfgIdentifierString =
                                                    string.Format("0x{0:X2}", pCfgIdentifier[cfgIdentIndex++]);
                                            }
                                            else
                                            {
                                                realCfgIdentifierString =
                                                    string.Format("0x{0:X2}", pCfgIdentifier[cfgIdentIndex++]); // !? the same ?
                                            }

                                            sbDpManagement.Append(" ");
                                            sbDpManagement.Append(realCfgIdentifierString);
                                        }

                                        Global.OutputString(OutputMode.OmFile, sbDpManagement.ToString());

                                    } // if REAL_CFG valid
                            } // else local master address
                        }// if not a preconfigured slave at this address
                        catch (Exception exc)
                        {
                            Global.DisplayError(exc);
                        }
                    }
                    else
                    {
                        slaveParaSet = _slaveDescr[nSlaveAdd].SlaveParaSet;
                        //pPrmData    = (T_DP_PRM_DATA*) (pSlaveParaSet + 1);
                        prmData = slaveParaSet.PrmData;

                        Global.OutputString(OutputMode.OmScreenFile | OutputMode.OmStartLf,
                            string.Format("DP slave [0x{0:X4}] preconfigured at address {1} with bytes: {2} in / {3} out",
                                 Global.Swap16(prmData.IdentNumber),
                                 nSlaveAdd,
                                 _slaveDescr[nSlaveAdd].nNumberInputs,
                                 _slaveDescr[nSlaveAdd].nNumberOutputs));

                    } // else new slave
                } // for nSlaveAdd


                if (_nNumberOfSlaves == 0)
                {
                    Thread.Sleep(10000);
                    Global.OutputString(OutputMode.OmScreenFile | OutputMode.OmStartLf,
                        "No DP slave found.");
                    Global.OutputString(OutputMode.OmScreenFile | OutputMode.OmStartLf,
                        "Scanning PROFIBUS again for not configured DP slave devices");
                }
            } while (_nNumberOfSlaves == 0);

            Global.OutputString(OutputMode.OmScreenFile | OutputMode.OmStartLf, "DP slave scan done successfully");

            //--- enter operation mode OFFLINE ---------------------------------------

            if (!DpExitMasterReq())
            {
                Global.DisplayLastError();
                return false;// (1);
            }

            if (!DpReadServiceInterface())
            {
                Global.DisplayLastError();
                return false;
            }

            DpHandleEvent();

            return true;

        } // DpSlaveDetection

        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This service brings the DP master in the operation mode OFFLINE.
        /// Also the operation characteristics of the master can be defined (e.g.
        /// master class 2 role). In WindowsNT environments there is no need to
        /// change one of the predefined parameters in 'init_master_req_data'
        /// except 'master_class2' if you wish to be a DP master (class 2)
        /// requester/client station.
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpInitMasterReq()
        {
            byte[] profiData = new byte[Global.MaxPduDataBufferLength];

            Global.OutputString("DP_Init_Master.req");

            C_PROFI_SERVICE_DESCR profiSdb      = new C_PROFI_SERVICE_DESCR();
            C_DP_INIT_MASTER_REQ  initMasterReq = new C_DP_INIT_MASTER_REQ();

            profiSdb.Layer     = (byte)LayerIdentifier.Dp;
            profiSdb.Service   = (byte)DPService.InitMaster;
            profiSdb.Primitive = (byte)ServicePrimitive.Req;

            initMasterReq.MasterDefaultAddress = Global.MasterAddr;
            initMasterReq.MasterDefaultAddress = 0;                        // master_default_address
            initMasterReq.MasterClass2         = Constants.PbFalse;                 // master_class2
            initMasterReq.MasterClass2Name     = Helper.ConvertStringToSByte("Softing DP-Master V5.45 ", 32);    // master_class2_name [32]
            initMasterReq.LowestSlaveAddress   = 0;                           // lowest_slave_address
            initMasterReq.SlaveIOAddressMode   = (byte)DPDpramAAM.IOBlocks;     // slave_io_address_mode
            initMasterReq.ClearOutputs         = Constants.PbFalse;                             // clear_outputs
            initMasterReq.AutoRemoteServices   = (byte)MasterDPAuto.RemoteServices;                       // auto_remote_services
            initMasterReq.CyclicDataTransfer   = Constants.PbTrue;                                       // cyclic_data_transfer

            if (Global.WriteServiceInterface(profiSdb, initMasterReq))
            {
                _nRequestedOpMode = DPOPMode.Offline;
                _nDesiredOpMode = DPOPMode.Offline;
                return true;
            }

            return false;

        } // DpInitMasterReq

        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function is used to bring the DP master into a new operation mode
        /// (STOP, CLEAR, OPERATE). Please consider that it is allowed to switch
        /// between neighbouring operation modes only (e.g. STOP->CLEAR, OPERATE->
        /// CLEAR, and not STOP->OPERATE).
        /// The operation mode is controlled via the variables nCurrentOpMode,
        /// nRequestedOpMode, nDesiredOpMode and will be handled within function
        /// 'DpHandleEvent' (see there for details).
        ///
        /// DP operation modes:
        /// - OFFLINE: master inactive, local services allowed only (init phase)
        /// - STOP   : master active on PROFIBUS, no data exchange with DP slaves
        ///            yet, all services allowed except reconfiguration of DP
        ///            slaves (i.e. DP_Download_Loc)
        /// - CLEAR  : master active on PROFIBUS, data exchange with all slaves,
        ///            output data is set to zero or length null (FAIL SAFE mode)
        /// - OPERATE: master active, normal data exchage with all slaves
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpActParamLocReq(DPOPMode nOpMode)
        {
            switch (nOpMode)
            {
                case DPOPMode.Offline: Global.OutputString("DP_Act_Param_Loc.req (OFFLINE)"); break;
                case DPOPMode.Stop: Global.OutputString("DP_Act_Param_Loc.req (STOP)"); break;
                case DPOPMode.Clear: Global.OutputString("DP_Act_Param_Loc.req (CLEAR)"); break;
                case DPOPMode.Operate: Global.OutputString("DP_Act_Param_Loc.req (OPERATE)"); break;

                default:
                    {
                        Global.OutputString(string.Format("DP_Act_Param_Loc.req (invalid operation mode = 0x%{0:X2})", nOpMode));
                        return false;
                    }
            }

            C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();
            C_DP_ACT_PARAM_REQ    actParam = new C_DP_ACT_PARAM_REQ();

            profiSdb.Layer = (byte)LayerIdentifier.Dp;
            profiSdb.Service = (byte)DPService.ActParamLoc;
            profiSdb.Primitive = (byte)ServicePrimitive.Req;

            actParam.AreaCode = (byte)DPAreaCodes.SetMode;
            actParam.Activate = (byte)nOpMode;

            if (Global.WriteServiceInterface(profiSdb, actParam))
            {
                _nRequestedOpMode = nOpMode;
                return true;
            }

            return false;

        } // DpActParamLocReq

        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function requests the current slave diagnostic data from a single
        /// DP slave. This service is handled in parallel to the standard diag
        /// history which is managed in a circular buffer of the DP master. Do not
        /// confuse this service with the 'DP_Get_Slave_Diag' service which is
        /// responsible for the circular buffer!
        ///
        /// This service provides among other things:
        /// - whether the slave is reachable via PROFIBUS communication,
        /// - the current state of the DP slave,
        /// - the IDENT_NUMBER of the DP slave,
        /// - the address of the DP master which has configured this slave.
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpSlaveDiagReq(byte nSlaveAdd)
        {
            C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();
            C_DP_SLAVE_DIAG_REQ   slaveDiag = new C_DP_SLAVE_DIAG_REQ();

            profiSdb.Layer     = (byte)LayerIdentifier.Dp;
            profiSdb.Service   = (byte)DPService.SlaveDiag;
            profiSdb.Primitive = (byte)ServicePrimitive.Req;

            slaveDiag.RemAdd   = nSlaveAdd;

            return Global.WriteServiceInterface(profiSdb, slaveDiag);

        } // DpSlaveDiagReq


        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function can be used to request the current configuration of a
        /// DP slave. The data is coded recording to EN 50170/DIN 19245-3 and
        /// describes the number of inputs/outputs of the device. The current
        /// configuration data (CFG data) is also called REAL CFG data.
        /// To decode the retrieved data use the auxiliary function
        /// 'DpConvertCfgToLength'.
        ///
        /// Please note that it is not possible to identify a specific module
        /// of a modular DP slave within the CFG data because the coding of the
        /// CFG is not clear at this point. This is a weakness of the DP protocol.
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpGetCfgReq(byte nSlaveAdd)
        {
            C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();
            C_DP_CHK_CFG_REQ      getCfg   = new C_DP_CHK_CFG_REQ();

            profiSdb.Layer     = (byte)LayerIdentifier.Dp;
            profiSdb.Service   = (byte)DPService.GetCfg;
            profiSdb.Primitive = (byte)ServicePrimitive.Req;

            getCfg.RemAdd      = nSlaveAdd;

            return Global.WriteServiceInterface(profiSdb, getCfg);

        } // DpGetCfgReq


        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// Use this service to bring the DP master in the operation mode OFFLINE
        /// just before master configuration. That means you have to call service
        /// DP_Init_Master before being able to execute local services again.
        /// This service can be called regardless of the current operation mode
        /// of the DP Master (see DpActParamLocReq and DpInitMasterReq for more
        /// information).
        ///
        /// This service is the reset service for the DP protocol component. It
        /// does not influence any other components like FMB or FMS.
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpExitMasterReq()
        {
            Global.OutputString("DP_Exit_Master.req");

            C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();

            profiSdb.Layer     = (byte)LayerIdentifier.Dp;
            profiSdb.Service   = (byte)DPService.ExitMaster;
            profiSdb.Primitive = (byte)ServicePrimitive.Req;

            if (Global.WriteServiceInterface(profiSdb, null))
            {
                _nRequestedOpMode = DPOPMode.Offline;
                _nDesiredOpMode   = DPOPMode.Offline;
                return true;
            }
            return false;

        } // DpExitMasterReq


        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function loads the DP specific part of the bus parameter set.
        /// By means of service FMB_Set_Busparameter the bus parameters for all
        /// protocol components are set (i.e. FDL, FMS, DP).
        /// The DP master has some additional parameters (e.g. Min_Slave_Interval)
        /// which can be set with this service.
        ///
        /// You should call this function if you want to change the default
        /// settings of the DP bus parameters. The use of this service is
        /// optional. Call this service right after DP_Init_Master.con [+].
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpSetBusparameterReq()
        {
            Global.OutputString("DP_Set_Busparameter.req");

            C_PROFI_SERVICE_DESCR       profiSdb             = new C_PROFI_SERVICE_DESCR();
            C_DP_SET_BUSPARAMETER_REQ   dpSetBusparameterReq = new C_DP_SET_BUSPARAMETER_REQ();

            profiSdb.Layer     = (byte)LayerIdentifier.Dp;
            profiSdb.Service   = (byte)DPService.SetBusParameter;
            profiSdb.Primitive = (byte)ServicePrimitive.Req;

            dpSetBusparameterReq = _dpBusparameter;

            return Global.WriteServiceInterface(profiSdb, dpSetBusparameterReq);

        } // DpSetBusparameterReq

        //**************************************************************************

        /// <summary>
        /// Starts sequence loc.
        /// </summary>
        private bool DpStartSeqLoc(byte nMasterAdd, byte nSlaveAdd)
        {
            byte[] profiData = new byte[Global.MaxPduDataBufferLength];

            Global.OutputString(string.Format("DP_StartSequence_Loc.req Master: ({0}) Slave: ({1})", nMasterAdd, nSlaveAdd));

            C_PROFI_SERVICE_DESCR profiSdb    = new C_PROFI_SERVICE_DESCR();
            C_DP_START_SEQ_REQ    startSeqLoc = new C_DP_START_SEQ_REQ();

            profiSdb.Layer       = (byte)LayerIdentifier.Dp;
            profiSdb.Service     = (byte)DPService.StartSeqLoc;
            profiSdb.Primitive   = (byte)ServicePrimitive.Req;

            startSeqLoc.RemAdd   = nMasterAdd;
            startSeqLoc.AreaCode = nSlaveAdd;
            startSeqLoc.Timeout  = 65000;

            if (!Global.WriteServiceInterface(profiSdb, startSeqLoc))
            {
                Global.DisplayLastError ();
                return false;
            }

            if (!DpReadServiceInterface())
            {
                Global.DisplayLastError();
                return false;
            }

            DpHandleEvent();

            return true;

        } // DpStartSeqLoc

        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function loads a slave parameter set to the DP master.
        /// In WindowsNT environment downloads are only possible before entering
        /// the operation mode STOP! After that you will get an 'access denied'
        /// error because the DP master and the NT driver are linked with each
        /// other and synchronize any data access. Thus, the common data base
        /// cannot be changed without reconfiguration.
        ///
        /// The DP master needs slave parameter sets to:
        /// - identify the type and station address of a DP slave,
        /// - to know about the number of inputs/outputs of the slave (CFG data),
        /// - to send the correct parameterization data incl. USER_PRM data.
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpDownloadLoc(byte nSlaveAdd, C_DP_SLAVE_PARA_SET slaveParaSet)
        {
            ushort i = 0;
            ushort nDataLength;
            ushort remainingLength;
            ushort offset;

            byte[] slaveParaSetAsArray    = slaveParaSet.ObjectToArray();
            nDataLength = remainingLength = (ushort) slaveParaSetAsArray.Length;
            do
            {
                Global.OutputString(string.Format("DP_Download_Loc.req ({0} {1})", i, nSlaveAdd));

                C_PROFI_SERVICE_DESCR profiSdb         = new C_PROFI_SERVICE_DESCR();
                C_DP_DOWNLOAD_REQ     dpDownload       = new C_DP_DOWNLOAD_REQ();

                profiSdb.Layer     = (byte)LayerIdentifier.Dp;
                profiSdb.Service   = (byte)DPService.DownloadLoc;
                profiSdb.Primitive = (byte)ServicePrimitive.Req;

                dpDownload.AreaCode  = nSlaveAdd;
                dpDownload.AddOffset = offset = (ushort)(i++ * DPConstants.MaxDownloaDataLen);

                if (remainingLength > (ushort)DPConstants.MaxDownloaDataLen)
                {
                   dpDownload.DataLen  = DPConstants.MaxDownloaDataLen;
                   remainingLength    -= DPConstants.MaxDownloaDataLen;
                }
                else
                {
                   dpDownload.DataLen = remainingLength;
                   remainingLength = 0;
                }
                dpDownload.Data = new byte[dpDownload.DataLen];

                Buffer.BlockCopy(slaveParaSetAsArray, dpDownload.AddOffset, dpDownload.Data, 0, dpDownload.DataLen);

                if (!Global.WriteServiceInterface(profiSdb, dpDownload))
                {
                    Global.DisplayLastError();
                    return false;
                }

                if (!DpReadServiceInterface())
                {
                    Global.DisplayLastError();
                    return false;
                }

                DpHandleEvent();

            } while (remainingLength > 0);

            return true;

        } // DpDownloadLoc

        //**************************************************************************

        /// <summary>
        /// End sequence loc.
        /// </summary>
        private bool DpEndSeqLoc(byte nMasterAdd)
        {
            Global.OutputString(string.Format("DP_EndSequence_Loc.req Master: ({0})", nMasterAdd));

            C_PROFI_SERVICE_DESCR profiSdb  = new C_PROFI_SERVICE_DESCR();
            C_DP_END_SEQ_REQ      endSeqLoc = new C_DP_END_SEQ_REQ();

            profiSdb.Layer     = (byte)LayerIdentifier.Dp;
            profiSdb.Service   = (byte)DPService.EndSeqLoc;
            profiSdb.Primitive = (byte)ServicePrimitive.Req;

            endSeqLoc.RemAdd = nMasterAdd;

            if (!Global.WriteServiceInterface(profiSdb, endSeqLoc))
            {
                Global.DisplayLastError();
                return false;
            }

            if (!DpReadServiceInterface())
            {
                Global.DisplayLastError();
                return false;
            }

            DpHandleEvent();

            return true;

        } // DpEndSeqLoc


        //**************************************************************************

        /// <summary>
        /// Get slave parameters
        /// </summary>
        private bool DpGetSlaveParam(byte nSlaveAdd)
        {
            C_PROFI_SERVICE_DESCR       profiSdb        = new C_PROFI_SERVICE_DESCR();

            C_DP_GET_SLAVE_PARAM_REQ    getSlaveParamReq = new C_DP_GET_SLAVE_PARAM_REQ();
            C_DP_GET_SLAVE_PARAM_CON    getSlaveParamCon = new C_DP_GET_SLAVE_PARAM_CON();
            C_DP_SLAVE_PARAM_SLAVE_INFO slaveParamInfo   = new C_DP_SLAVE_PARAM_SLAVE_INFO();

            profiSdb.Layer     = (byte)LayerIdentifier.Dp;
            profiSdb.Service   = (byte)DPService.GetSlaveParam;
            profiSdb.Primitive = (byte)ServicePrimitive.Req;

            getSlaveParamReq.Identifier = (byte) DPSlaveParam.SlaveInfo;
            getSlaveParamReq.RemAdd     = nSlaveAdd;

            if (!Global.WriteServiceInterface(profiSdb, getSlaveParamReq))
            {
                Global.DisplayLastError ();
                return false;
            }

            if (!DpReadServiceInterface())
            {
                Global.DisplayLastError();
                return false;
            }

            try
            {
                //Global.BuildObjectData(profiData, ref profiSdb);
                profiSdb = _profiSdb;

                if ((profiSdb.Service   != (byte)DPService.GetSlaveParam) ||
                    (profiSdb.Primitive != (byte)ServicePrimitive.Con   )    )
                {
                    Global.OutputString(OutputMode.OmScreenFile | OutputMode.OmStartLf, string.Format("invalid service response at slave {0} on DP_Get_Slave_Param.req", nSlaveAdd));
                    return false;
                }

                getSlaveParamCon = profiSdb.Data as C_DP_GET_SLAVE_PARAM_CON;
                if (getSlaveParamCon != null)
                {
                    slaveParamInfo = getSlaveParamCon.slaveParamSlaveInfo;
                    if (slaveParamInfo != null)
                    {
                        _slaveDescr[nSlaveAdd].nNumberInputs  = slaveParamInfo.NumberInputs;
                        _slaveDescr[nSlaveAdd].nNumberOutputs = slaveParamInfo.NumberOutputs;
                        _slaveDescr[nSlaveAdd].OffsetInputs   = slaveParamInfo.OffsetInputs;
                        _slaveDescr[nSlaveAdd].OffsetOutputs  = slaveParamInfo.OffsetOutputs;

                        return true;
                    }
                }
            }
            catch (Exception exc)
            {
               Global.DisplayError(exc);
            }

            return false;

        } // DpGetSlaveParam

        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function displays diagnostic messages.
        /// The standard DP diagnostic data (the first 6 bytes of the message)
        /// are decoded in readable text format. The following bytes (if present),
        /// the so called Extended Diagnostic Data, cannot be displayed. Their
        /// meaning is described in the Device Data Base (i.e. GSD- or GSx- files)
        /// of the respective DP slave.
        /// GSD files are always assigned to a DP slave by means of the IDENT
        /// NUMBER.
        /// German: GSD == GeraeteStammDaten, the letter 'x' in GSx represents the
        /// language (D = Default, E = English, G = German, etc.)
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpGetSlaveDiagConInd(C_PROFI_SERVICE_DESCR profiSdb, C_DP_GET_SLAVE_DIAG_CON getSlaveDiagConInd)
        {
            C_DP_DIAG_DATA diagData = new C_DP_DIAG_DATA();

            byte[] extDiagDataArray = new byte[0];
            int extDiagIndex;

            if (profiSdb.Result != (short)ServiceResult.Pos)
            {
                return false;
            }

            //pDiagData = (T_DP_DIAG_DATA*) (pGetSlaveDiagConInd + 1);
            diagData = getSlaveDiagConInd.DiagData;

            StringBuilder dpManagementSb = new StringBuilder();
            dpManagementSb.Append(string.Format("Slave [{0}] [{1:X4}]: Master = {2}, Ext_Diag_Len = {3}, Remaining diags = {4}\r\n",
                                  getSlaveDiagConInd.RemAdd,
                                  Global.Swap16(diagData.IdentNumber),
                                  diagData.MasterAdd,
                                  getSlaveDiagConInd.DiagDataLen - DPConstants.MinSlaveDiagLen,
                                  getSlaveDiagConInd.DiagEntries));

            if ((diagData.StationStatus1 & (byte)DPSlaveDiagBits.Diag1MasterLock) == (byte)DPSlaveDiagBits.Diag1MasterLock)
            {
                dpManagementSb.Append("\tSlave locked by another DP master\r\n");
            }
            if ((diagData.StationStatus1 & (byte)DPSlaveDiagBits.Diag1PrmFault) == (byte)DPSlaveDiagBits.Diag1PrmFault)
            {
                dpManagementSb.Append("\tPRM fault detected (i.e. slave requires special PRM data from GSD file)\r\n");
            }
            if ((diagData.StationStatus1 & (byte)DPSlaveDiagBits.Diag1InvalidSlaveRes) == (byte)DPSlaveDiagBits.Diag1InvalidSlaveRes)
            {
                dpManagementSb.Append("\tInvalid slave response (e.g. slave in wrong state)\r\n");
            }
            if ((diagData.StationStatus1 & (byte)DPSlaveDiagBits.Diag1NotSupported) == (byte)DPSlaveDiagBits.Diag1NotSupported)
            {
                dpManagementSb.Append("\tRequested feature not supported (e.g. SYNC, FREEZE)\r\n");
            }
            if ((diagData.StationStatus1 & (byte)DPSlaveDiagBits.Diag1ExtDiag) == (byte)DPSlaveDiagBits.Diag1ExtDiag)
            {
                dpManagementSb.Append("\tEXT diag bit set (usually an alarm indication, else a status message)\r\n");
            }
            if ((diagData.StationStatus1 & (byte)DPSlaveDiagBits.Diag1CfgFault) == (byte)DPSlaveDiagBits.Diag1CfgFault)
            {
                dpManagementSb.Append("\tCFG fault detected (check CFG bytes again -> use service DP_Get_Cfg)\r\n");
            }
            if ((diagData.StationStatus1 & (byte)DPSlaveDiagBits.Diag1StationNotReady) == (byte)DPSlaveDiagBits.Diag1StationNotReady)
            {
                dpManagementSb.Append("\tSlave not ready for data exchange (e.g. no input data available yet)\r\n");
            }
            if ((diagData.StationStatus1 & (byte)DPSlaveDiagBits.Diag1StationNonExistent) == (byte)DPSlaveDiagBits.Diag1StationNonExistent)
            {
                dpManagementSb.Append("\tSlave does not exist at this address (e.g. communication failure)\r\n");
            }
            if ((diagData.StationStatus2 & (byte)DPSlaveDiagBits.Diag2Deactivated) == (byte)DPSlaveDiagBits.Diag2Deactivated)
            {
                dpManagementSb.Append("\tSlave deactivated and released\r\n");
            }
            if ((diagData.StationStatus2 & (byte)DPSlaveDiagBits.Diag2SyncMode) ==(byte)DPSlaveDiagBits.Diag2SyncMode)
            {
                dpManagementSb.Append("\tSync mode enabled\r\n");
            }
            if ((diagData.StationStatus2 & (byte)DPSlaveDiagBits.Diag2FreezeMode) == (byte)DPSlaveDiagBits.Diag2FreezeMode)
            {
                dpManagementSb.Append("\tFreeze mode enabled\r\n");
            }
            if ((diagData.StationStatus2 & (byte)DPSlaveDiagBits.Diag2WDOn) == (byte)DPSlaveDiagBits.Diag2WDOn)
            {
                dpManagementSb.Append("\tWatchdog enabled\r\n");
            }
//            if ((diagData.StationStatus2 & (byte)DPSlaveDiagBits.Diag2Default) == (byte)DPSlaveDiagBits.Diag2Default)
//            {
//                dpManagementSb.Append("\tStatic diag reported by DP slave (i.e. default)\r\n");
//            }
            if ((diagData.StationStatus2 & (byte)DPSlaveDiagBits.Diag2StatDiag) == (byte)DPSlaveDiagBits.Diag2StatDiag)
            {
                dpManagementSb.Append("\tStatic diag reported by DP slave (i.e. no data exchange from now on)\r\n");
            }
            if ((diagData.StationStatus2 & (byte)DPSlaveDiagBits.Diag2PrmReq) == (byte)DPSlaveDiagBits.Diag2PrmReq)
            {
                dpManagementSb.Append("\tSlave requires new parameter data\r\n");
            }
            if ((diagData.StationStatus3 & (byte)DPSlaveDiagBits.Diag3ExtDiagOverflow) == (byte)DPSlaveDiagBits.Diag3ExtDiagOverflow)
            {
                dpManagementSb.Append("\tDiagnostic buffer overflow (e.g. diag buffer too small)\r\n");
            }

            dpManagementSb.Remove(dpManagementSb.Length - 2, 2);
            Global.OutputString(dpManagementSb.ToString());

            //--- write EXT_DIAG data to log file ------------------------------------

            if (getSlaveDiagConInd.DiagDataLen > (ushort)DPConstants.MinSlaveDiagLen)
            {
                //pExtDiagData = (OCTET*) pDiagData;
                extDiagDataArray = diagData.ExtDiagData;

                dpManagementSb.Remove(0, dpManagementSb.Length);
                dpManagementSb.Append(string.Format("EXT_DIAG [{0}]: ", getSlaveDiagConInd.RemAdd));

                int minSlaveDiagLen = (int)DPConstants.MinSlaveDiagLen;
                for (extDiagIndex = minSlaveDiagLen; extDiagIndex < getSlaveDiagConInd.DiagDataLen; extDiagIndex++)
                {
                    if (extDiagIndex == getSlaveDiagConInd.DiagDataLen - 1)
                    {
                        dpManagementSb.Append(string.Format("0x{0:X2}", extDiagDataArray[extDiagIndex - minSlaveDiagLen]));
                    }
                    else
                    {
                        dpManagementSb.Append(string.Format("0x{0:X2},", extDiagDataArray[extDiagIndex - minSlaveDiagLen]));
                    }

                } // for ExtDiagIndex

                dpManagementSb.Append("\r\n");
                Global.OutputString(OutputMode.OmFile, dpManagementSb.ToString());

            } // if ExtDiagData

            return (getSlaveDiagConInd.DiagEntries > 0);

        } // DpGetSlaveDiagConInd


        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This service retrieves the last entry (oldest in order) of the DP
        /// master diagnostic data history (circular buffer). Call this function
        /// as long as there are remaining entries within the list (element
        /// 'diag_entries > 0' within the response).
        /// With the diagnostic data the master keeps you informed about the state
        /// of all configured DP slaves. The master will automatically indicate
        /// the presence of new diagostic data by means of a DP_Get_Slave_Diag
        /// indication message. This message already contains the last diagnostic
        /// entry.
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool DpGetSlaveDiagReq()
        {
            Global.OutputString("DP_Get_Slave_Diag.req");

            C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();

            profiSdb.Layer     = (byte)LayerIdentifier.Dp;
            profiSdb.Service   = (byte)DPService.GetSlaveDiag;
            profiSdb.Primitive = (byte)ServicePrimitive.Req;

            return Global.WriteServiceInterface(profiSdb, null);

        } // DpGetSlaveDiagReq


        //**************************************************************************

        /// <summary>
        /// Write (set) data to the PAPI
        /// </summary>
        private bool DpWriteDataInterface(byte[] data, uint offset, uint dataSize, out uint pBytesWritten)
        {
           bool retVal       = true;
           uint retryCounter = DpMaxServiceRetry;

           do
           {
              retVal = Global.SetData(data, offset, dataSize, out pBytesWritten);
           }  while ((!retVal) && (retryCounter-- > 0));

           return retVal;
        }

        //**************************************************************************

        /// <summary>
        /// Read (get) data from the PAPI
        /// </summary>
        private bool DpReadDataInterface(out byte[] data, uint offset, uint dataSize, out uint pBytesRead)
        {
           return Global.GetData(out data, offset, dataSize, out pBytesRead);
        }

        //**************************************************************************

        private void DpHandleEvent()
        {
           lock (_profiSdb)
           {
              DpHandleEvent(_profiSdb);
           }
        }

        static bool bGetSlaveDiagReqPending = false;
        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// This function is used to identify and display any response message
        /// from the DP master. Further this function controls the operation mode
        /// of the DP master (STOP, CLEAR, OPERATE).
        /// The variables:
        /// - nCurrentOpMode  : present operation mode of the DP master,
        /// - nRequestedOpMode: operation mode of the last DP_Act_Param_Loc.req,
        /// - nDesiredOpMode  : operation mode that should be reached finally
        ///                     control the operation mode handling.
        /// Call this function whenever you have read a message from the DP
        /// management device.
        /// ------------------------------------------------------------------------
        /// </summary>
        private void DpHandleEvent(C_PROFI_SERVICE_DESCR profiSdb)
        {
            C_DP_GET_SLAVE_DIAG_CON diagConInd  = new C_DP_GET_SLAVE_DIAG_CON();
            C_DP_ACT_PARAM_IND      actParamInd = new C_DP_ACT_PARAM_IND();

            string dpManagementString;
            StringBuilder resultString = new StringBuilder();

            try
            {
                if (profiSdb != null)
                {
                    if (profiSdb.Result == (short)ServiceResult.Pos)
                    {
                       resultString.Append(".con [+]");
                    }
                    else
                    {
                       resultString.Append(".con [-]");
                    } // else POS

                    switch ((ServicePrimitive)profiSdb.Primitive)
                    {
                        //--- confirmations services -----------------------------------------

                        case ServicePrimitive.Con:
                        {
                           switch (profiSdb.Service)
                           {
                              case (byte) DPService.InitMaster:
                              {
                                 if (profiSdb.Result == (short)ServiceResult.Pos)
                                 {
                                    dpManagementString = string.Format("DP_Init_Master{0}", resultString.ToString());
                                    _nCurrentOpMode   = DPOPMode.Offline;
                                    _nRequestedOpMode = DPOPMode.Offline;
                                    _nDesiredOpMode   = DPOPMode.Offline;
                                 }
                                 else
                                 {
                                    C_DP_INIT_MASTER_CON initMasterCon = profiSdb.Data as C_DP_INIT_MASTER_CON;
                                    resultString.Append(string.Format(" 0x{0:X4} ", initMasterCon.Status));
                                    dpManagementString = string.Format("DP_Init_Master{0}", resultString.ToString());
                                 }
                                 Global.OutputString(dpManagementString);
                              }
                              break;

                              case (byte)DPService.ExitMaster:
                              {
                                 if (profiSdb.Result == (short)ServiceResult.Pos)
                                 {
                                    dpManagementString = string.Format("DP_Exit_Master{0}", resultString.ToString());
                                    _nCurrentOpMode    = DPOPMode.Offline;
                                    _nRequestedOpMode = DPOPMode.Offline;
                                    _nDesiredOpMode   = DPOPMode.Offline;
                                 }
                                 else
                                 {
                                    C_DP_EXIT_MASTER_CON exitMasterCon = profiSdb.Data as C_DP_EXIT_MASTER_CON;
                                    resultString.Append(string.Format(" 0x{0:X4} ", exitMasterCon.Status));
                                    dpManagementString = string.Format("DP_Exit_Master{0}", resultString.ToString());
                                 }
                                 Global.OutputString(dpManagementString);
                              }
                              break;

                              case (byte)DPService.SetBusParameter:
                              {
                                 if (profiSdb.Result == (short)ServiceResult.Pos)
                                 {
                                    dpManagementString = string.Format("DP_Set_Busparameter{0}", resultString.ToString());
                                 }
                                 else
                                 {
                                    C_DP_SET_BUSPARAMETER_CON setBusParamCon = profiSdb.Data as C_DP_SET_BUSPARAMETER_CON;
                                    resultString.Append(string.Format(" 0x{0:X4} ", setBusParamCon.Status));
                                    dpManagementString = string.Format("DP_Set_Busparameter{0}", resultString.ToString());
                                    Global.OutputString(dpManagementString);
                                 }
                              }
                              break;

                              case (byte)DPService.StartSeqLoc:
                              {
                                 if (profiSdb.Result == (short)ServiceResult.Pos)
                                 {
                                    dpManagementString = string.Format("DP_StartSequence_Loc{0}", resultString.ToString());
                                 }
                                 else
                                 {
                                    C_DP_START_SEQ_RES_CON startSeqCon = profiSdb.Data as C_DP_START_SEQ_RES_CON;
                                    resultString.Append(string.Format(" 0x{0:X4} ", startSeqCon.Status));
                                    dpManagementString = string.Format("DP_StartSequence_Loc{0}", resultString.ToString());
                                 }
                                 Global.OutputString(dpManagementString);
                              }
                              break;

                              case (byte)DPService.EndSeqLoc:
                              {
                                 if (profiSdb.Result == (short)ServiceResult.Pos)
                                 {
                                    dpManagementString = string.Format("DP_EndSequence_Loc{0}", resultString.ToString());
                                 }
                                 else
                                 {
                                    C_DP_END_SEQ_RES_CON endSeqCon = profiSdb.Data as C_DP_END_SEQ_RES_CON;
                                    resultString.Append(string.Format(" 0x{0:X4} ", endSeqCon.Status));
                                    dpManagementString = string.Format("DP_EndSequence_Loc{0}", resultString.ToString());
                                 }
                                 Global.OutputString(dpManagementString);
                              }
                              break;

                              case (byte)DPService.DownloadLoc:
                              {
                                 if (profiSdb.Result == (short)ServiceResult.Pos)
                                 {
                                    dpManagementString = string.Format("DP_Download_Loc{0}", resultString.ToString());
                                 }
                                 else
                                 {
                                    C_DP_DOWNLOAD_RES_CON downlCon = profiSdb.Data as C_DP_DOWNLOAD_RES_CON;
                                    resultString.Append(string.Format(" 0x{0:X4} ", downlCon.Status));
                                    dpManagementString = string.Format("DP_Download_Loc{0}", resultString.ToString());
                                 }
                                 Global.OutputString(dpManagementString);
                              }
                              break;

                              case (byte)DPService.ActParamLoc:
                              {
                                 if (profiSdb.Result == (short)ServiceResult.Pos)
                                 {
                                    dpManagementString = string.Format("DP_Act_Param_Loc{0}", resultString.ToString());
                                    Global.OutputString(dpManagementString);
                                    _nCurrentOpMode = _nRequestedOpMode;

                                    if (_nCurrentOpMode < _nDesiredOpMode)
                                    {
                                       if (!DpActParamLocReq((DPOPMode)(_nCurrentOpMode + (DPOPMode.Operate - DPOPMode.Clear))))
                                       {
                                         Global.DisplayLastError();
                                       }
                                    } // if CurrentOpMode < OPERATE
                                 } // if POS
                                 else
                                 {
                                    C_DP_ACT_PARAM_RES_CON actParamCon = profiSdb.Data as C_DP_ACT_PARAM_RES_CON;
                                    resultString.Append(string.Format(" 0x{0:X4} ", actParamCon.Status));
                                    if (_nNumberOfSlaves == 0)
                                    {
                                       dpManagementString = string.Format("DP_Act_Param_Loc{0}\tNo slaves found or defined", resultString.ToString());
                                    }
                                    else
                                    {
                                       dpManagementString = string.Format("DP_Act_Param_Loc{0}", resultString.ToString());
                                    }
                                    Global.OutputString(dpManagementString);
                                  } // else POS

                              } // case DP_ACT_PARAM_LOC
                              break;

                              case (byte)DPService.GetSlaveDiag:
                              {
                                 if (profiSdb.Result == (short)ServiceResult.Pos)
                                 {
                                    dpManagementString = string.Format("DP_Get_Slave_Diag{0}", resultString.ToString());
                                    Global.OutputString(dpManagementString);

                                    diagConInd = profiSdb.Data as C_DP_GET_SLAVE_DIAG_CON;
                                    if (diagConInd != null)
                                    {
                                       if (DpGetSlaveDiagConInd(profiSdb, diagConInd))
                                       {
                                          bGetSlaveDiagReqPending = true;

                                          if (!DpGetSlaveDiagReq())
                                          {
                                             Global.DisplayLastError();
                                             bGetSlaveDiagReqPending = false;
                                          }
                                       }
                                       else
                                       {
                                          bGetSlaveDiagReqPending = false;
                                       }
                                    }
                                 }
                                 else
                                 {
                                    C_DP_GET_SLAVE_DIAG_CON getSlaveDiagCon = profiSdb.Data as C_DP_GET_SLAVE_DIAG_CON;
                                    resultString.Append(string.Format(" 0x{0:X4} ", getSlaveDiagCon.Status));
                                    dpManagementString = string.Format("DP_Get_Slave_Diag{0}", resultString.ToString());
                                    Global.OutputString(dpManagementString);
                                 }
                              } // case DP_GET_SLAVE_DIAG
                              break;

                              case (byte)DPService.GlobalControl:
                              {
                                 if (profiSdb.Result == (short)ServiceResult.Pos)
                                 {
                                    dpManagementString = string.Format("DP_Global_Control{0}", resultString.ToString());
                                 }
                                 else
                                 {
                                    C_DP_GLOBAL_CONTROL_CON globalControlCon = profiSdb.Data as C_DP_GLOBAL_CONTROL_CON;
                                    resultString.Append(string.Format(" 0x{0:X4} ", globalControlCon.Status));
                                    dpManagementString = string.Format("DP_Global_Control{0}", resultString.ToString());
                                 }
                                 Global.OutputString(dpManagementString);
                              } // case DP_GLOBAL_CONTROL
                              break;

                              default:
                              {
                                 dpManagementString = string.Format("DP: invalid service confirmation = 0x{0:X2}", profiSdb.Service);
                                 Global.OutputString(dpManagementString);
                              }
                              break;
                           } // switch service
                        } // case CON
                        break;

                        //--- indication services --------------------------------------------

                        case ServicePrimitive.Ind:
                            {
                                switch (profiSdb.Service)
                                {
                                    case (byte)DPService.ActParamLoc:
                                        {
                                            // pActParamInd = (T_DP_ACT_PARAM_IND*)(pProfiSdb + 1);
                                            actParamInd = profiSdb.Data as C_DP_ACT_PARAM_IND;
                                            if (actParamInd != null)
                                            {
                                                if (actParamInd.AreaCode == (byte)DPAreaCodes.SetMode)
                                                {
                                                    switch ((DPOPMode)actParamInd.Activate)
                                                    {
                                                        case DPOPMode.Offline: Global.OutputString("DP_Act_Param_Loc.ind [OFFLINE]"); break;
                                                        case DPOPMode.Stop:    Global.OutputString("DP_Act_Param_Loc.ind [STOP]"); break;
                                                        case DPOPMode.Clear:   Global.OutputString("DP_Act_Param_Loc.ind [CLEAR]"); break;
                                                        case DPOPMode.Operate: Global.OutputString("DP_Act_Param_Loc.ind [OPERATE]"); break;

                                                        default:
                                                            {
                                                                dpManagementString = string.Format("DP_Act_Param_Loc.ind [op_mode = 0x{0:X2}]", actParamInd.Activate);
                                                                Global.OutputString(dpManagementString);
                                                            }
                                                            break;
                                                    }

                                                    _nCurrentOpMode = (DPOPMode)actParamInd.Activate;

                                                } // if DP_AREA_SET_MODE
                                                else
                                                {
                                                    dpManagementString = string.Format("DP_Act_Param_Loc.ind [area_code = 0x{0:X2}] [op_mode = 0x{0:X2}]",
                                                           actParamInd.AreaCode,
                                                           actParamInd.Activate);
                                                    Global.OutputString(dpManagementString);

                                                } // else DP_AREA_SET_MODE
                                            }
                                            break;
                                        } // case DP_ACT_PARAM_LOC

                                    case (byte)DPService.GetSlaveDiag:
                                        {
                                            Global.OutputString("DP_Get_Slave_Diag.ind");

                                            // pDiagConInd = (T_DP_GET_SLAVE_DIAG_CON*) (pProfiSdb + 1);
                                            diagConInd = profiSdb.Data as C_DP_GET_SLAVE_DIAG_CON;
                                            if (diagConInd != null)
                                            {
                                                if ((DpGetSlaveDiagConInd(profiSdb, diagConInd)) &&
                                                     (!bGetSlaveDiagReqPending))
                                                {
                                                    bGetSlaveDiagReqPending = true;
                                                    if (!DpGetSlaveDiagReq())
                                                    {
                                                        Global.DisplayLastError();
                                                        bGetSlaveDiagReqPending = false;
                                                    }
                                                }
                                            }
                                            break;
                                        } // case DP_GET_SLAVE_DIAG

                                    default:
                                        {
                                            Global.OutputString(string.Format("DP: invalid service indication = 0x{0:X2}", profiSdb.Service));
                                        }
                                        break;
                                } // switch service

                                break;
                            } // case IND

                        //--------------------------------------------------------------------

                        default:
                            {
                                dpManagementString = string.Format("DP: invalid primitive = 0x{0:X2}", profiSdb.Primitive);
                                Global.OutputString(dpManagementString);
                            }
                            break;
                    } // switch primitive
                } // if nNumberOfBytesRead
                else
                {
                    Global.OutputString("DP: service event with zero length received");

                } // else nNumberOfBytesRead
            }
            catch (Exception exc)
            {
                Global.DisplayError(exc);
            }

            return;

        } // DpHandleEvent

        //**************************************************************************

        /// <summary>
        /// DP service configuration indication flag received
        /// </summary>
        private void DPServiceConIndReceived(object sender, EventArgs<C_PROFI_SERVICE_DESCR> e)
        {
           _profiSdb = e.Object as C_PROFI_SERVICE_DESCR;
           Global.EwhDpRead.Set();

        } // DPServiceConIndReceived

        //**************************************************************************

        /// <summary>
        /// Reads periodically the data from the pipe where it is saved periodically the received
        /// indication/confirmation buffer data
        /// </summary>
        private bool DpReadServiceInterface()
        {
           bool bResult = true;
           int waitResult = -1;
           do
           {
              waitResult = WaitHandle.WaitAny(_waitDpReadServiceHandles, 10, false);
              if (0 == waitResult)
              {
                  Global.CanReadServiceInterface = true;
                  return true;
              }
              else if (1 == waitResult)
              {
                Global.EwhExit.Set();
                return false;
              }
           }  while (waitResult == WaitHandle.WaitTimeout);

           return bResult;
        }  // DpReadServiceInterface


        //**************************************************************************

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// - This function display the data up to 16 bytes per line in hexadecimal
        ///   format. If the length of data is greater then 16 bytes, the data
        ///   will be splitted in blocks of 16 bytes and each block will be
        ///   displayed in a separate line.
        /// ------------------------------------------------------------------------
        /// </summary>
        private void ShowHexDump(byte[] dump, byte nLength)
        {
            int i;
            int line;

            StringBuilder lineSb = new StringBuilder();

            line = 0;
            lineSb.Append(string.Format("[{0:D2}]: ", line));

            for (i = 0; i < nLength; i++)
            {
                lineSb.Append(string.Format("0x{0:X2} ", dump[i]));

                if ((i % 16 == 0) && (i > 0))
                {
                    Global.OutputString(OutputMode.OmScreen, lineSb.ToString());
                    line++;
                    lineSb.Remove(0, lineSb.Length);
                    lineSb.Append(string.Format("{0:D2}: ", line));
                }
            }

            Global.OutputString(OutputMode.OmScreen, lineSb.ToString());

            return;

        } // ShowHexDump

        //**************************************************************************

        #endregion
    }
}
