using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

using PapiWrapper.Common;
using PapiWrapper.Fmb;
using System.IO;
using System.Runtime.Serialization;
using System.Xml;
using System.Runtime.InteropServices;
using System.ComponentModel;
using System.Collections;
using PapiWrapper.Dp;

using GlobalModule;
using DpModule;

namespace FmbModule
{
  /// <summary>
  /// This FmbMmodule contains all functionality which is relevant to the Basic Management (FMB).
  /// It implements the 'BasicManagementThread' which is responsible for initializing
  /// the PROFIBUS protocol software and will start the DP protocol component if successful.
  ///
  /// The Basic Management contains all 'system-wide' functionality like memory
  /// configuration or bus parameter settings. It will also handle the system events
  /// (FMB events).
  /// </summary>
  public class FmbLayer
  {
      #region Fields

      #region FMB data

      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataKbaud_9_6;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataKbaud_19_2;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataKbaud_93_75;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataKbaud_187_5;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataKbaud_500;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataMbaud_1_5;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataMbaud_3;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataMbaud_6;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataMbaud_12;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataKbaud_45_45;
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameterReqDataKbaud_93_75_PF;


      //--- main configuration and bus parameter administration memory --------------------------------
      private C_FMB_SET_BUSPARAMETER_REQ  _fmbSetBusparameter;

      #endregion

      private static FmbLayer _instance;
      private Thread          _dpManagementThread;

      /// <summary>
      /// wait handles for: fmb read service + exit
      /// </summary>
      private static WaitHandle[] _waitFmbReadServiceHandles;

      private C_PROFI_SERVICE_DESCR _profiSdb = new C_PROFI_SERVICE_DESCR();
      #endregion


      #region Constructor
      private FmbLayer()
      {
         _fmbSetBusparameter                 = new C_FMB_SET_BUSPARAMETER_REQ();
         _fmbSetBusparameterReqDataKbaud_9_6 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.KBaud_9_6,                                         // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 100,                                                                 // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 60,                                                                  // Max Tsdr: maximum station delay time responder
            Tqui          = 0,                                                                   // Tqui: quiet time
            Tset          = 1,                                                                   // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 1,                                                                   // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 1                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataKbaud_19_2 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.KBaud_19_2,                                        // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 100,                                                                 // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 60,                                                                  // Max Tsdr: maximum station delay time responder
            Tqui          = 0,                                                                   // Tqui: quiet time
            Tset          = 1,                                                                   // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 1,                                                                   // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 1                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataKbaud_93_75 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.KBaud_93_75,                                       // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 100,                                                                 // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 60,                                                                  // Max Tsdr: maximum station delay time responder
            Tqui          = 0,                                                                   // Tqui: quiet time
            Tset          = 1,                                                                   // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 1,                                                                   // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 1                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataKbaud_187_5 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.KBaud_187_5,                                       // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 100,                                                                 // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 60,                                                                  // Max Tsdr: maximum station delay time responder
            Tqui          = 0,                                                                   // Tqui: quiet time
            Tset          = 1,                                                                   // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 1,                                                                   // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 1                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataKbaud_500 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.KBaud_500,                                         // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 100,                                                                 // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 100,                                                                 // Max Tsdr: maximum station delay time responder
            Tqui          = 0,                                                                   // Tqui: quiet time
            Tset          = 1,                                                                   // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 1,                                                                   // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 1                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataMbaud_1_5 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.MBaud_1_5,                                         // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 300,                                                                 // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 150,                                                                 // Max Tsdr: maximum station delay time responder
            Tqui          = 0,                                                                   // Tqui: quiet time
            Tset          = 1,                                                                   // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 10,                                                                  // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 3                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataMbaud_3 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.MBaud_3,                                           // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 400,                                                                 // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 250,                                                                 // Max Tsdr: maximum station delay time responder
            Tqui          = 3,                                                                   // Tqui: quiet time
            Tset          = 4,                                                                   // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 10,                                                                  // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 3                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataMbaud_6 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.MBaud_6,                                           // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 600,                                                                 // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 450,                                                                 // Max Tsdr: maximum station delay time responder
            Tqui          = 6,                                                                   // Tqui: quiet time
            Tset          = 8,                                                                   // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 10,                                                                  // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 3                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataMbaud_12 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.MBaud_12,                                          // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 1000,                                                                // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 800,                                                                 // Max Tsdr: maximum station delay time responder
            Tqui          = 9,                                                                   // Tqui: quiet time
            Tset          = 16,                                                                  // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 10,                                                                  // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 4                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataKbaud_45_45 = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.KBaud_45_45,                                       // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 640,                                                                 // Tsl: slot time */
            MinTsdr       = 11,                                                                  // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 400,                                                                 // Max Tsdr: maximum station delay time responder
            Tqui          = 0,                                                                   // Tqui: quiet time
            Tset          = 95,                                                                  // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 10,                                                                  // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 1                                                                    // Retry: maximum retry limit on telegram failure
         };

         _fmbSetBusparameterReqDataKbaud_93_75_PF = new C_FMB_SET_BUSPARAMETER_REQ()
         {
            LocAdd        = 0,                                                                   // local station address (FDL address)
            LocSegm       = 0,                                                                   // local segment: not used - do not change
            Baudrate      = (byte)FmbBaudrate.KBaud_93_75_PF,                                    // baud rate
            MediumRed     = (byte)MediumRedundancy.NoRedundancy,                                 // medium redundancy: do not change
            Tsl           = 8000,                                                                // Tsl: slot time */
            MinTsdr       = 250,                                                                 // Min Tsdr: minimum station delay time responder
            MaxTsdr       = 1000,                                                                // Max Tsdr: maximum station delay time responder
            Tqui          = 150,                                                                 // Tqui: quiet time
            Tset          = 200,                                                                 // Tset: setup time
            Ttr           = 300000u,                                                             // Ttr: target token rotation time
            G             = 10,                                                                  // GAP: gap update factor - ignored by ASPC2-Asic
            InRingDesired = (byte)Constants.PbTrue,                                              // active or passive station - Master/Slave
            Hsa           = 126,                                                                 // HSA: highest station address
            MaxRetryLimit = 4                                                                    // Retry: maximum retry limit on telegram failure
         };


         _waitFmbReadServiceHandles = new WaitHandle[2] { Global.EwhFmbRead, Global.EwhExit };

         Global.FMBServiceConIndReceived += new EventHandler<EventArgs<C_PROFI_SERVICE_DESCR>>(FMBServiceConIndReceived);
      }

      #endregion

      #region Properties
      public static FmbLayer Instance
      {
         get
         {
            if (_instance == null)
            {
              _instance = new FmbLayer();
            }
            return _instance;
         }
      }
      #endregion


      #region Public methods

      /// <summary>
      /// BasicManagementThread
      /// This thread is responsible for initializing the entire PROFIBUS
      /// protocol components. All system-wide parameters like memory settings,
      /// bus parameters are handled here.
      /// This function will also receive the management events (FMB events).
      /// These events inform about general problems (e.g. with the PROFIBUS
      /// physical layer).
      ///
      /// The read timeout for all devices will be set to WAIT_FOREVER. Then all
      /// threads are blocked as long as no new message is available.
      /// The write timeout determines the time within the driver automatically
      /// retries a service request when the previous one failed. If no response
      /// is received during this time (i.e. any DP service) then the operation
      /// will be aborted (e.g. error E_SERVICE_CONSTRAIN_CONFLICT). Please
      /// consider that this timeout can be dependent on the bus frequency.
      /// Some service need considerably more time with a slow baudrate. During
      /// this time no message will be available at the DP interface.
      ///
      /// If the initialization is successful the DpManagementThread will be
      /// created.
      ///
      /// </summary>
      public void BasicManagementThread()
      {
         Global.OutputSetThreadIdString("  FMB");
         Global.OutputString("BasicManagementThread started");

         // increase thread priority for fmb layer thread
         Thread.CurrentThread.Priority++;

         // set basic configuration -----------------------------------------------------------
         if (!FmbSetConfigurationReq())
         {
            Global.OutputString("BasicManagementThread terminated");
            return;
         }
         if (!FmbReadServiceInterface())
         {
            Global.OutputString("BasicManagementThread terminated");
            return;
         }
         FmbHandleEvent();


         // set bus parameters ----------------------------------------------------------------
         if (!FmbSetBusparameterReq())
         {
            // todo: display last error
            Global.OutputString("BasicManagementThread terminated");
            return;
         }
         if (!FmbReadServiceInterface())
         {
            Global.OutputString("BasicManagementThread terminated");
            return;
         }
         FmbHandleEvent();

         // --- wait some time to allow baud rate detection of the DP slaves -------

         Global.OutputString("wait some time to allow baud rate detection of the DP slaves");

         for (int nDelay = 20; nDelay >= 0; nDelay--)                // units of 100ms
         {
             string fmbManagementString = string.Format("time left for slave baud rate detection: {0}.{1}s\r", nDelay / 10,nDelay % 10);
             Global.OutputString(OutputMode.OmSuppressId | OutputMode.OmScreen | OutputMode.OmSuppressLf, fmbManagementString);

             Thread.Sleep(100);  /* wait 100ms */
         }

         Global.OutputString(OutputMode.OmSuppressId | OutputMode.OmScreen, "");


         _dpManagementThread = new Thread(DpLayer.Instance.DpManagementThread);
         _dpManagementThread.Start();

         // handle Basic Management events ----------------------------------------------------
         bool result;
         do
         {
            result = FmbReadServiceInterface();
            if (result)
            {
               FmbHandleEvent();
            }

         } while (result);                    // terminates in case of errors only

         Global.DisplayLastError();
         Global.OutputString("BasicManagementThread terminated");

      }  // BasicManagementThread


      //**************************************************************************************
      //**************************************************************************************

      /// <summary>
      /// ------------------------------------------------------------------------
      /// FUNCTIONAL_DESCRIPTION
      /// ------------------------------------------------------------------------
      /// Place all module specific initialization here. This function is called
      /// before any communication is started or initialized.
      /// ------------------------------------------------------------------------
      /// </summary>
      public void InitFmbLayer(int baudrate)
      {
         switch ((FmbBaudrate)baudrate)
         {
            case FmbBaudrate.KBaud_9_6:      _fmbSetBusparameter = _fmbSetBusparameterReqDataKbaud_9_6;      break;
            case FmbBaudrate.KBaud_19_2:     _fmbSetBusparameter = _fmbSetBusparameterReqDataKbaud_19_2;     break;
            case FmbBaudrate.KBaud_93_75:    _fmbSetBusparameter = _fmbSetBusparameterReqDataKbaud_93_75;    break;
            case FmbBaudrate.KBaud_187_5:    _fmbSetBusparameter = _fmbSetBusparameterReqDataKbaud_187_5;    break;
            case FmbBaudrate.KBaud_500:      _fmbSetBusparameter = _fmbSetBusparameterReqDataKbaud_500;      break;
            case FmbBaudrate.MBaud_1_5:      _fmbSetBusparameter = _fmbSetBusparameterReqDataMbaud_1_5;      break;
            case FmbBaudrate.MBaud_3:        _fmbSetBusparameter = _fmbSetBusparameterReqDataMbaud_3;        break;
            case FmbBaudrate.MBaud_6:        _fmbSetBusparameter = _fmbSetBusparameterReqDataMbaud_6;        break;
            case FmbBaudrate.MBaud_12:       _fmbSetBusparameter = _fmbSetBusparameterReqDataMbaud_12;       break;
            case FmbBaudrate.KBaud_45_45:    _fmbSetBusparameter = _fmbSetBusparameterReqDataKbaud_45_45;    break;
            case FmbBaudrate.KBaud_93_75_PF: _fmbSetBusparameter = _fmbSetBusparameterReqDataKbaud_93_75_PF; break;
            default:                         _fmbSetBusparameter = _fmbSetBusparameterReqDataMbaud_1_5;      break;
         }

      }  //InitFmbLayer

      //**************************************************************************************

      /// <summary>
      /// /*------------------------------------------------------------------------*/
      /// /* FUNCTIONAL_DESCRIPTION                                                 */
      /// /*------------------------------------------------------------------------*/
      /// /* This function converts the standard DP bus parameter set into the      */
      /// /* format needed for calling FMB_Set_Busparameter. Special DP parameters  */
      /// /* which are not common with the FMB can be set with DP services          */
      /// /* (see function DpSetBusparameterReq in DPLayer.c)                       */
      /// /*------------------------------------------------------------------------*/
      /// </summary>
      public bool FmbSetBusParaSet(C_DP_BUS_PARA_SET dpBusParaSet)
      {
          Global.OutputString("- import FMB bus parameters ");

          _fmbSetBusparameter.LocAdd        = dpBusParaSet.FdlAdd;
          _fmbSetBusparameter.LocSegm       = 0;
          _fmbSetBusparameter.Baudrate      = dpBusParaSet.BaudRate;
          _fmbSetBusparameter.MediumRed     = (byte)MediumRedundancy.NoRedundancy;
          _fmbSetBusparameter.Tsl           = Global.Swap16(dpBusParaSet.Tsl);
          _fmbSetBusparameter.MinTsdr       = Global.Swap16(dpBusParaSet.MinTsdr);
          _fmbSetBusparameter.MaxTsdr       = Global.Swap16(dpBusParaSet.MaxTsdr);
          _fmbSetBusparameter.Tqui          = dpBusParaSet.Tqui;
          _fmbSetBusparameter.Tset          = dpBusParaSet.Tset;
          _fmbSetBusparameter.Ttr           = Global.Swap32(dpBusParaSet.Ttr);
          _fmbSetBusparameter.G             = dpBusParaSet.G;
          _fmbSetBusparameter.InRingDesired = (byte)Constants.PbTrue;
          _fmbSetBusparameter.Hsa           = dpBusParaSet.Hsa;
          _fmbSetBusparameter.MaxRetryLimit = dpBusParaSet.MaxRetryLimit;

          Global.Baudrate                   = dpBusParaSet.BaudRate;
          Global.MasterAddr                 = dpBusParaSet.FdlAdd;

          return true;

      } /* FmbSetBusParaSet */

      //**************************************************************************************

      public void TerminateFmbLayer()
      {
         FmbExitReq();

      } /* TerminateFmbLayer */

      #endregion

      //**************************************************************************************
      //**************************************************************************************


      #region Private methods

      /// <summary>
      /// ------------------------------------------------------------------------
      /// FUNCTIONAL_DESCRIPTION
      /// ------------------------------------------------------------------------
      /// - coding and sending of FMB_Set_Configuration.req service
      /// - the configuration data is defined above
      ///
      /// FMB_Set_Configuration defines the memory and resource usage of the
      /// PROFIBUS protocol software. In case of multiprotocol operation (e.g.
      /// FMS and DP protocol in parallel) this service defines the amount of
      /// resources which can be used by the different components.
      /// The service belongs to the Basic Management component (FMB) and has to
      /// be executed before any other service.
      /// ------------------------------------------------------------------------
      /// </summary>
      private bool FmbSetConfigurationReq()
      {
         Global.OutputString("FMB_Set_Configuration.req");

         C_PROFI_SERVICE_DESCR       profiSdb                   = new C_PROFI_SERVICE_DESCR();
         C_FMB_SET_CONFIGURATION_REQ fmbSetConfigurationReqData = new C_FMB_SET_CONFIGURATION_REQ();

         profiSdb.Layer     = (byte) LayerIdentifier.Fmb;
         profiSdb.Service   = (byte) FMBService.SetConfiguration;
         profiSdb.Primitive = (byte) ServicePrimitive.Req;

         fmbSetConfigurationReqData.DpActive               = (byte)  Constants.PbTrue;           // dp_active
         fmbSetConfigurationReqData.FdlEvtReceiver         = (ushort)LayerIdentifier.FmbUsr;     // fdl_evt_receiver
         fmbSetConfigurationReqData.DataBufferLength       = Global.MaxPduDataBufferLength;      // data_buffer_length


         fmbSetConfigurationReqData.Dp.MaxNumberSlaves     = Global.MaxNumberDPSlaves;           // max_number_slaves
         fmbSetConfigurationReqData.Dp.MaxSlaveOutputLen   = Global.MaxIoTelegramLength;         // max_slave_output_len
         fmbSetConfigurationReqData.Dp.MaxSlaveInputLen    = Global.MaxIoTelegramLength;         // max_slave_input_len
         fmbSetConfigurationReqData.Dp.MaxSlaveDiagLen     = Global.MaxSlaveDiagLength;          // max_slave_diag_len
         fmbSetConfigurationReqData.Dp.MaxSlaveDiagEntries = Global.MaxSlaveDiagEntries;         // max_slave_diag_entries
         fmbSetConfigurationReqData.Dp.MaxBusParaLen       = Global.MaxBusParaLen;               // max_bus_para_len: not used here, see FMB
         fmbSetConfigurationReqData.Dp.MaxSlaveParaLen     = Global.MaxSlaveParaLen;             // max_slave_para_len

         if (Global.WriteServiceInterface(profiSdb, fmbSetConfigurationReqData))
         {
           return true;
         }

         return false;
      }  // FmbSetConfigurationReq


      /// <summary>
      /// ------------------------------------------------------------------------
      /// FUNCTIONAL_DESCRIPTION
      /// ------------------------------------------------------------------------
      /// - coding and sending of FMB_Set_Busparameter.req service
      /// - the bus parameter are defined above
      ///
      /// FMB_Set_Busparameter sets the bus parameters for all PROFIBUS
      /// protocol components. It replaces the protocol specific bus parameter
      /// procedures of all components (e.g. DP_Download_Loc (Bus_Para)).
      /// ------------------------------------------------------------------------
      /// </summary>
      private bool FmbSetBusparameterReq()
      {
        Global.OutputString("FMB_Set_Busparameter.req");

        C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();

        profiSdb.Layer     = (byte)LayerIdentifier.Fmb;
        profiSdb.Service   = (byte)FMBService.SetBusParameter;
        profiSdb.Primitive = (byte)ServicePrimitive.Req;

        _fmbSetBusparameter.LocAdd = Global.MasterAddr;

        if (Global.WriteServiceInterface(profiSdb, _fmbSetBusparameter))
        {
           return true;
        }

        return false;

      } /* FmbSetBusparameterReq */


      /****************************************************************************/

      /// <summary>
      /// Reads periodically the data from the pipe where it is saved periodically the received
      /// indication/confirmation buffer data
      /// </summary>
      private bool FmbReadServiceInterface()
      {
         bool bResult    = true;
         int  waitResult = -1;
         do
         {
            waitResult = WaitHandle.WaitAny(_waitFmbReadServiceHandles, 10, false);
            if (0 == waitResult)
            {
               return true;
            }
            else if (1 == waitResult)
            {
               Global.EwhExit.Set();
               return false;
            }
         } while (waitResult == WaitHandle.WaitTimeout);

         return bResult;

      } /* FmbReadServiceInterface */

      /****************************************************************************/

      /// <summary>
      /// /*------------------------------------------------------------------------*/
      /// /* FUNCTIONAL_DESCRIPTION                                                 */
      /// /*------------------------------------------------------------------------*/
      /// /* - coding and sending of FMB_EXIT.req service                           */
      /// /*                                                                        */
      /// /* FMB_EXIT terminates the PROFIBUS protocol software                     */
      /// /*------------------------------------------------------------------------*/
      /// </summary>
      private bool FmbExitReq()
      {
        Global.OutputString("FMB_Exit.req");

        C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();

        profiSdb.Layer     = (byte)LayerIdentifier.Fmb;
        profiSdb.Service   = (byte)FMBService.Exit;
        profiSdb.Primitive = (byte)ServicePrimitive.Req;

        return Global.WriteServiceInterface(profiSdb, null);

      } /* FmbExitReq */

      /****************************************************************************/

      private void FmbHandleEvent()
      {
        lock (_profiSdb)
        {
          FmbHandleEvent(_profiSdb);
        }
      }

      static ushort nPrevFmbEvent = 0;
      /// <summary>
      ///  ------------------------------------------------------------------------
      ///   FUNCTIONAL_DESCRIPTION
      ///  ------------------------------------------------------------------------
      ///   Call this function whenever you have read a message from the FMB
      ///   management device.
      ///   The function will display any message in readable text format.
      ///   This function is used to identify and decode any response of the
      ///   PROFIBUS software dedicated to the FMB management device.
      ///  ------------------------------------------------------------------------
      /// </summary>
      private void FmbHandleEvent(C_PROFI_SERVICE_DESCR profiSdb)
      {
         C_ERROR                     error          = new C_ERROR();
         C_FMB_FM2_EVENT_IND         fmbFm2EventInd = new C_FMB_FM2_EVENT_IND();
         C_EXCEPTION                 exception      = new C_EXCEPTION();
         C_FMB_GET_LIVE_LIST_CNF     getLiveListCon = new C_FMB_GET_LIVE_LIST_CNF();
         C_FMB_READ_BUSPARAMETER_CNF readBusParameterCon = new C_FMB_READ_BUSPARAMETER_CNF();
         try
         {
            if (profiSdb != null)
            {
               switch ((ServicePrimitive)profiSdb.Primitive)
               {
                  case ServicePrimitive.Con:
                  {
                     switch ((FMBService)profiSdb.Service)
                     {
                        case FMBService.SetConfiguration:
                        {
                          if (profiSdb.Result == (short)ServiceResult.Pos)
                          {
                             Global.OutputString("FMB_Set_Configuration.con [+]");
                          }
                          else
                          {
                             error = profiSdb.Data as C_ERROR;
                             if (error != null)
                             {
                                Global.OutputString(string.Format("FMB_Set_Configuration.con [-] class_code = {0:4X}, add_detail = {1}", error.ClassCode, error.AddDetail));
                             }
                           }
                        } /* case FMB_SET_CONFIGURATION */
                        break;

                        case FMBService.ValidateMaster:
                        {
                            if (profiSdb.Result == (short)ServiceResult.Pos)
                            {
                                Global.OutputString("FMB_Validate_Master.con [+]");
                            }
                            else
                            {
                                error = profiSdb.Data as C_ERROR;
                                if (error != null)
                                {
                                    Global.OutputString(string.Format("FMB_Validate_Master.con [-] class_code = {0:4X}, add_detail = {1}", error.ClassCode, error.AddDetail));
                                }
                            }
                        } /* case FMB_ValidateMaster */
                        break;

                        case FMBService.GetLiveList:
                        {
                            if (profiSdb.Result == (short)ServiceResult.Pos)
                            {
                                getLiveListCon = profiSdb.Data as C_FMB_GET_LIVE_LIST_CNF;
                                Global.OutputString("FMB_GetLiveList.con [+]");
                            }
                            else
                            {
                                error = profiSdb.Data as C_ERROR;
                                if (error != null)
                                {
                                    Global.OutputString(string.Format("FMB_GetLiveList.con [-] class_code = {0}, add_detail = {1}", error.ClassCode, error.AddDetail));
                                }
                            }
                        } /* case FMB_GetLiveList */
                        break;

                        case FMBService.Exit:
                        {
                           if (profiSdb.Result == (short)ServiceResult.Pos)
                           {
                              Global.OutputString("FMB_Exit.con [+]");
                           }
                           else
                           {
                              error = profiSdb.Data as C_ERROR;
                              if (error != null)
                              {
                                 Global.OutputString(string.Format("FMB_Exit.con [-] class_code = {0:4X}, add_detail = {1}", error.ClassCode, error.AddDetail));
                              }
                           }
                        } /* case FMB_EXIT */
                        break;

                        case FMBService.Reset:
                        {
                            if (profiSdb.Result == (short)ServiceResult.Pos)
                            {
                               Global.OutputString("FMB_Reset.con [+]");
                            }
                            else
                            {
                                error = profiSdb.Data as C_ERROR;
                                if (error != null)
                                {
                                   Global.OutputString(string.Format("FMB_Reset.con [-] class_code = {0:4X}, add_detail = {1}", error.ClassCode, error.AddDetail));
                                }
                            }
                        } /* case FMB_RESET */
                        break;

                        case FMBService.SetBusParameter:
                        {
                           if (profiSdb.Result == (short)ServiceResult.Pos)
                           {
                              Global.OutputString("FMB_Set_Busparameter.con [+]");
                           }
                           else
                           {
                              error = profiSdb.Data as C_ERROR;
                              if (error != null)
                              {
                                 Global.OutputString(string.Format("FMB_Set_Busparameter.con [-] class_code = {0:X4}, add_detail = {1}", error.ClassCode, error.AddDetail));
                              }
                           }

                        } /* case FMB_SET_BUSPARAMETER */
                        break;

                        case FMBService.ReadBusParameter:
                        {
                            if (profiSdb.Result == (short)ServiceResult.Pos)
                            {
                                readBusParameterCon = profiSdb.Data as C_FMB_READ_BUSPARAMETER_CNF;
                                Global.OutputString("FMB_Read_Busparameter.con [+]");
                            }
                            else
                            {
                                error = profiSdb.Data as C_ERROR;
                                if (error != null)
                                {
                                    Global.OutputString(string.Format("FMB_Set_Busparameter.con [-] class_code = {0}, add_detail = {1}", error.ClassCode, error.AddDetail));
                                }
                            }

                        } /* case FMB_Read_Busparameter */
                        break;

                         default:
                        {
                           Global.OutputString(string.Format("FMB: invalid service confirmation = {0:2X}", profiSdb.Service));
                        }
                        break;
                     } /* switch service */
                  }
                  break;  // ServicePrimitive.Con


                  case ServicePrimitive.Ind:
                  {
                     switch ((FMBService)profiSdb.Service)
                     {
                        case FMBService.Fm2Event:
                        {
                           fmbFm2EventInd = profiSdb.Data as C_FMB_FM2_EVENT_IND;
                           if (fmbFm2EventInd != null)
                           {
                              if (fmbFm2EventInd.Reason != nPrevFmbEvent)
                              {
                                 switch ((FM2Fault)fmbFm2EventInd.Reason)
                                 {
                                    case FM2Fault.Address:   Global.OutputString("FMB_Event.ind: duplicate address detected (another station uses same bus address)"); break;
                                    case FM2Fault.PHY:       Global.OutputString("FMB_Event.ind: fatal error in physical layer");                                      break;
                                    case FM2Fault.TTO:       Global.OutputString("FMB_Event.ind: timeout on bus detected, master entered token ring");                 break;
                                    case FM2Fault.FaultSyn:  Global.OutputString("FMB_Event.ind: no receiver synchronization (e.g. disturbed bus)");                   break;
                                    case FM2Fault.OutOfRing: Global.OutputString("FMB_Event.ind: station out of token ring (e.g. disturbed bus)");                     break;

                                    default:                 Global.OutputString(string.Format("FMB_Event.ind: {0:4X}", fmbFm2EventInd.Reason));                       break;
                                 } /* switch reason */

                                 nPrevFmbEvent = fmbFm2EventInd.Reason;

                              } /* if event changed */
                           }
                        } /* case FMB_FM2_EVENT */
                        break;

                        case FMBService.Exception:
                        {
                           exception = profiSdb.Data as C_EXCEPTION;
                           if (exception != null)
                           {
                              Global.OutputString(string.Format("FMB_Exception.ind: \r\ntask_id: {0:X2}, {1}\r\ntask_id: {2:X2}, {3}\r\ntask_id: {4:X4}, {5}\r\ntask_id: {6:X4}, {7}",
                                                  exception.TaskId, exception.TaskId, exception.Par1, exception.Par1, exception.Par2, exception.Par2, exception.Par3, exception.Par3));
                              Global.OutputString("Please report this exception to Softing PROFIBUS support (E-mail: support.profibus@softing.com)");
                           }
                        } /* case FMB_EXCEPTION */
                        break;

                        default:
                        {
                          Global.OutputString(string.Format("FMB: invalid service indication = {0:X2}", profiSdb.Service));
                        }
                        break;

                     } // switch service
                     break;
                  } // case ServicePrimitive.Ind


                  default:
                  {
                     Global.OutputString(string.Format("FMB: invalid primitive = 0x{0:X2}", profiSdb.Primitive));
                  }
                  break;
               } // switch primitive
            } // if nNumberOfBytesRead
            else
            {
               Global.OutputString("FMB: service event with zero length received");
            } // else nNumberOfBytesRead

         }
         catch (Exception exc)
         {
            Global.DisplayError(exc);
         }

         return;

      } // FmbHandleEvent


      private void FMBServiceConIndReceived(object sender, EventArgs<C_PROFI_SERVICE_DESCR> e)
      {
         _profiSdb = e.Object as C_PROFI_SERVICE_DESCR;

         Global.EwhFmbRead.Set();
      }

      #endregion
    }
}
