﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

using PapiWrapper;
using PapiWrapper.Common;
using PapiWrapper.Dp;
using System.Runtime.InteropServices;
using System.ComponentModel;

using HelperModule;

namespace GlobalModule
{
   #region Enums
   public enum OutputMode              // bit coded to select different streams
   {
      OmScreen = 1,                    // screen output enabled
      OmFile = 2,                      // file output enabled
      OmScreenFile = 3,                // screen and file output enabled
      OmStartLf = 4,                   // insert line feed before string
      OmSuppressId = 8,                // do not show thread identifier
      OmSuppressLf = 16                // no line feed at end of line
   };
   #endregion

   /// <summary>
   /// Generic event args class
   /// </summary>
   public class EventArgs<T> : EventArgs
   {
      #region fields

      private T m_object;

      #endregion

      #region constructors

      public EventArgs()
      {
      }

      public EventArgs(T obj)
      {
         m_object = obj;
      }

      #endregion

      #region properties

      /// <summary>
      /// Get/Set File field
      /// </summary>
      public T Object
      {
         get { return m_object; }
         set { m_object = value; }
      }

      #endregion
   }

   /// <summary>
   /// Global methods holder
   /// </summary>
   public class Global
   {
      #region Events
      public static event EventHandler<EventArgs<C_PROFI_SERVICE_DESCR>> DPServiceConIndReceived;
      public static event EventHandler<EventArgs<C_PROFI_SERVICE_DESCR>> FMBServiceConIndReceived;
      public static event EventHandler<EventArgs<C_PROFI_SERVICE_DESCR>> FMSServiceConIndReceived;
      public static event EventHandler<EventArgs<C_PROFI_SERVICE_DESCR>> FDLServiceConIndReceived;
      #endregion

      #region Constants
      private const string LogFileName = "DpDemo.log";
      private const int MaxNumberThreadItems = 12;
      private const int MaxNumberThreadChars = 12;

      public const ushort MaxPduDataBufferLength = 512;

      public const ushort MaxSlaveIoImageLength = 14336;    /* max. DPRAM size available */
      public const byte   MaxNumberDPSlaves     = DPConstants.MaxNumberSlaves;
      public const byte   MaxIoTelegramLength   = ((MaxSlaveIoImageLength / MaxNumberDPSlaves / 4) * 2);
      public const byte   MaxSlaveDiagLength    = DPConstants.MaxTelegramLen;
      public const ushort MaxSlaveDiagEntries   = (DPConstants.MaxNumberSlaves * 2 + DPConstants.MaxNumberSlaves / 4);
      public const ushort MaxBusParaLen         = DPConstants.MinBusParaLen;
      public const ushort MaxSlaveParaLen       = 512;

      #endregion

      #region Fields
      private static object _LoggerOut = new object();
      private static object _readService = new object();
      private static CPapi  _papi;
      private static Global _instance;

      private static Dictionary<int, string> _threadIdStringTable;

      private static IntPtr _hLogFile = new IntPtr(); // handle for log file

      private C_PROFI_SERVICE_DESCR  _pSdb;
              byte[]                 _profiReadServiceData;

      private static EventWaitHandle _ewhDpRead;
      private static EventWaitHandle _ewhFmbRead;
      private static EventWaitHandle _ewhRead;
      private static EventWaitHandle _ewhExit;

      #endregion

      #region Constructor
      private Global()
      {
         _pSdb                 = new C_PROFI_SERVICE_DESCR();
         _profiReadServiceData = new byte[MaxPduDataBufferLength];

         _threadIdStringTable  = new Dictionary<int, string>();

         _ewhDpRead            = new EventWaitHandle(false, EventResetMode.AutoReset);
         _ewhFmbRead           = new EventWaitHandle(false, EventResetMode.AutoReset);
         _ewhRead              = new EventWaitHandle(false, EventResetMode.AutoReset);
         _ewhExit              = new EventWaitHandle(false, EventResetMode.AutoReset);
      }
      #endregion

      #region Properties
      public static Global Instance
      {
         get
         {
            if (_instance == null)
            {
               _instance = new Global();
            }
            return _instance;
         }
      }

      public static EventWaitHandle EwhFmbRead
      {
         get { return _ewhFmbRead; }
      }
      public static EventWaitHandle EwhDpRead
      {
         get { return _ewhDpRead; }
      }
      public static EventWaitHandle EwhRead
      {
         get { return _ewhRead; }
      }
      public static EventWaitHandle EwhExit
      {
         get { return _ewhExit; }
      }

      public static byte MasterAddr  { get; set; }
      public static byte BoardNumber { get; set; }
      public static byte Baudrate    { get; set; }

      public static bool CanReadServiceInterface { get; set; }

      #endregion


      #region PAPI methods

      /// <summary>
      /// open logger and pipes
      /// </summary>
      public static bool OpenPapiAndLogger()
      {
         _hLogFile = Helper.CreateFile(LogFileName,
                                       EFileAccess.GenericWrite,
                                       EFileShare.Read,
                                       IntPtr.Zero,
                                       ECreationDisposition.CreateAlways,
                                       EFileAttributes.Normal,
                                       IntPtr.Zero);

         if (_hLogFile == IntPtr.Zero)
         {
            OutputString(string.Format("Create log file %s failed.", LogFileName));
            DisplayLastError();
            return false;
         }

         Global._papi = new CPapi();
         if (_papi == null)
         {
            OutputString("Papi interface not initialized.");
            return false;
         }

         return true;
      }

      /// <summary>
      /// ReadServiceInterface
      /// - read service indication/confirmation from the PAPI
      /// </summary>
      public static bool ReadServiceInterface()
      {
         try
         {
            lock (_readService)
            {
               ServiceConfirmationIndication confInd = _papi.ReceiveServiceConfirmIndication(Global.BoardNumber);
               if (confInd != null)
               {
                  if (confInd.IsConInd)
                  {
                     if (confInd.ServiceDescriptor.Layer == (byte)LayerIdentifier.FmbUsr)
                     {
                        if (FMBServiceConIndReceived != null)
                        {
                            FMBServiceConIndReceived(null, new EventArgs<C_PROFI_SERVICE_DESCR>(confInd.ServiceDescriptor));
                        }
                     }
                     else if (confInd.ServiceDescriptor.Layer == (byte)LayerIdentifier.DpUsr)
                     {
                         if (DPServiceConIndReceived != null)
                        {
                            Global.CanReadServiceInterface = false;
                            DPServiceConIndReceived(null, new EventArgs<C_PROFI_SERVICE_DESCR>(confInd.ServiceDescriptor));
                        }
                     }
                     else if (confInd.ServiceDescriptor.Layer == (byte)LayerIdentifier.FdlIfUsr)
                     {
                        if (FDLServiceConIndReceived != null)
                        {
                          FDLServiceConIndReceived(null, new EventArgs<C_PROFI_SERVICE_DESCR>(confInd.ServiceDescriptor));
                        }
                     }
                     else if (confInd.ServiceDescriptor.Layer == (byte)LayerIdentifier.FmsUsr)
                     {
                        if (FMSServiceConIndReceived != null)
                        {
                          FMSServiceConIndReceived(null, new EventArgs<C_PROFI_SERVICE_DESCR>(confInd.ServiceDescriptor));
                        }
                     }
                     else
                     {
                       ;
                     }
                     Thread.Sleep(1);
                  }
                  else
                  {
                       ;
                  }
               }
               else
               {
                   ;
               }
			}
         }
         catch (CPapiException papiExc)
         {
            DisplayPapiError(papiExc);
            return false;
         }
         return true;
      }


      /// <summary>
      /// WriteServiceInterface
      /// - write (send) service request/response to the PAPI
      /// </summary>
      public static bool WriteServiceInterface(C_PROFI_SERVICE_DESCR profiServiceDesc, object sendData)
      {
         try
         {
            _papi.SendServiceRequestResponse(Global.BoardNumber, profiServiceDesc, sendData);
         }
         catch (CPapiException papiExc)
         {
            DisplayPapiError(papiExc);
            return false;
         }
         catch (Exception)
         {
            return false;
         }

         return true;
      }


      /// <summary>
      /// GetData
      /// - read data from CMI data area
      /// </summary>
      public static bool GetData(out byte[] pData, uint offset, uint dataSize, out uint pBytesRead)
      {
         pBytesRead = 0;
         pData = null;
         try
         {
             UInt16 locDataSize = (UInt16) dataSize;
            _papi.GetData(Global.BoardNumber, (byte) DataImage.IdDPSlaveIo, (ushort) offset, ref  locDataSize, ref pData);
            pBytesRead = locDataSize;
         }
         catch (CPapiException papiExc)
         {
            DisplayPapiError(papiExc);
            return false;
         }
         return true;
      }

      /// <summary>
      /// Write data to CMI data area
      /// </summary>
      public static bool SetData(byte[] pData, uint offset, uint dataSize, out uint pBytesRead)
      {
         pBytesRead = 0;
         try
         {
            _papi.SetData(Global.BoardNumber, (byte) DataImage.IdDPSlaveIo, (ushort) offset, (ushort)dataSize, pData);
         }
         catch (CPapiException papiExc)
         {
            DisplayPapiError(papiExc);
            return false;
         }
         return true;
      }

      /// <summary>
      /// Get serial number
      /// </summary>
      private static void GetSerialDeviceNumber()
      {
         OutputString("Get serial number.");

         try
         {
            uint serialDeviceNumber = _papi.GetSerialDeviceNumber(Global.BoardNumber);
            OutputString("Serial device number: " + serialDeviceNumber);
         }
         catch (CPapiException papiExc)
         {
            DisplayPapiError(papiExc);
         }
      }

      /// <summary>
      /// Get PAPI version string
      /// </summary>
      public static string GetPapiVersion()
      {
         string papiVersion     = string.Empty;
         string firmwareVersion = string.Empty;

         try
         {
            _papi.GetVersions(Global.BoardNumber, ref papiVersion, ref firmwareVersion);
         }
         catch (CPapiException papiExc)
         {
            DisplayPapiError(papiExc);
         }
         return papiVersion;
      }

      /// <summary>
      /// Get Firmware version string
      /// </summary>
      public static string GetFirmwareVersion()
      {
          string papiVersion = string.Empty;
          string firmwareVersion = string.Empty;

          try
          {
              _papi.GetVersions(Global.BoardNumber, ref papiVersion, ref firmwareVersion);
          }
          catch (CPapiException papiExc)
          {
              DisplayPapiError(papiExc);
          }
          return firmwareVersion;
      }


      /// <summary>
      /// Call init papi in order to use PAPI facilities
      /// </summary>
      public static bool PapiInit(byte boardNumber, uint readTimeout, uint writeTimeout)
      {
         try
         {
            _papi.Init(boardNumber, readTimeout, readTimeout);
         }
         catch (CPapiException papiExc)
         {
            DisplayPapiError(papiExc);
            DisplayLastError();
            return false;
         }
         return true;
      }

      /// <summary>
      /// Call the PAPI to end.
      /// </summary>
      public static void PapiEnd(byte boardNumber)
      {
         try
         {
            _papi.End(boardNumber);
            OutputString("Papi ended");
         }
         catch (CPapiException papiExc)
         {
            DisplayPapiError(papiExc);
         }
      }

        /// <summary>
        /// /*------------------------------------------------------------------------*/
        /// /* FUNCTIONAL_DESCRIPTION                                                 */
        /// /*------------------------------------------------------------------------*/
        /// /* This function imports a Softing configuration file. These files are    */
        /// /* created by the Softing PROFIBUS DP Configurator. Also the MS DOS       */
        /// /* application CNF2BIN is able to convert ASCII configuration files (CNF) */
        /// /* into the binary configuration format.                                  */
        /// /*                                                                        */
        /// /* The routine will load all descriptions (bus and slave parameter sets)  */
        /// /* found in the binary file into the respective working structures of     */
        /// /* this application. This is done by means of the functions:              */
        /// /* - FmbSetBusParaSet,                                                    */
        /// /* - DpSetBusParaSet,                                                     */
        /// /* - DpSetSlaveParaSet.                                                   */
        /// /*                                                                        */
        /// /* Note, that there exist two versions of the binary configuration        */
        /// /* format: an aligned and a not aligned version. Both formats can be      */
        /// /* handled by this routine.                                               */
        /// /*                                                                        */
        /// /* The PAPI wrapper includes the file reading functionality.              */
        /// /*------------------------------------------------------------------------*/
        /// </summary>
        /// <param name="binFile"></param>
        /// <param name="slaveParaSets"></param>
        /// <returns></returns>
        public static bool ImportBinaryDpConfigurationFile(string binFile,
                                                           out C_DP_BUS_PARA_SET busParaSet,
                                                           Dictionary<byte, C_DP_SLAVE_PARA_SET> slaveParaSets
                                                          )
        {
           busParaSet = null;
           bool retVal = _papi.ImportBinaryDpConfigurationFile(binFile, ref busParaSet, slaveParaSets);
           return (retVal);
        }

        //***************************************************************************

        /// <summary>
        /// /*------------------------------------------------------------------------*/
        /// /* FUNCTIONAL_DESCRIPTION                                                 */
        /// /*------------------------------------------------------------------------*/
        /// /* This function checks the given slave parameter set whether it is valid */
        /// /* or not.                                                                */
        /// /*                                                                        */
        /// /* Note: The PAPI wrapper includes the checking functionality.            */
        /// /*------------------------------------------------------------------------*/
        /// </summary>
        public static bool DpSetSlaveParaSet(byte nSlaveAdd, C_DP_SLAVE_PARA_SET slaveParaSet)
        {
           return _papi.DpSetSlaveParaSet(nSlaveAdd, slaveParaSet);
        }

      #endregion

      #region Methods

      /// <summary>
      /// Close logger and wait handlers
      /// </summary>
      public static void CloseLoggerAndHandlers()
      {
         _ewhDpRead.Close();
         _ewhFmbRead.Close();
         _ewhRead.Close();
         _ewhExit.Close();

         if (_hLogFile != IntPtr.Zero)
         {
            Helper.CloseHandle(_hLogFile);
            _hLogFile = IntPtr.Zero;
         }
      }

      /// <summary>
      /// Add thread prefix for output string sormat
      /// </summary>
      public static void OutputSetThreadIdString(string threadIdString)
      {
         if (!_threadIdStringTable.ContainsKey(Thread.CurrentThread.ManagedThreadId))
         {
            _threadIdStringTable.Add(Thread.CurrentThread.ManagedThreadId,string.Format("[{0:5}] ", threadIdString));
         }
      }

      /// <summary>
      /// Shows a message on the screen or/and into the log file
      /// </summary>
      public static void OutputString(OutputMode outputMode, string msg)
      {
         lock (_LoggerOut)
         {
            if (_hLogFile == IntPtr.Zero)
            {
              return;
            }

            uint nNumberOfBytesWritten;

            // look for assigned thread ID string ------------------------------------------------

            string threadString = string.Empty;
            if (_threadIdStringTable.ContainsKey(Thread.CurrentThread.ManagedThreadId))
            {
               threadString = _threadIdStringTable[Thread.CurrentThread.ManagedThreadId];

               if ((outputMode & OutputMode.OmStartLf) == OutputMode.OmStartLf)
               {
                  Console.WriteLine("");
               }

               string tempStringEol = "\r\n";
               if ((outputMode & OutputMode.OmSuppressLf) == OutputMode.OmSuppressLf)
               {
                  tempStringEol = string.Empty;
               }

               if ((outputMode & OutputMode.OmScreen) == OutputMode.OmScreen)
               {
                  if (tempStringEol.Equals(string.Empty))
                  {
                     Console.Write("{0}{1}", threadString, msg);
                  }
                  else
                  {
                     Console.WriteLine("{0}{1}", threadString, msg);
                  }
               }

               if (((outputMode & OutputMode.OmFile      ) == OutputMode.OmFile      ) ||
                   ((outputMode & OutputMode.OmScreenFile) == OutputMode.OmScreenFile) &&
                   ( _hLogFile                             != IntPtr.Zero            )    )
               {
                  if ((!Helper.WriteFile(_hLogFile, threadString,  (uint)threadString.Length,  out nNumberOfBytesWritten, IntPtr.Zero)) ||
                      (!Helper.WriteFile(_hLogFile, msg,           (uint)msg.Length,           out nNumberOfBytesWritten, IntPtr.Zero)) ||
                      (!Helper.WriteFile(_hLogFile, tempStringEol, (uint)tempStringEol.Length, out nNumberOfBytesWritten, IntPtr.Zero))    )
                  {
                     if (!Helper.CloseHandle(_hLogFile))
                     {
                        _hLogFile = IntPtr.Zero;
                     }

                     Console.WriteLine(string.Format("Output to log file {0} failed.", LogFileName));
                     Console.WriteLine(new Win32Exception(Marshal.GetLastWin32Error()).Message);

                  } // if not WriteFile
               } // if LogFile
            }
         }
      }


      /// <summary>
      /// OutputString
      /// - shows a message on the screen and into the log file by default
      /// </summary>
      public static void OutputString(string msg)
      {
         OutputString(OutputMode.OmScreenFile, msg);
      }


      /// <summary>
      /// DisplayLastError
      /// - retrieves the error code of the last function which failed
      /// - displays readable strings if an error message string is available
      public static void DisplayLastError()
      {
         int    errorCode    = Marshal.GetLastWin32Error();
         string errorMessage = new Win32Exception(errorCode).Message;

         if (errorMessage.StartsWith("Unknown error"))
         {
            switch ((InterfaceErrorCode)errorCode)
            {
               case InterfaceErrorCode.OK | InterfaceErrorCode.ErrorFlag: break;
               case InterfaceErrorCode.FatalError | InterfaceErrorCode.ErrorFlag:                   OutputString("fatal error on board detected"); break;
               case InterfaceErrorCode.InitInvalidParameter | InterfaceErrorCode.ErrorFlag:         OutputString("invalid 'parameter' during initialization"); break;
               case InterfaceErrorCode.NoControllerResponse | InterfaceErrorCode.ErrorFlag:         OutputString("controller/firmware does not respond"); break;
               case InterfaceErrorCode.InvalidControllerTypeVersion | InterfaceErrorCode.ErrorFlag: OutputString("invalid controller type or firmware version"); break;
               case InterfaceErrorCode.InvalidLayer | InterfaceErrorCode.ErrorFlag:                 OutputString("invalid 'layer' in Service Description Block (SDB)"); break;
               case InterfaceErrorCode.InvalidService | InterfaceErrorCode.ErrorFlag:               OutputString("invalid 'service' in Service Description Block (SDB)"); break;
               case InterfaceErrorCode.InvalidPrimitive | InterfaceErrorCode.ErrorFlag:             OutputString("invalid 'primitive' in Service Description Block (SDB)"); break;
               case InterfaceErrorCode.InvalidDataSize | InterfaceErrorCode.ErrorFlag:              OutputString("invalid data size in message detected (e.g. check 'length' parameters)"); break;
               case InterfaceErrorCode.InvalidCommRef | InterfaceErrorCode.ErrorFlag:               OutputString("invalid 'comm_ref' in Service Description Block (SDB)"); break;
               case InterfaceErrorCode.InvalidCmiCall | InterfaceErrorCode.ErrorFlag:               OutputString("invalid interface call (e.g. invalid SDB memory; check interface conventions)"); break;
               case InterfaceErrorCode.CmiError | InterfaceErrorCode.ErrorFlag:                     OutputString("internal error in Common Memory Interface (CMI)"); break;
               case InterfaceErrorCode.ResourceUnavailable | InterfaceErrorCode.ErrorFlag:          OutputString("no resource available (e.g. check memory configuration or try again later)"); break;
               case InterfaceErrorCode.NoParallelServices | InterfaceErrorCode.ErrorFlag:           OutputString("no parallel service execution allowed (e.g. a service is already running)"); break;
               case InterfaceErrorCode.ServiceConstraintConflict | InterfaceErrorCode.ErrorFlag:    OutputString("service temporary not executable (i.e. try again later or serve pending tasks first)"); break;
               case InterfaceErrorCode.ServiceNotSupported | InterfaceErrorCode.ErrorFlag:          OutputString("service not supported"); break;
               case InterfaceErrorCode.ServiceNotExecutable | InterfaceErrorCode.ErrorFlag:         OutputString("service not executable (reason not specified)"); break;
               case InterfaceErrorCode.InvalidVersion | InterfaceErrorCode.ErrorFlag:               OutputString("invalid version (interface/firmware version conflict)"); break;

               case InterfaceErrorCode.NoCntrlPresent | InterfaceErrorCode.ErrorFlag:               OutputString("board/controller not found"); break;
               case InterfaceErrorCode.InvalidParameter | InterfaceErrorCode.ErrorFlag:             OutputString("invalid 'parameter' in request or response"); break;
               case InterfaceErrorCode.InitFailed | InterfaceErrorCode.ErrorFlag:                   OutputString("initialization of API failed"); break;
               case InterfaceErrorCode.ExitFailed | InterfaceErrorCode.ErrorFlag:                   OutputString("shutdown of API failed"); break;
               case InterfaceErrorCode.PapiNotInitialized | InterfaceErrorCode.ErrorFlag:           OutputString("API not initialized"); break;
               case InterfaceErrorCode.NoDeviceConnection | InterfaceErrorCode.ErrorFlag:           OutputString("connection to PROFIBUS device is lost"); break;

               case InterfaceErrorCode.SlaveDiagData | InterfaceErrorCode.ErrorFlag:                OutputString("new DP slave diagnostic data available at DP management device"); break;
               case InterfaceErrorCode.SlaveError | InterfaceErrorCode.ErrorFlag:                   OutputString("DP slave communication error detected"); break;
               case InterfaceErrorCode.InvalidDpState | InterfaceErrorCode.ErrorFlag:               OutputString("DP is not in state clear/operate"); break;
               case InterfaceErrorCode.ReadingRegistry | InterfaceErrorCode.ErrorFlag:              OutputString("error reading registry"); break;
               case InterfaceErrorCode.OsError | InterfaceErrorCode.ErrorFlag:                      OutputString("operating system error"); break;

               default:
               {
                  OutputString(string.Format("last error code = 0x{0:X8}", errorCode));
                  OutputString(errorMessage);
               }
               break;
            } // switch ErrorCode
         } // else MessageLength
         else
         {
            OutputString(errorMessage);
         }

         return;

      } // DisplayLastError


      /// <summary>
      /// Display papi wrapper error
      /// </summary>
      public static void DisplayPapiError(CPapiException papiExc)
      {
        OutputString(OutputMode.OmFile, string.Format("Papi Error: {0} {1}", papiExc.Error, papiExc.StackTrace));
      }


      /// <summary>
      /// display the exception error
      /// </summary>
      public static void DisplayError(Exception exc)
      {
         OutputString(string.Format("Error: {0} {1}", exc.Message, exc.StackTrace));
      }


      /****************************************************************************/
      // MACROS related
      /****************************************************************************/

      private static InterfaceErrorCode PROFI_ERROR(InterfaceErrorCode errorCode)
      {
         return (errorCode | InterfaceErrorCode.ErrorFlag);
      }

      public static ushort Swap16(ushort x)
      {
         return (ushort)((x >> 8) | (x << 8));
      }

      public static UInt32 Swap32(UInt32 x)
      {
         return ((UInt32)((((UInt32)(x)) >> 24) |
                         ((((UInt32)(x)) & 0x00FF0000u) >> 8) |
                                   ((((UInt32)(x)) & 0x0000FF00u) << 8) |
                                    (((UInt32)(x)) << 24)));
      }
      #endregion

 
  }
}
