using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Threading;
using PapiWrapper.Common;
using PapiWrapper.Dp;

using GlobalModule;
using HelperModule;

namespace DpModule
{
   /// <summary>
   /// This module contains all functionality which is relevant to the DPV1 protocol.
   /// The module covers DPV1 master initialization and handling of the DPV1 master 
   /// operational states (STOP)
   ///
   /// Responsible for all theses features is the DpManagementThread.
   /// </summary>
   public class DpLayer
   {
      #region Fields

      #region DP data area
      private static DPOPMode              _nCurrentOpMode   = DPOPMode.Offline;
      private static DPOPMode              _nRequestedOpMode = DPOPMode.Offline;
      #endregion

      private static DpLayer               _instance;

      private static WaitHandle[]          _waitDpReadServiceHandles;

      private        C_PROFI_SERVICE_DESCR _profiSdb = new C_PROFI_SERVICE_DESCR();
      #endregion

      #region Constructor
      private DpLayer()
      {
         _nCurrentOpMode   = DPOPMode.Offline;
         _nRequestedOpMode = DPOPMode.Offline;

         _waitDpReadServiceHandles = new WaitHandle[2] { Global.EwhDpRead, Global.EwhExit };

         Global.DPServiceConIndReceived += new EventHandler<EventArgs<C_PROFI_SERVICE_DESCR>>(DPServiceConIndReceived);
      }
      #endregion

      #region Properties
      public static DpLayer Instance
      {
         get
         {
            if (_instance == null)
            {
               _instance = new DpLayer();
            }
            return _instance;
         }
      }
      #endregion


      #region Public methods

      /// <summary>
      ///  DpManagementThread
      ///  This thread is responsible for initializing the DPV1 protocol component
      ///  and handles all DPV1 specific events (e.g. receive DPV1 specific service
      ///  indications and confirmations.
      ///
      ///  This function initializes the DPV1 master (DP_Init_Master service).
      ///  After that the master will be be brought in the operation mode STOP
      ///  and than it waits for DPV1 specific events
      /// </summary>
      public void DpManagementThread()
      {
         byte[] profiData = new byte[Global.MaxPduDataBufferLength];

         Global.OutputSetThreadIdString("   DP");
         Global.OutputString("DpManagementThread started.");

         // increase thread priority for dp layer thread
         Thread.CurrentThread.Priority++;

         // initialize DP Master -------------------------------------------------------
         if (!DpInitMasterReq())
         {
            Global.DisplayLastError();
            return;
         }
         if (!DpReadServiceInterface())
         {
            Global.DisplayLastError();
            return;
         }
         DpHandleEvent();


         // enter operation mode STOP to go Online ------------------------------------
         if (!DpActParamLocReq(DPOPMode.Stop))
         {
            Global.DisplayLastError();
            return;
         }
         if (!DpReadServiceInterface())
         {
            Global.DisplayLastError();
            return;
         }
         DpHandleEvent();

         // start event handling for DP -----------------------------------------------
         bool bResult;
         do
         {
            bResult = DpReadServiceInterface();
            if (bResult)
            {
               DpHandleEvent();
            }
         } while (bResult);                   // terminates in case of errors only

         Global.DisplayLastError();
         Global.OutputString("DpManagementThread terminated");
      }




      /****************************************************************************/

       /// <summary>
      /// DpC2InitiateReq
      /// This service opens a master acyclic class 2 connection to a DP-Slave
      /// </summary>
      public bool DpC2InitiateReq(byte remAdd, UInt16 sendTimeout)
      {
         Global.OutputString(string.Format("DP_C2_Initiate.req Slave Address = {0}  Send Timeout = {1}", remAdd, sendTimeout));
                                 
         C_PROFI_SERVICE_DESCR profiSdb      = new C_PROFI_SERVICE_DESCR();
         C_DP_INITIATE_REQ     initiateReq   = new C_DP_INITIATE_REQ();

         profiSdb.Layer     = (byte)LayerIdentifier.Dp;
         profiSdb.Service   = (byte)DPService.DpInitiate;
         profiSdb.Primitive = (byte)ServicePrimitive.Req;

         initiateReq.RemAdd                      = remAdd;
         initiateReq.SendTimeout                 = sendTimeout;
         initiateReq.FeaturesSupported[0]        = 1;
         initiateReq.FeaturesSupported[1]        = 0;
         initiateReq.ProfileFeaturesSupported[0] = 0;
         initiateReq.ProfileFeaturesSupported[1] = 0;
         initiateReq.ProfileIdentNumber          = 0;

         return Global.WriteServiceInterface(profiSdb, initiateReq);

      } // DpC2InitiateReq



      /****************************************************************************/

      /// <summary>
      /// DpC2ReadReq
      /// This service reads acyclic data from a slave
      /// </summary>

      public bool DpC2ReadReq(byte remAdd, byte slotNumber, byte index, byte length)
      {
         Global.OutputString(string.Format("DP_C2_Read.req: Slave Address = {0}  Slotnumber = {1} Index = {2} Length = {3}", remAdd, slotNumber, index, length));

         C_PROFI_SERVICE_DESCR profiSdb   = new C_PROFI_SERVICE_DESCR();
         C_DP_READ_REQ         readReq    = new C_DP_READ_REQ();

         profiSdb.Layer     = (byte)LayerIdentifier.Dp;
         profiSdb.Service   = (byte)DPService.DpRead;
         profiSdb.Primitive = (byte)ServicePrimitive.Req;

         readReq.RemAdd      = remAdd;
         readReq.SlotNumber  = slotNumber;
         readReq.Index       = index;
         readReq.Length      = length;

         return Global.WriteServiceInterface(profiSdb, readReq);

      } // DpC2ReadReq


     /****************************************************************************/

      /// <summary>
      /// DpC2WriteReq
      /// This service writes acyclic data from a slave
      /// </summary>
      public bool DpC2WriteReq(byte remAdd, byte slotNumber, byte index, byte length)
      {
          Global.OutputString(string.Format("DP_C2_Write.req: Slave Address = {0}  Slotnumber = {1} Index = {2} Length = {3}", remAdd, slotNumber, index, length));

         C_PROFI_SERVICE_DESCR profiSdb      = new C_PROFI_SERVICE_DESCR();
         C_DP_WRITE_REQ        writeReq      = new C_DP_WRITE_REQ();
                               writeReq.Data = new byte[length];
    
         profiSdb.Layer     = (byte)LayerIdentifier.Dp;
         profiSdb.Service   = (byte)DPService.DpWrite;
         profiSdb.Primitive = (byte)ServicePrimitive.Req;

         writeReq.RemAdd      = remAdd;
         writeReq.SlotNumber  = slotNumber;
         writeReq.Index       = index;
         writeReq.Length      = length;

         for (int i = 0; i < length; i++)
         {
             writeReq.Data[i] = (byte) i;
         }

         return Global.WriteServiceInterface(profiSdb, writeReq);

      } // DpC2WriteReq


      /****************************************************************************/

      /// <summary>
      /// DpC2AbortReq
      /// This service closes a open master acyclic class 2 connection.
      /// </summary>
      public bool DpC2AbortReq()
      {
         Global.OutputString("DP_C2_Abort.req");

         C_PROFI_SERVICE_DESCR    profiSdb  = new C_PROFI_SERVICE_DESCR();
         C_DP_ABORT_REQ           abortReq  = new C_DP_ABORT_REQ();

         profiSdb.Layer     = (byte)LayerIdentifier.Dp;
         profiSdb.Service   = (byte)DPService.DpAbort;
         profiSdb.Primitive = (byte)ServicePrimitive.Req;

         abortReq.Subnet  = (byte) DPAbortSubnet.No;
         abortReq.Reason  = 0;

         return Global.WriteServiceInterface(profiSdb, abortReq);

      } // DpC2AbortReq


/****************************************************************************/


      /// <summary>
      /// TerminateDpLayer
      /// - clean up module
      /// </summary>
      public void TerminateDpLayer()
      {

      } // TerminateDpLayer

      #endregion

      #region Private methods

      /// <summary>
      /// DpInitMasterReq
      /// This service brings the DPV1 master in the operation mode OFFLINE.
      /// Also the operation characteristics of the master will be defined to
      /// master class 2 role.
      /// </summary>
      private bool DpInitMasterReq()
      {
         Global.OutputString("DP_Init_Master.req");

         C_PROFI_SERVICE_DESCR profiSdb      = new C_PROFI_SERVICE_DESCR();
         C_DP_INIT_MASTER_REQ  initMasterReq = new C_DP_INIT_MASTER_REQ();

         profiSdb.Layer     = (byte)LayerIdentifier.Dp;
         profiSdb.Service   = (byte)DPService.InitMaster;
         profiSdb.Primitive = (byte)ServicePrimitive.Req;

         initMasterReq.MasterDefaultAddress = Global.MasterAddr;
         initMasterReq.MasterClass2         = Constants.PbTrue;
         initMasterReq.MasterClass2Name     = Helper.ConvertStringToSByte("Softing DP-Master V5.45 ", 32);
         initMasterReq.LowestSlaveAddress   = 0;
         initMasterReq.SlaveIOAddressMode   = (byte)DPDpramAAM.IOBlocks;
         initMasterReq.ClearOutputs         = Constants.PbFalse;
         initMasterReq.AutoRemoteServices   = (byte)MasterDPAuto.RemoteServices;
         initMasterReq.CyclicDataTransfer   = Constants.PbTrue;

         return Global.WriteServiceInterface(profiSdb, initMasterReq);

      } // DpV1InitMasterReq

      /****************************************************************************/

      /// <summary>
      /// DpActParamLocReq
      /// This function is used to bring the DPV1 master into a new operation mode
      /// (STOP, CLEAR, OPERATE). Please consider that it is allowed to switch
      /// between neighbouring operation modes only (e.g. STOP->CLEAR, OPERATE->
      /// CLEAR, and not STOP->OPERATE).
      /// The operation mode is controlled via the variables nCurrentOpMode and,
      /// nRequestedOpMode.
      ///
      /// DP operation modes:
      /// - OFFLINE: master inactive, local services allowed only (init phase)
      /// - STOP   : master active on PROFIBUS, no data exchange with DP slaves
      ///            yet, all services allowed except reconfiguration of DP
      ///            slaves (i.e. DP_Download_Loc)
      /// - CLEAR  : master active on PROFIBUS, data exchange with all slaves,
      ///            output data is set to zero or length null (FAIL SAFE mode)
      /// - OPERATE: master active, normal data exchage with all slaves
      ///       
      /// </summary>
      private bool DpActParamLocReq(DPOPMode nOpMode)
      {
         switch (nOpMode)
         {
            case DPOPMode.Offline: Global.OutputString("DP_Act_Param_Loc.req (OFFLINE)"); break;
            case DPOPMode.Stop:    Global.OutputString("DP_Act_Param_Loc.req (STOP)");    break;
            case DPOPMode.Clear:   Global.OutputString("DP_Act_Param_Loc.req (CLEAR)");   break;
            case DPOPMode.Operate: Global.OutputString("DP_Act_Param_Loc.req (OPERATE)"); break;
            default:               Global.OutputString(string.Format("DP_Act_Param_Loc.req (invalid operation mode = 0x%{0:X2})", nOpMode));
                                   return false;
         }

         C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();
         C_DP_ACT_PARAM_REQ    actParam = new C_DP_ACT_PARAM_REQ();

         profiSdb.Layer     = (byte)LayerIdentifier.Dp;
         profiSdb.Service   = (byte)DPService.ActParamLoc;
         profiSdb.Primitive = (byte)ServicePrimitive.Req;

         actParam.AreaCode  = (byte)DPAreaCodes.SetMode;
         actParam.Activate  = (byte)nOpMode;

         if (Global.WriteServiceInterface(profiSdb, actParam))
         {
            _nRequestedOpMode = nOpMode;
            return true;
         }

         return false;

      } // DpActParamLocReq


      /****************************************************************************/

      /// <summary>
      ///  DpExitMasterReq
      ///  Use this service to bring the DP master in the operation mode OFFLINE
      ///  just before master configuration. That means you have to call service
      ///  DP_Init_Master before being able to execute local services again.
      ///  This service can be called regardless of the current operation mode
      ///  of the DP Master (see DpActParamLocReq and DpInitMasterReq for more
      ///  information).
      ///
      ///  This service is the reset service for the DP protocol component. It
      ///  does not influence any other components like FMB or FMS.
      /// </summary>
      private bool DpExitMasterReq()
      {
         Global.OutputString("DP_Exit_Master.req");

         C_PROFI_SERVICE_DESCR profiSdb = new C_PROFI_SERVICE_DESCR();

         profiSdb.Layer     = (byte)LayerIdentifier.Dp;
         profiSdb.Service   = (byte)DPService.ExitMaster;
         profiSdb.Primitive = (byte)ServicePrimitive.Req;

         if (Global.WriteServiceInterface(profiSdb, null))
         {
            _nRequestedOpMode = DPOPMode.Offline;
            _nCurrentOpMode   = DPOPMode.Offline;
            return true;
         }

         return false;

      } // DpExitMasterReq



      /// <summary>
      ///  DpC2InitiateCon
      /// </summary>
      private void DpC2InitiateCon(C_DP_INITIATE_CON initiateCon)
      {
         Global.OutputString(string.Format(" slave address        =  {0}        ", initiateCon.RemAdd));
         Global.OutputString(string.Format(" features supported [0]={0} [1]={1} ", initiateCon.FeaturesSupported[0],initiateCon.FeaturesSupported[1]));
         Global.OutputString(string.Format(" max length data unit = {0}         ", initiateCon.MaxLenDataUnit));
 
         return;
      } // DpC2InitiateCon



      /// <summary>
      ///  DpC2ReadCon
      /// </summary>
      private void DpC2ReadCon(C_DP_READ_CON readCon)
      {
         Global.OutputString(string.Format(" Status         = {0}        ", readCon.Status));
         Global.OutputString(string.Format(" Slave address  = {0}        ", readCon.RemAdd));
         Global.OutputString(string.Format(" Slot           = {0}        ", readCon.SlotNumber));
         Global.OutputString(string.Format(" Index          = {0}        ", readCon.Index));
         Global.OutputString(string.Format(" Length         = {0}        ", readCon.Length));
         Global.OutputString(string.Format(" Data           = {0} {1} {2} {3}  ", readCon.Data[0], readCon.Data[1], readCon.Data[2], readCon.Data[3]));

         return;
      } // DpC2ReadCon

      /// <summary>
      ///  DpC2WriteCon
      /// </summary>
      private void DpC2WriteCon(C_DP_WRITE_CON writeCon)
      {
         Global.OutputString(string.Format(" Status         = {0}        ", writeCon.Status));
         Global.OutputString(string.Format(" Slave address  = {0}        ", writeCon.RemAdd));
         Global.OutputString(string.Format(" Slot           = {0}        ", writeCon.SlotNumber));
         Global.OutputString(string.Format(" Index          = {0}        ", writeCon.Index));
         Global.OutputString(string.Format(" Length         = {0}        ", writeCon.Length));

         return;
      } // DpC2WriteCon


      /// <summary>
      ///  DpC2AbortInd
      /// </summary>
      private void DpC2AbortInd(C_DP_ABORT_IND abortInd)
      {
         Global.OutputString(string.Format(" locally generated = {0}  ", abortInd.LocallyGenerated));
         Global.OutputString(string.Format(" reason            = {0}  ", abortInd.Reason));
         Global.OutputString(string.Format(" subnet            = {0}  ", abortInd.Subnet));
         Global.OutputString(string.Format(" additional detail = {0}  ", abortInd.AdditionalDetail));

         return;
      } // DpC2AbortInd


      private void DpHandleEvent()
      {
          lock (_profiSdb)
          {
              DpHandleEvent(_profiSdb);
          }

      }

      /// <summary>
      /// DpHandleEvent
      /// This function is used to identify and display any response message
      /// from the DP master. 
      /// all this function whenever you have read a message from the DP
      /// management device.
      /// </summary>
      private void DpHandleEvent(C_PROFI_SERVICE_DESCR profiSdb)
      {
         string        dpManagementString;
         StringBuilder resultString = new StringBuilder();

         C_DP_ERROR_CON dpError = new C_DP_ERROR_CON();
         ushort nStatus = 0;
         ushort nErrorCode = 0;
         ushort nErrorExtension = 0;

         try
         {
            if (profiSdb != null)
            {
               if (profiSdb.Result == (short)ServiceResult.Pos)
               {
                   if (profiSdb.Primitive == (byte) ServicePrimitive.Ind)
                   {
                      resultString.Append(".ind");
                   }
                   else
                   {
                      resultString.Append(".con [+]");
                   }
               }
               else
               {
                   if (profiSdb.Data is ushort)
                   {
                       dpError.Status = (ushort)profiSdb.Data;
                   }
                   else
                   {
                       dpError = profiSdb.Data as C_DP_ERROR_CON; 
                   }

                   nStatus = dpError.Status;
                   resultString.Append(string.Format(".con [-] {0} ", nStatus));


                     nErrorCode      = (ushort)(nStatus & 0x00FF);
                     nErrorExtension = (ushort)(nStatus & 0xFF00);

                     if (nErrorCode > 0)
                     {
                        switch ((DPServiceResult)nErrorCode)
                        {
                           case DPServiceResult.UE: resultString.Append("remote user error ");               break;
                           case DPServiceResult.RR: resultString.Append("remote resource insufficient ");    break;
                           case DPServiceResult.RS: resultString.Append("remote service/SAP deactivated ");  break;
                           case DPServiceResult.RA: resultString.Append("access of remote SAP blocked ");    break;
                           case DPServiceResult.NA: resultString.Append("no reaction from remote station "); break;
                           case DPServiceResult.DS: resultString.Append("local entity disconnected ");       break;
                           case DPServiceResult.NO: resultString.Append("not possible in this state ");      break;
                           case DPServiceResult.LR: resultString.Append("local resource not available ");    break;
                           case DPServiceResult.IV: resultString.Append("invalid parameter in request ");    break;
                           case DPServiceResult.TO: resultString.Append("timeout expired ");                 break;
                           case DPServiceResult.FE: resultString.Append("format error in request frame ");   break;
                           case DPServiceResult.NI: resultString.Append("function not implemented ");        break;
                           case DPServiceResult.AD: resultString.Append("access denied ");                   break;
                           case DPServiceResult.EA: resultString.Append("area too large ");                  break;
                           case DPServiceResult.LE: resultString.Append("data block length exceeded ");      break;
                           case DPServiceResult.RE: resultString.Append("format error in response frame ");  break;
                           case DPServiceResult.IP: resultString.Append("invalid parameter ");               break;
                           case DPServiceResult.SC: resultString.Append("sequence conflict ");               break;
                           case DPServiceResult.SE: resultString.Append("sequence error ");                  break;
                           case DPServiceResult.NE: resultString.Append("area non-existent ");               break;
                           case DPServiceResult.DI: resultString.Append("data incomplete ");                 break;
                           case DPServiceResult.NC: resultString.Append("not connected ");                   break;
                           default:                 resultString.Append(string.Format("ErrorCode = 0x{0:X2} ", nErrorCode)); break;
                        } // switch nErrorCode
                     } // if nErrorCode

                     if (nErrorExtension > 0)
                     {
                        switch ((DPServiceResult)nErrorExtension)
                        {
                           case DPServiceResult.DataAlignment:      resultString.Append("[DATA_ALIGNMENT] ");        break;
                           case DPServiceResult.TooManySlaves:      resultString.Append("[TOO_MANY_SLAVES] ");       break;
                           case DPServiceResult.WrongSlaveAdd:      resultString.Append("[WRONG_SLAVE_ADD] ");       break;
                           case DPServiceResult.AAMNotSupported:    resultString.Append("[AAM_NOT_SUPPORTED] ");     break;
                           case DPServiceResult.TooFewDiagEntries:  resultString.Append("[TOO_FEW_DIAG_ENTRIES] ");  break;
                           case DPServiceResult.WrongPrmDataLen:    resultString.Append("[WRONG_PRM_DATA_LEN] ");    break;
                           case DPServiceResult.WrongCfgDataLen:    resultString.Append("[WRONG_CFG_DATA_LEN] ");    break;
                           case DPServiceResult.WrongDiagLen:       resultString.Append("[WRONG_DIAG_LEN] ");        break;
                           case DPServiceResult.WrongBusParaLen:    resultString.Append("[WRONG_BUS_PARA_LEN] ");    break;
                           case DPServiceResult.WrongSlaveParaLen:  resultString.Append("[WRONG_SLAVE_PARA_LEN] ");  break;
                           case DPServiceResult.WrongIODataLen:     resultString.Append("[WRONG_IO_DATA_LEN] ");     break;
                           case DPServiceResult.NotEnoughMemory:    resultString.Append("[NOT_ENOUGH_MEMORY] ");     break;
                           case DPServiceResult.WrongUsifState:     resultString.Append("[WRONG_USIF_STATE] ");      break;
                           case DPServiceResult.SlaveAccessDenied:  resultString.Append("[SLAVE_ACCESS_DENIED] ");   break;
                           case DPServiceResult.WrongAreaCode:      resultString.Append("[WRONG_AREA_CODE] ");       break;
                           case DPServiceResult.NotSupported:       resultString.Append("[NOT_SUPPORTED] ");         break;
                           case DPServiceResult.PrmDataFault:       resultString.Append("[PRM_DATA_FAULT] ");        break;
                           case DPServiceResult.CfgDataFault:       resultString.Append("[CFG_DATA_FAULT] ");        break;
                           case DPServiceResult.AATDataFault:       resultString.Append("[AAT_DATA_FAULT] "); 	     break;
                           case DPServiceResult.UserDataFault:      resultString.Append("[USER_DATA_FAULT] ");       break;
                           case DPServiceResult.SlaveParaFault:     resultString.Append("[SLAVE_PARA_FAULT] ");      break;
                           case DPServiceResult.AreaNotAccessed:    resultString.Append("[AREA_NOT_ACCESSED] ");     break;
                           case DPServiceResult.WrongBaudrate:      resultString.Append("[WRONG_BAUDRATE] ");        break;
                           case DPServiceResult.WrongBPFlag:        resultString.Append("[WRONG_BP_FLAG] ");         break;
                           case DPServiceResult.WrongFDLState:      resultString.Append("[WRONG_FDL_STATE] ");       break;
                           case DPServiceResult.WrongActivation:    resultString.Append("[WRONG_ACTIVATION] ");      break;
                           case DPServiceResult.WrongMasterAdd:     resultString.Append("[WRONG_MASTER_ADD] ");      break;
                           case DPServiceResult.DPRAMInitError:     resultString.Append("[DPRAM_INIT_ERROR] ");      break;
                           case DPServiceResult.WrongLen:           resultString.Append("[WRONG_LEN] ");             break;
                           case DPServiceResult.WrongIdentifier:    resultString.Append("[WRONG_IDENTIFIER] ");      break;
                           case DPServiceResult.LoadBusParameter:   resultString.Append("[LOAD_BUS_PARAMETER] ");    break;
                           case DPServiceResult.ActivateSAP:        resultString.Append("[ACTIVATE_SAP] ");          break;
                           case DPServiceResult.WrongRemoteService: resultString.Append("[WRONG_REMOTE_SERVICE] ");  break;
                           case DPServiceResult.IllegalIndex:       resultString.Append("[E_DP_ILLEGAL_INDEX]");     break;
                           case DPServiceResult.IllegalSlot:        resultString.Append("[E_DP_ILLEGAL_SLOT]");      break;
                           case DPServiceResult.IllegalLength:      resultString.Append("[E_DP_ILLEGAL_LENGTH]");    break;
                           case DPServiceResult.IllegalExtention:   resultString.Append("[E_DP_ILLEGAL_EXTENSION]"); break;
                           default:                                 resultString.Append(string.Format("ErrorExtension = 0x{0:X2} ", nErrorExtension)); break;
                        } // switch nErrorExtension
                     } // if nErrorExtension

                     resultString.Append("\r\n");
               } // POS


               switch ((ServicePrimitive)profiSdb.Primitive)
               {
                  // handle confirmations --------------------------------------------------------------
                  case ServicePrimitive.Con:
                  {
                     switch (profiSdb.Service)
                     {
                        case (byte)DPService.InitMaster:
                        {
                           dpManagementString = string.Format("DP_Init_Master{0}", resultString.ToString());
                           Global.OutputString(dpManagementString);
                        }
                        break;

                        case (byte)DPService.ExitMaster:
                        {
                           dpManagementString = string.Format("DP_Exit_Master{0}", resultString.ToString());
                           Global.OutputString(dpManagementString);
                        }
                        break;

                        case (byte)DPService.ActParamLoc:
                        {
                           if (profiSdb.Result == (short)ServiceResult.Pos)
                           {
                              _nCurrentOpMode = _nRequestedOpMode;
                           }

                           dpManagementString = string.Format("DP_Act_Param_Loc{0}", resultString.ToString());
                           Global.OutputString(dpManagementString);
                        }
                        break;

                        case (byte)DPService.DpInitiate:
                        {
                           dpManagementString = string.Format("DP_C2_Initiate{0}", resultString.ToString());
                           Global.OutputString(dpManagementString);
                           if (profiSdb.Result == (short)ServiceResult.Pos)
                           {
                              C_DP_INITIATE_CON initiateCon = profiSdb.Data as C_DP_INITIATE_CON;
                              DpC2InitiateCon(initiateCon);
                           }
                        }
                        break;

                        case (byte)DPService.DpRead:
                        {
                           dpManagementString = string.Format("DP_C2_Read{0}", resultString.ToString());
                           Global.OutputString(dpManagementString);
                           if (profiSdb.Result == (short)ServiceResult.Pos)
                           {
                              C_DP_READ_CON readCon = profiSdb.Data as C_DP_READ_CON;
                              DpC2ReadCon(readCon);
                           }
                        }
                        break;

                        case (byte)DPService.DpWrite:
                        {
                           dpManagementString = string.Format("DP_C2_Write{0}", resultString.ToString());
                           Global.OutputString(dpManagementString);
                           if (profiSdb.Result == (short)ServiceResult.Pos)
                           {
                              C_DP_WRITE_CON writeCon = profiSdb.Data as C_DP_WRITE_CON;
                              DpC2WriteCon(writeCon);
                           }
                        }
                        break;

                        default:
                        {
                           dpManagementString = string.Format("DP: invalid service confirmation = 0x{0:X2}", profiSdb.Service);
                           Global.OutputString(dpManagementString);
                        }
                        break;
                     } //
                     break;
                  } // case CON

                  // handle indications ---------------------------------------------------------------------
                  case ServicePrimitive.Ind:
                  {
                     switch (profiSdb.Service)
                     {
                        case (byte)DPService.DpAbort:
                        {
                           dpManagementString = string.Format("DP_Abort{0}", resultString.ToString());
                           Global.OutputString(dpManagementString);
                           C_DP_ABORT_IND abortInd = profiSdb.Data as C_DP_ABORT_IND;
                           DpC2AbortInd(abortInd);
                        }
                        break;

                        default:
                        {
                           dpManagementString = string.Format("DP: invalid service indication = 0x{0:X2}", profiSdb.Service);
                           Global.OutputString(dpManagementString);
                        }
                        break;
                     } // switch service
                     break;
                  } // case IND

                  default:
                  {
                     dpManagementString = string.Format("DP: invalid service primitive = 0x{0:X2}", profiSdb.Primitive);
                     Global.OutputString(dpManagementString);
                  }
                  break;
               } // switch primitive
            } // if nNumberOfBytesRead
            else
            {
               Global.OutputString("DP: service event with zero length received");

            } // else nNumberOfBytesRead
         }
         catch (Exception exc)
         {
            Global.DisplayError(exc);
         }

         return;

      } // DpHandleEvent

        /****************************************************************************/  
        /// <summary>
        /// DPServiceConIndReceived
        /// DP service configuration indication flag received
        /// </summary>

       private void DPServiceConIndReceived(object sender, EventArgs<C_PROFI_SERVICE_DESCR> e)
       {
           _profiSdb = e.Object as C_PROFI_SERVICE_DESCR;
           Global.EwhDpRead.Set();
        } // DPServiceConIndReceived


        /****************************************************************************/

        /// <summary>
        /// Reads periodically the data from the pipe where it is saved periodically the received
        /// indication/confirmation buffer data
        /// </summary>
        private bool DpReadServiceInterface()
        {
           bool bResult    = true;
           int  waitResult = -1;
           do
           {
              waitResult = WaitHandle.WaitAny(_waitDpReadServiceHandles, 10, false);
              if (0 == waitResult)
              {
                 return true;
              }
              else if (1 == waitResult)
              {
                 Global.EwhExit.Set();
                 return false;
              }
           } while (waitResult == WaitHandle.WaitTimeout);

           return bResult;
        }  // DpReadServiceInterface

    #endregion
    }
}
