﻿using System;
using System.Collections.Generic;
using System.Text;

using PapiWrapper.Common;
using PapiWrapper.Fmb;
using Microsoft.Win32.SafeHandles;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Runtime.Serialization;
using System.Threading;
using System.Runtime.InteropServices;
using System.Security.AccessControl;
using System.Reflection;

using PapiWrapper.Dp;

using GlobalModule;
using FmbModule;
using DpModule;

namespace DpV1Demo
{
   public class DpV1Demo
    {
        #region Fields

        private Global _global;

        public  Thread _keyboardThread;
        private Thread _readServiceInterfaceThread;
        private Thread _basicManagementThread;

        private static DpV1Demo _instance;

        private static WaitHandle[] _waitReadServiceHandles;

        #endregion

        #region Constructor
        private DpV1Demo()
        {
            _global = Global.Instance;
            _waitReadServiceHandles = new WaitHandle[2] { Global.EwhRead, Global.EwhExit };
        }
        #endregion

        #region Properties
        public static DpV1Demo Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new DpV1Demo();
                }
                return _instance;
            }
        }
        public static int CommandSetBaudrate { get; set; }

        #endregion

        #region Public methods
        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// The main function controls the initialization and termination of the
        /// program.
        /// The BasicManagementThread starts the whole PROFIBUS activity.
        /// The KeyboardThread checks the keyboard for a termination character to
        /// quit the program at any time.
        /// If one of these two threads terminates then the whole application will
        /// be terminated.
        /// This function is responsible for calling the nested initialization and
        /// termination routines of the sub modules.
        ///
        /// -----------------------------------------------------------------------
        /// </summary>
        public void Run(string[] args)
        {
            if (InitDpV1Demo())
            {
                if (ParseCommandLine(args))
                {
                    ShowProgramInfo();
                    InitFmbLayer();

                    if (!OpenProfibusApi(2000,0))
                    {
                        Global.OutputString("Initializing PROFIBUS controller failed.");
                        DpLayer.Instance.TerminateDpLayer();
                        FmbLayer.Instance.TerminateFmbLayer();

                        TerminateDpV1Demo();
                        return;
                    }

                    /*--- show board / release information ---------------------------------*/
                    ShowReleaseInfo();

                    /*--- show board / release information ---------------------------------*/
                    _keyboardThread             = new Thread(KeyboardThread);
                    _readServiceInterfaceThread = new Thread(ReadServiceInterfaceThread);
                    _basicManagementThread      = new Thread(FmbLayer.Instance.BasicManagementThread);

                    _readServiceInterfaceThread.Start();
                    _basicManagementThread.Start();


                    _keyboardThread.Start();


                    _keyboardThread.Join();
                    _readServiceInterfaceThread.Join();
                    _basicManagementThread.Join();

                    /*--- terminate sub modules and free globals -----------------------------*/
                    DpLayer.Instance.TerminateDpLayer();
                    FmbLayer.Instance.TerminateFmbLayer();

                    CloseProfibusApi();

                    TerminateDpV1Demo();
                }
            }
        }

        #endregion

        #region Private methods

        /// <summary>
        /// ------------------------------------------------------------------------
        /// FUNCTIONAL_DESCRIPTION
        /// ------------------------------------------------------------------------
        /// The DpDemo application allows several command line parameters.
        /// Please refer to function ShowCommandLineUsage for details.
        /// ------------------------------------------------------------------------
        /// </summary>
        private bool ParseCommandLine(string[] args)
        {
            int argc = args.Length;
            if (argc > 4)
            {
              ShowCommandLineUsage(args[0]);
            }

            for (int argIndex = 0; argIndex < argc; argIndex++)
            {
                if (args[argIndex].StartsWith("-b"))
                {
                    Global.BoardNumber = Convert.ToByte(args[argIndex].Remove(0,2));
                }
                else if (args[argIndex].StartsWith("-m"))
                {
                    Global.MasterAddr = Convert.ToByte(args[argIndex].Remove(0, 2));
                }
                else if (args[argIndex] == "-r0")
                {
                    CommandSetBaudrate = 0x00;  // 9.6 kBaud
                }
                else if (args[argIndex] == "-r1")
                {
                    CommandSetBaudrate = 0x01;  // 19.2 kBaud
                }
                else if (args[argIndex] == "-r2")
                {
                    CommandSetBaudrate = 0x02;  // 93.75 kBaud
                }
                else if (args[argIndex] == "-r2PF")
                {
                    CommandSetBaudrate = 0x0C;  // 93.75 kBaud
                }
                else if (args[argIndex] == "-r3")
                {
                    CommandSetBaudrate = 0x03;  // 187.5 kBaud
                }
                else if (args[argIndex] == "-r4")
                {
                    CommandSetBaudrate = 0x04;  // 500 kBaud
                }
                else if (args[argIndex] == "-r5")
                {
                    CommandSetBaudrate = 0x05;  // 750 kBaud
                }
                else if (args[argIndex] == "-r6")
                {
                    CommandSetBaudrate = 0x06;  // 1.5 MBaud
                }
                else if (args[argIndex] == "-r7")
                {
                    CommandSetBaudrate = 0x07;  // 3 MBaud
                }
                else if (args[argIndex] == "-r8")
                {
                    CommandSetBaudrate = 0x08;  // 6 MBaud
                }
                else if (args[argIndex] == "-r9")
                {
                    CommandSetBaudrate = 0x09;  // 12 MBaud
                }
                else if (args[argIndex] == "-rB")
                {
                    CommandSetBaudrate = 0x0B;  // 45.45 kBaud
                }
                else if (args[argIndex].Contains("-h") ||
                         args[argIndex].Contains("-?") ||
                         args[argIndex][0] == '-')
                {
                    ShowCommandLineUsage(args[0]);
                    return false;
                }
           }

           return true;
        }  /* ParseCommandLine */

        /****************************************************************************/
        private void ShowCommandLineUsage(string  pProgramPath)
        {
            Console.WriteLine("usage: [-b<BoardNumber>] [-m<MasterAddress>] [-r<baudrate>] [binfile]\n");//, pProgramPath);
            Console.WriteLine("       -b       use PROFIBUS board with \"Board Number\"\n");
            Console.WriteLine("       -m       master address \n");
            Console.WriteLine("       -r       baudrate  (default 1.5 MBaud)                   \n");
            Console.WriteLine("                -r0 (  9.6 kBaud)   -r1 ( 19.2 kBaud)   -r2   (93.75 kBaud) \n");
            Console.WriteLine("                                                        -r2PF (93.75 kBaud Pepperl&Fuchs) \n");
            Console.WriteLine("                -r3 (187.5 kBaud)   -r4 (  500 kBaud)   -r6   (  1.5 MBaud) \n");
            Console.WriteLine("                -r7 (    3 MBaud)   -r8 (    6 MBaud)   -r9   (   12 MBaud) \n");
            Console.WriteLine("       -h -?    show this message\n");
            Console.WriteLine("\n");
        } /* ShowCommandLineUsage */

        /****************************************************************************/

      private void ShowProgramInfo()
      {
        Global.OutputString("DpV1Demo V5.46 Copyright (C) SOFTING Industrial Automation GmbH 2011-2012");
        Global.OutputString(string.Format("Board Number:     {0}",Global.BoardNumber));
        Global.OutputString(string.Format("Master Address:   {0}",Global.MasterAddr));


        switch (CommandSetBaudrate)
        {
          case 0x00: Global.OutputString("Baudrate:         9.6 kBaud") ; break;
          case 0x01: Global.OutputString("Baudrate:         19.2 kBaud"); break;
          case 0x02: Global.OutputString("Baudrate:         93.75 kBaud"); break;
          case 0x03: Global.OutputString("Baudrate:         187.5 kBaud"); break;
          case 0x04: Global.OutputString("Baudrate:         500 kBaud"); break;
          case 0x05: Global.OutputString("Baudrate:         750 kBaud"); break;
          case 0x06: Global.OutputString("Baudrate:         1.5 MBaud"); break;
          case 0x07: Global.OutputString("Baudrate:         3 MBaud"); break;
          case 0x08: Global.OutputString("Baudrate:         6 MBaud"); break;
          case 0x09: Global.OutputString("Baudrate:         12 MBaud"); break;
          case 0x0B: Global.OutputString("Baudrate:         45.45 kBaud"); break;
          case 0x0C: Global.OutputString("Baudrate:         93.75 kBaud Pepperl&Fuchs"); break;
          default:   Global.OutputString("wrong Baudrate"); break;
        }

      } /* ShowProgramInfo */


      private void InitFmbLayer()
      {
        FmbLayer.Instance.InitFmbLayer(CommandSetBaudrate);
      }

      private bool InitDpV1Demo()
      {
         CommandSetBaudrate = 0x6; // 1.5 MB by default

         bool bResult = Global.OpenPapiAndLogger();
         if (bResult)
         {
            Global.OutputSetThreadIdString(" MAIN");
         }
         return bResult;
      }


      static bool bStopped;
      private void KeyboardThread()
      {
        Global.OutputSetThreadIdString("  KEY");
        Global.OutputString("KeyboardThread started");

        System.Threading.Thread.Sleep(4000);

        Console.WriteLine("");
        Console.WriteLine(" <I> ....... Initialize Connection");
        Console.WriteLine(" <R> ....... Read Object in Slave ");
        Console.WriteLine(" <W> ....... Write Object in Slave");
        Console.WriteLine(" <A> ....... Abort Connection     ");
        Console.WriteLine(" <X> ....... Exit                 ");

        while (true)
        {
           ConsoleKeyInfo pressedKey = Console.ReadKey(true);
           if (pressedKey.KeyChar == 'Q' || pressedKey.KeyChar == 'X' ||
               pressedKey.KeyChar == 'q' || pressedKey.KeyChar == 'x')
           {
             bStopped = true;
             Global.EwhExit.Set();
             break;
           }
           else if (pressedKey.KeyChar == 'I' || pressedKey.KeyChar == 'i')
           {
              byte   remAdd;
              UInt16 sendTimeout;
              Console.WriteLine("");
              Console.WriteLine(" Initiate ... ");
              Console.Write(" Enter slave address ");
              byte.TryParse(Console.ReadLine(), out remAdd);
              Console.Write(" Enter send timeout  ");
              UInt16.TryParse(Console.ReadLine(), out sendTimeout);

              if (!DpLayer.Instance.DpC2InitiateReq(remAdd,sendTimeout))
              {
                Global.DisplayLastError();
              }

           }
           else if (pressedKey.KeyChar == 'R' || pressedKey.KeyChar == 'r')
           {
               byte remAdd;
               byte slotNumber;
               byte index;
               byte length;

               Console.WriteLine("");
               Console.WriteLine(" Read ... \n");
               Console.Write(" Enter slave address ");
               byte.TryParse(Console.ReadLine(), out remAdd);
               Console.Write(" Enter slot number   ");
               byte.TryParse(Console.ReadLine(), out slotNumber);
               Console.Write(" Enter index         ");
               byte.TryParse(Console.ReadLine(), out index);
               Console.Write(" Enter bytes to read ");
               byte.TryParse(Console.ReadLine(), out length);

               DpLayer.Instance.DpC2ReadReq(remAdd,slotNumber,index,length);
           }
           else if (pressedKey.KeyChar == 'W' || pressedKey.KeyChar == 'w')
           {
               byte remAdd;
               byte slotNumber;
               byte index;
               byte length;

               Console.WriteLine("");
               Console.WriteLine(" Write ... \n");
               Console.Write(" Enter slave address ");
               byte.TryParse(Console.ReadLine(), out remAdd);
               Console.Write(" Enter slot number   ");
               byte.TryParse(Console.ReadLine(), out slotNumber);
               Console.Write(" Enter index         ");
               byte.TryParse(Console.ReadLine(), out index);
               Console.Write(" Enter bytes to write ");
               byte.TryParse(Console.ReadLine(), out length);

               DpLayer.Instance.DpC2WriteReq(remAdd, slotNumber, index, length);
           }
           else if (pressedKey.KeyChar == 'A' || pressedKey.KeyChar == 'a')
           {
              DpLayer.Instance.DpC2AbortReq();
           }
           else if (pressedKey.KeyChar == ' ' || pressedKey.KeyChar == ' ')
           {
              Console.WriteLine("");
              Console.WriteLine(" <I> ....... Initialize Connection ");
              Console.WriteLine(" <R> ....... Read Object in Slave  ");
              Console.WriteLine(" <W> ....... Write Object in Slave ");
              Console.WriteLine(" <A> ....... Abort Connection      ");
              Console.WriteLine(" <X> ....... Exit                  ");
           }
           else
           {
           }

         }

         Global.OutputString("KeyboardThread terminated");

       }

       public void ReadServiceInterfaceThread()
       {
          Global.OutputSetThreadIdString("RDSRV");
          Global.OutputString("ReadServiceInterfaceThread started");

          while (!bStopped)
          {
            int waitResult = WaitHandle.WaitAny(_waitReadServiceHandles, 10, false);
            if (waitResult == 0)
            {
               Global.ReadServiceInterface();
            }
            else if (waitResult == 1)
            {
              Global.EwhExit.Set();
              break;
            }
            else if (waitResult == WaitHandle.WaitTimeout)
            {
              Global.EwhRead.Set();
            }
          }

          Global.OutputString("ReadServiceInterfaceThread terminated");

       } /* ReadServiceInterfaceThread */


       public void ShowReleaseInfo()
       {
           Global.OutputString("PAPI Version:     " + Global.GetPapiVersion().Substring(0, 50));
           Global.OutputString("Firmware Version: " + Global.GetFirmwareVersion().Substring(0, Math.Min(50, Global.GetFirmwareVersion().Length)));
           Global.OutputString("");
       } // ShowReleaseInfo


       private bool OpenProfibusApi(uint readTimeout, uint writeTimeout)
       {
           return Global.PapiInit(Global.BoardNumber, readTimeout, writeTimeout);
       } /* OpenProfibusDevices */


       private void CloseProfibusApi()
       {
          Global.PapiEnd(Global.BoardNumber);

       }


       private void TerminateDpV1Demo()
       {
          Global.OutputString("main thread terminated");
          Global.CloseLoggerAndHandlers();

       }

       #endregion
    }
}
